/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "forward.h"
#include "canvasuid.h"

#include "session.h"
#include "views.h"

namespace Crow {

// UIDefinitionCanvasEditorWidget ----------------------------------------------

UIDefinitionCanvasEditorWidget::UIDefinitionCanvasEditorWidget() :
	Glib::ObjectBase(typeid(UIDefinitionCanvasEditorWidget))
{

	actionGroup= createActionGroup();
	uiManager= createUIManager(actionGroup);

	add(box);
	box.show();

	Gtk::Toolbar * toolbar= dynamic_cast<Gtk::Toolbar *>(uiManager->get_widget("/Toolbar"));
	toolbar->set_toolbar_style(Gtk::TOOLBAR_ICONS);
	toolbar->set_orientation(Gtk::ORIENTATION_VERTICAL);
	box.pack_start(*toolbar, Gtk::PACK_SHRINK);
	toolbar->show();

	scroll.set_policy(Gtk::POLICY_NEVER, Gtk::POLICY_ALWAYS);
	box.pack_end(scroll, Gtk::PACK_EXPAND_WIDGET);
	scroll.show();

	treeView.setImmediateEditing(false);
	scroll.add(treeView);
	treeView.show();

	treeView.createModel(3);
	treeView.addColumn(0, _("Element"));
	treeView.addColumn(1, _("Name"));
	treeView.addColumn(2, _("Action"));

	treeView.get_column_cell_renderer(0)->property_ypad()= 0;
	treeView.get_column_cell_renderer(1)->property_ypad()= 0;
	treeView.get_column_cell_renderer(2)->property_ypad()= 0;

}

PGtkActionGroup UIDefinitionCanvasEditorWidget::getActions() {
	return actionGroup;
}

PGtkUIManager UIDefinitionCanvasEditorWidget::getManager() {
	return uiManager;
}

PGtkActionGroup UIDefinitionCanvasEditorWidget::createActionGroup() {

	PGtkActionGroup actions= Gtk::ActionGroup::create();

	actions->add(Gtk::Action::create("New", Gtk::Stock::NEW, "", _("Create a new child")));
	actions->add(Gtk::Action::create("Add", Gtk::Stock::ADD, "", _("Create a new sibling")));
	actions->add(Gtk::Action::create("Remove", Gtk::Stock::REMOVE, "", _("Remove the element")));
	actions->add(Gtk::Action::create("MoveUp", Gtk::Stock::GO_UP, "", _("Move the element up")));
	actions->add(Gtk::Action::create("MoveDown", Gtk::Stock::GO_DOWN, "", _("Move the element down")));

	actions->add(Gtk::Action::create("Cancel", _("Cancel")));
	actions->add(Gtk::Action::create("menubar", "menubar"));
	actions->add(Gtk::Action::create("toolbar", "toolbar"));
	actions->add(Gtk::Action::create("popup", "popup"));
	actions->add(Gtk::Action::create("accelerator", "accelerator"));
	actions->add(Gtk::Action::create("menuitem", "menuitem"));
	actions->add(Gtk::Action::create("separator", "separator"));
	actions->add(Gtk::Action::create("placeholder", "placeholder"));
	actions->add(Gtk::Action::create("menu", "menu"));
	actions->add(Gtk::Action::create("toolitem", "toolitem"));

	return actions;
}

PGtkUIManager UIDefinitionCanvasEditorWidget::createUIManager(PGtkActionGroup actions) {

	PGtkUIManager manager= Gtk::UIManager::create();
	manager->insert_action_group(actions);

	Glib::ustring uistr=
		"<ui>"
		"  <toolbar name='Toolbar'>"
		"    <toolitem action='New'/>"
		"    <toolitem action='Add'/>"
		"    <toolitem action='Remove'/>"
		"    <toolitem action='MoveUp'/>"
		"    <toolitem action='MoveDown'/>"
		"  </toolbar>"
		"  <popup name='NewUiChild'>"
		"    <menuitem action='Cancel'/>"
		"    <separator/>"
		"    <menuitem action='menubar'/>"
		"    <menuitem action='toolbar'/>"
		"    <menuitem action='popup'/>"
		"    <menuitem action='accelerator'/>"
		"  </popup>"
		"  <popup name='NewMenuChild'>"
		"    <menuitem action='Cancel'/>"
		"    <separator/>"
		"    <menuitem action='menuitem'/>"
		"    <menuitem action='separator'/>"
		"    <menuitem action='placeholder'/>"
		"    <menuitem action='menu'/>"
		"  </popup>"
		"  <popup name='NewToolbarChild'>"
		"    <menuitem action='Cancel'/>"
		"    <separator/>"
		"    <menuitem action='toolitem'/>"
		"    <menuitem action='separator'/>"
		"    <menuitem action='placeholder'/>"
		"  </popup>"
		"  <popup name='NewPlaceholderChild'>"
		"    <menuitem action='Cancel'/>"
		"    <separator/>"
		"    <menuitem action='menuitem'/>"
		"    <menuitem action='toolitem'/>"
		"    <menuitem action='separator'/>"
		"    <menuitem action='placeholder'/>"
		"    <menuitem action='menu'/>"
		"  </popup>"
		"</ui>"
	;

	try {
		manager->add_ui_from_string(uistr);
	} catch(const Glib::Error & x) {
		FatalError(_("Creating UIManager failed: ") + x.what());
	}

	return manager;
}

// UIDefinitionCanvasEditor ----------------------------------------------------

UIElement::UIElement(
	const std::vector<int> & path,
	string_cr element,
	string_cr name,
	string_cr action,
	bool expanded
) {
	remove= false;
	this->path= path;
	this->element= element;
	this->name= name;
	this->action= action;
	this->expanded= expanded;
}

bool UIElement::equals(PObject obj) {
	PUIElement elem= PUIElement::cast_dynamic(obj);
	if(!elem)
		return false;
	return
		path==elem->path &&
		element==elem->element &&
		name==elem->name &&
		action==elem->action
	;
}

UIDefinitionCanvasEditor::UIDefinitionCanvasEditor() {
	lock= false;
}

void UIDefinitionCanvasEditor::created() {
	PUIDefinitionCanvasEditorWidget ew= getEditorWidget();
	ew->getActions()->get_action("New")->signal_activate().connect(sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onActionNew));
	ew->getActions()->get_action("Add")->signal_activate().connect(sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onActionAdd));
	ew->getActions()->get_action("Remove")->signal_activate().connect(sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onActionRemove));
	ew->getActions()->get_action("MoveUp")->signal_activate().connect(sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onActionMoveUp));
	ew->getActions()->get_action("MoveDown")->signal_activate().connect(sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onActionMoveDown));
	ew->getActions()->get_action("menubar")->signal_activate().connect(sigc::bind<const char *>(sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onActionElement), "menubar"));
	ew->getActions()->get_action("toolbar")->signal_activate().connect(sigc::bind<const char *>(sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onActionElement), "toolbar"));
	ew->getActions()->get_action("popup")->signal_activate().connect(sigc::bind<const char *>(sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onActionElement), "popup"));
	ew->getActions()->get_action("accelerator")->signal_activate().connect(sigc::bind<const char *>(sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onActionElement), "accelerator"));
	ew->getActions()->get_action("menuitem")->signal_activate().connect(sigc::bind<const char *>(sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onActionElement), "menuitem"));
	ew->getActions()->get_action("separator")->signal_activate().connect(sigc::bind<const char *>(sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onActionElement), "separator"));
	ew->getActions()->get_action("placeholder")->signal_activate().connect(sigc::bind<const char *>(sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onActionElement), "placeholder"));
	ew->getActions()->get_action("menu")->signal_activate().connect(sigc::bind<const char *>(sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onActionElement), "menu"));
	ew->getActions()->get_action("toolitem")->signal_activate().connect(sigc::bind<const char *>(sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onActionElement), "toolitem"));
	ew->getTreeView()->connect_selection_changed(sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onSelectionChanged));
	ew->getTreeView()->signal_row_collapsed().connect(sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onRowCollapsed));
	ew->getTreeView()->signal_row_expanded().connect(sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onRowExpanded));

	ew->getTreeView()->setColumnWidth(0, getController()->getContext()->getKeyValue<int>("$uidce_col1"));
	ew->getTreeView()->setColumnWidth(1, getController()->getContext()->getKeyValue<int>("$uidce_col2"));

	ew->getTreeView()->setSetRowValues(sigc::mem_fun(*this, &UIDefinitionCanvasEditor::setRowValues));

}

void UIDefinitionCanvasEditor::load(bool reload) {
	Any value= getManager()->getModel()->getScalar(getSession()->getNode1(), "ui");
	fromString(value->toString());
	//dump("reload()");
	updateTree(reload);
	updateActions();
	if(!reload) {
		for(Elements::iterator it= elements.begin(); it!=elements.end(); ++it)
			it->second->expanded= false;
		PObject state= getManager()->getStates()->get(getSession()->getNode1());
		if(state)
			getEditorWidget()->getTreeView()->loadState(state);
		else
			getEditorWidget()->getTreeView()->expandAll();
	}
}

void UIDefinitionCanvasEditor::storeSettings() {
	Polytree * treeView= getEditorWidget()->getTreeView();
	int cw1= treeView->getColumnWidth(0);
	int cw2= treeView->getColumnWidth(1);
	if(cw1!=0 && cw2!=0) {
		getController()->getContext()->setKeyValue("$uidce_col1", cw1);
		getController()->getContext()->setKeyValue("$uidce_col2", cw2);
	}
}

void UIDefinitionCanvasEditor::deleted() {
	Polytree * treeView= getEditorWidget()->getTreeView();
	getManager()->getStates()->set(getSession()->getNode1(), treeView->saveState());
	treeView->prepare();
	treeView->submit();
}

PUIElement UIDefinitionCanvasEditor::getSelected() {
	const std::vector<PObject> & selection= *getEditorWidget()->getTreeView()->getSelection();
	PUIElement selected= selection.empty() ? PUIElement() : PUIElement::cast_dynamic(selection.front());
	return selected;
}

PUIElement UIDefinitionCanvasEditor::getParent(PUIElement elem) {
	CHECK(elem);
	std::vector<int> path= elem->path;
	CHECK(!path.empty());
	path.pop_back();
	return path.empty() ? PUIElement() : getElement(path);
}

void UIDefinitionCanvasEditor::onSelectionChanged() {
	updateActions();
}

void UIDefinitionCanvasEditor::updateActions() {
	PUIElement selected= getSelected();
	PGtkActionGroup actions= getEditorWidget()->getActions();
	actions->get_action("New")->set_sensitive(
		!selected || selected->element=="menubar" || selected->element=="menu" ||
		selected->element=="popup" || selected->element=="toolbar" || selected->element=="placeholder"
			 || selected->element=="toolitem"
	);
	actions->get_action("Add")->set_sensitive(selected);
	actions->get_action("Remove")->set_sensitive(selected);
	actions->get_action("MoveUp")->set_sensitive(selected && selected->path.back() > 0);
	if(!selected)
		actions->get_action("MoveDown")->set_sensitive(false);
	else {
		std::vector<int> path= selected->path;
		++path.back();
		actions->get_action("MoveDown")->set_sensitive(getElement(path));
	}
}

bool UIDefinitionCanvasEditor::anchestry(PUIElement obj, string_cr element) {
	std::vector<int> path= obj->path;
	while(!path.empty()) {
		if(getElement(path)->element==element)
			return true;
		path.pop_back();
	}
	return false;
}

void UIDefinitionCanvasEditor::updateElementActions(PUIElement parent) {
	PGtkActionGroup actions= getEditorWidget()->getActions();
	actions->get_action("menubar")->set_sensitive(!parent);
	actions->get_action("toolbar")->set_sensitive(!parent);
	actions->get_action("popup")->set_sensitive(!parent);
	actions->get_action("accelerator")->set_sensitive(!parent);
	actions->get_action("menuitem")->set_sensitive(
		parent && (anchestry(parent, "menubar") || anchestry(parent, "popup") || anchestry(parent, "toolitem"))
	);
	actions->get_action("separator")->set_sensitive(
		parent && parent->element!="menuitem" && parent->element!="toolitem" &&
		parent->element!="separator" && parent->element!="accelerator"
	);
	actions->get_action("placeholder")->set_sensitive(parent && (
		parent->element=="menubar" || parent->element=="menu" ||
		parent->element=="popup" || parent->element=="toolbar" || parent->element=="placeholder"
	));
	actions->get_action("menu")->set_sensitive(
		parent && (anchestry(parent, "menubar") || anchestry(parent, "menu")
			|| anchestry(parent, "popup") || anchestry(parent, "toolitem"))
	);
	actions->get_action("toolitem")->set_sensitive(parent && anchestry(parent, "toolbar"));
}

void UIDefinitionCanvasEditor::popupNewChild(PUIElement parent) {
	std::string menu;
	if(!parent)
		menu= "/NewUiChild";
	else if(
		parent->element=="menubar" || parent->element=="menu" ||
		parent->element=="popup" || parent->element=="toolitem"
	)
		menu= "/NewMenuChild";
	else if(parent->element=="toolbar")
		menu= "/NewToolbarChild";
	else if(parent->element=="placeholder")
		menu= "/NewPlaceholderChild";
	else
		CHECK(false);
	dynamic_cast<Gtk::Menu *>(getEditorWidget()->getManager()->get_widget(menu))->popup(0, GDK_CURRENT_TIME);
}

void UIDefinitionCanvasEditor::onActionNew() {
	PUIElement parent= getSelected();
	actionElement= parent ? parent->path : UiPath();
	updateElementActions(parent);
	popupNewChild(parent);
}

void UIDefinitionCanvasEditor::onActionAdd() {
	PUIElement parent= getParent(getSelected());
	actionElement= parent ? parent->path : UiPath();
	updateElementActions(parent);
	popupNewChild(parent);
}

void UIDefinitionCanvasEditor::onActionRemove() {
	PUIElement selected= getSelected();
	std::vector<int> spath= selected->path;
	std::vector<int> bot= selected->path;
	++bot.back();
	cut(spath, false);
	std::vector<PUIElement> vbot= cut(bot, true);
	paste(vbot, -1);
	if(!getElement(spath)) {
		if(spath.back() > 0)
			--spath.back();
		else
			spath.pop_back();
	}
	updateModel(spath);
}

void UIDefinitionCanvasEditor::onActionMoveUp() {
	PUIElement selected= getSelected();
	std::vector<int> top= selected->path;
	--top.back();
	const std::vector<int> & bot= selected->path;
	std::vector<PUIElement> vtop= cut(top, false);
	std::vector<PUIElement> vbot= cut(bot, false);
	paste(vtop, 1);
	paste(vbot, -1);
	updateModel(top);
}

void UIDefinitionCanvasEditor::onActionMoveDown() {
	PUIElement selected= getSelected();
	const std::vector<int> & top= selected->path;
	std::vector<int> bot= selected->path;
	++bot.back();
	std::vector<PUIElement> vtop= cut(top, false);
	std::vector<PUIElement> vbot= cut(bot, false);
	paste(vtop, 1);
	paste(vbot, -1);
	updateModel(bot);
}

void UIDefinitionCanvasEditor::onActionElement(const char * element) {
	PUIElement parent= actionElement.empty() ? PUIElement() : getElement(actionElement);
	PUIElement selected= getSelected();
	std::vector<int> path= selected ? selected->path : std::vector<int>();
	if(selected && parent != selected)
		++path.back();
	else
		path.push_back(0);
	std::vector<PUIElement> vec= cut(path, true);
	setElement(path, MakeRefPtr(new UIElement(path, element, "", "")));
	paste(vec, 1);
	updateModel(path);
}

std::vector<PUIElement> UIDefinitionCanvasEditor::cut(const UiPath & path, bool siblings) {
	std::vector<PUIElement> result;
	UiPath spath= path;
	for(;;) {
		Elements::iterator it= elements.find(spath);
		if(it==elements.end())
			break;
		for(Elements::iterator cit= it; ; ) {
			result.push_back(cit->second);
			cit= EraseAndNext(elements, cit);
			if(cit==elements.end() || !Begins(cit->second->path, spath))
				break;
		}
		if(!siblings)
			break;
		++spath.back();
	}
	return result;
}

void UIDefinitionCanvasEditor::paste(const std::vector<PUIElement> & vec, int shift) {
	if(!vec.empty()) {
		int index= vec.front()->path.size() - 1;
		for(int i=0; i<SIGNED(vec.size()); ++i) {
			std::vector<int> path= vec[i]->path;
			path[index] += shift;
			setElement(path, MakeRefPtr(new UIElement(
				path, vec[i]->element, vec[i]->name, vec[i]->action, vec[i]->expanded
			)));
		}
	}
}

void UIDefinitionCanvasEditor::updateModel(const std::vector<int> & path) {
	std::string str= toString();
	Any value= CAny::createString(str);
	if(!value->equals(getManager()->getModel()->getScalar(getSession()->getNode1(), "ui"))) {
		getManager()->begin();
		getManager()->getModel()->setScalar(getSession()->getNode1(), "ui", value);
		getManager()->commit();
	} else
		updateTree();
	PUIElement elem= getElement(path);
	if(elem)
		getEditorWidget()->getTreeView()->setSelection(std::vector<PObject>(1, elem), false);
	else
		getEditorWidget()->getTreeView()->setSelection(std::vector<PObject>(), false);
	updateActions();
}

std::string UIDefinitionCanvasEditor::toString() {

	std::string result;
	std::vector<PUIElement> stack;

	result += "<ui>\n";

	for(Elements::iterator it= elements.begin(); it!=elements.end(); ++it) {

		PUIElement elem= it->second;
		Elements::iterator next= it;
		++next;
		bool push= next!=elements.end() && next->second->path.size() > elem->path.size();
		int pop= next==elements.end() ? elem->path.size()-1 : elem->path.size() - next->second->path.size();

		result += std::string(stack.size()+1, '\t') + "<" + elem->element;

		if(push)
			stack.push_back(elem);

		std::string name= elem->name;
		std::string action= elem->action;
		if(!name.empty()) {
			CHECK(ValidName(name));
			result += " name='" + name + "'";
		}
		if(!action.empty()) {
			CHECK(ValidName(action));
			result += " action='" + action + "'";
		}
		result += push ? ">\n" : "/>\n";

		for(int i=0; i<pop; ++i) {
			result += std::string(stack.size(), '\t') + "</" + stack.back()->element + ">\n";
			stack.pop_back();
		}

	}

	result += "</ui>\n";

	return result;
}

void UIDefinitionCanvasEditor::fromString(string_cr str) {

	for(Elements::iterator it= elements.begin(); it!=elements.end(); ++it)
		it->second->remove= true;

	pstack.push_back(-1);

	try {
		Parser::parse(this, str);
	} catch(const Glib::MarkupError & x) {
		FatalError(_("Parsing failed in UIDCE::fromString(): ") + x.what());
	}

	pstack.pop_back();

	for(Elements::iterator it= elements.begin(); it!=elements.end(); )
		if(it->second->remove)
			it= EraseAndNext(elements, it);
		else
			++it;

}

void UIDefinitionCanvasEditor::onStartElement(ustring_cr element_name, const Glib::Markup::Parser::AttributeMap & attributes) {
	if(element_name=="ui")
		return;
	++pstack.back();
	Glib::Markup::Parser::AttributeMap::const_iterator iname= attributes.find("name");
	Glib::Markup::Parser::AttributeMap::const_iterator iaction= attributes.find("action");
	std::string name= iname==attributes.end() ? "" : iname->second;
	std::string action= iaction==attributes.end() ? "" : iaction->second;
	PUIElement elem= MakeRefPtr(new UIElement(pstack, element_name, name, action));
	Elements::iterator it= elements.find(pstack);
	PUIElement curr= it==elements.end() ? PUIElement() : it->second;
	if(curr) {
		if(!elem->equals(curr)) {
			elements.erase(it);
			setElement(pstack, elem);
		} else
			curr->remove= false;
	} else
		setElement(pstack, elem);
	pstack.push_back(-1);
}

void UIDefinitionCanvasEditor::onEndElement(ustring_cr element_name) {
	if(element_name=="ui")
		return;
	pstack.pop_back();
}

void UIDefinitionCanvasEditor::updateTree(bool expand) {
	CHECK(!lock);
	lock= true;
	Polytree * treeView= getEditorWidget()->getTreeView();
	treeView->prepare();
	for(Elements::iterator it= elements.begin(); it!=elements.end(); ++it)
		treeView->update(IdPath(it->first.begin(), it->first.end()), it->second);
	treeView->submit();
	if(expand)
		for(Elements::iterator it= elements.begin(); it!=elements.end(); ++it)
			if(it->second->expanded)
				treeView->expandRow(it->second);
	lock= false;
}

void UIDefinitionCanvasEditor::setRowValues(PPolyelem obj) {

	PUIElement elem= PUIElement::cast_dynamic(obj->getObject());

	PPolycellText pcelement= MakeRefPtr<PolycellText>();
	pcelement->setText(elem->element);
	obj->setValue(0, pcelement);

	Glib::RefPtr<PolycellInput> pcname= MakeRefPtr<PolycellInput>();
	pcname->setText(elem->name);
	pcname->setAllowEditing(true);
	pcname->setAutoStop(true);
	pcname->connect_accepted(sigc::bind<PolycellInput *, PUIElement>(
		sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onSetName), GetPtr(pcname), elem
	));
	pcname->setValidator(sigc::mem_fun(*this, &UIDefinitionCanvasEditor::validator));
	obj->setValue(1, pcname);

	Glib::RefPtr<PolycellInput> pcaction= MakeRefPtr<PolycellInput>();
	pcaction->setText(elem->action);
	pcaction->setAllowEditing(true);
	pcaction->setAutoStop(true);
	pcaction->setPopupEnabled(true);
	pcaction->connect_accepted(sigc::bind<PolycellInput *, PUIElement>(
		sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onSetAction), GetPtr(pcaction), elem
	));
	pcaction->connect_popup(sigc::bind<PUIElement>(
		sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onPopupActions), elem
	));
	pcaction->setValidator(sigc::mem_fun(*this, &UIDefinitionCanvasEditor::validator));
	obj->setValue(2, pcaction);

}

bool UIDefinitionCanvasEditor::validator(ustring_cr oldtext, ustring_cr newtext) {
	return newtext.empty() || ValidName(newtext);
}

void UIDefinitionCanvasEditor::onSetName(PolycellInput * polycell, PUIElement elem) {
	std::string name= *polycell->getText();
	elem->name= name;
	updateModel(elem->path);
}

void UIDefinitionCanvasEditor::onSetAction(PolycellInput * polycell, PUIElement elem) {
	elem->action= *polycell->getText();
	updateModel(elem->path);
}

void UIDefinitionCanvasEditor::onPopupActions(PUIElement elem) {
	Gtk::Menu * menu= Gtk::manage(new Gtk::Menu());
	menu->items().push_back(Gtk::Menu_Helpers::MenuElem(_("Cancel")));

	Model * model= getManager()->getModel();
	PNode manager= model->findMaster(getSession()->getNode1());
	std::set<std::string> choices;
	if(manager) {

		PNode groups= model->find(manager, "action-groups");
		const Nodes & out= *groups->getOut();
		for(Nodes::const_iterator it= out.begin(); it!=out.end(); ++it) {
			PNode group= (*it)->getLink();
			if(group) {
				PNode actions= model->find(group, "actions");
				const Nodes & out= *actions->getOut();
				for(Nodes::const_iterator it= out.begin(); it!=out.end(); ++it) {
					PNode action= *it;
					if(action->getRole()==nrEntity) {
						PNode name= model->find(action, "name");
						choices.insert(*name->getValue()->getString());
					}
				}
			}
		}

		choices.erase(elem->action);

	}

	if(!choices.empty()) {
		menu->items().push_back(Gtk::Menu_Helpers::SeparatorElem());
		for(std::set<std::string>::iterator it= choices.begin(); it!=choices.end(); ++it)
			menu->items().push_back(Gtk::Menu_Helpers::MenuElem(
				*it,
				sigc::bind<PUIElement, std::string>(
					sigc::mem_fun(*this, &UIDefinitionCanvasEditor::onActionSelect), elem, *it
				)
			));
	}

	menu->popup(0, GDK_CURRENT_TIME);
}

void UIDefinitionCanvasEditor::onActionSelect(PUIElement elem, string_cr action) {
	Polycell * polycell= getEditorWidget()->getTreeView()->getEditingPolycell();
	getEditorWidget()->getTreeView()->rejectEditing();
	dynamic_cast<PolycellInput *>(polycell)->setText(action);
	elem->action= action;
	updateModel(elem->path);
}

void UIDefinitionCanvasEditor::onRowCollapsed(const Gtk::TreeModel::iterator &, const Gtk::TreeModel::Path & path) {
	if(!lock) {
		PUIElement elem= PUIElement::cast_dynamic(getEditorWidget()->getTreeView()->getByTreePath(path));
		if(elem) {
			UiPath spath= elem->path;
			for(Elements::iterator it= elements.find(spath); it != elements.end(); ++it) {
				if(Begins(it->second->path, spath))
					it->second->expanded= false;
				else
					break;
			}
		}
	}
}

void UIDefinitionCanvasEditor::onRowExpanded(const Gtk::TreeModel::iterator &, const Gtk::TreeModel::Path & path) {
	if(!lock) {
		PUIElement elem= PUIElement::cast_dynamic(getEditorWidget()->getTreeView()->getByTreePath(path));
		if(elem) {
			std::vector<int> vpath= elem->path;
			while(!vpath.empty()) {
				getElement(vpath)->expanded= true;
				vpath.pop_back();
			}
		}
	}
}

PUIElement UIDefinitionCanvasEditor::getElement(const UiPath & path) {
	Elements::iterator it= elements.find(path);
	return it!=elements.end() ? it->second : PUIElement();
}

void UIDefinitionCanvasEditor::setElement(const UiPath & path, PUIElement element) {
	elements[path]= element;
}

void UIDefinitionCanvasEditor::dump(string_cr str) {
	std::cout << "--------------------" << str << std::endl;
	for(Elements::iterator it= elements.begin(); it!=elements.end(); ++it) {
		UiPath path= it->first;
		PUIElement elem= it->second;
		std::cout << IdPathToString(IdPath(path.begin(), path.end()))
			<< " " << (elem ? elem->element : "NULL")
			<< " " << (elem ? elem->name : "")
			<< std::endl;
	}
}

}
