/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "forward.h"
#include "canvaswid.h"

#include "session.h"
#include "views.h"
#include "document.h"
#include "palette.h"

namespace Crow {

// Paint -----------------------------------------------------------------------

bool Paint::on_expose_event(GdkEventExpose * event) {
	PGdkWindow window= get_window();
	PGdkGC gc= Gdk::GC::create(window);
	Rectangle R= get_allocation();
	gc->set_rgb_fg_color(frame);
	window->draw_rectangle(gc, true, 0, 0, R.w, R.h);
	R.grow(1, 1, -2, -2);
	if(!R.empty()) {
		gc->set_rgb_fg_color(area);
		window->draw_rectangle(gc, true, 1, 1, R.w, R.h);
	}
	return true;
}

// WidgetCanvasEditorWidget ----------------------------------------------------

WidgetCanvasEditorWidget::WidgetCanvasEditorWidget() :
	Glib::ObjectBase(typeid(WidgetCanvasEditorWidget))
{

	setFrameNeeded(true);

	GAP= 5;
	scrollx= scrolly= 0;

	scroll.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
	add(scroll);
	scroll.show();

	glass.set_above_child(true);
	glass.set_visible_window(false);
	glass.add_events(
		Gdk::POINTER_MOTION_MASK |
		Gdk::POINTER_MOTION_HINT_MASK |
		Gdk::BUTTON_PRESS_MASK |
		Gdk::BUTTON_RELEASE_MASK |
		Gdk::ENTER_NOTIFY_MASK |
		Gdk::LEAVE_NOTIFY_MASK
	);
	glass.set_can_focus(true);
	scroll.add(glass);
	dynamic_cast<Gtk::Viewport *>(scroll.get_child())->set_shadow_type(Gtk::SHADOW_NONE);
	glass.show();

	glass.add(grabber);
	grabber.show();

	grabber.add(fixed);
	fixed.show();

	fixed.put(border, 0, 0);
	border.show();

	container.set_border_width(GAP);
	border.add(container);
	container.show();

	container.add(containment);
	containment.show();

	lastCursor= Gdk::LAST_CURSOR;

	scrollc= scroll.signal_size_allocate().connect(sigc::mem_fun(*this, &WidgetCanvasEditorWidget::setupState));

}

WidgetCanvasEditorWidget::~WidgetCanvasEditorWidget() {
	unsetWidget();
	clearPaintList();
}

void WidgetCanvasEditorWidget::setWidget(Gtk::Widget * widget) {
	containment.setWidget(widget);
	glass.grab_focus();
}

Gtk::Widget * WidgetCanvasEditorWidget::getWidget() {
	return containment.getWidget();
}

void WidgetCanvasEditorWidget::unsetWidget() {
	containment.unsetWidget();
}

void WidgetCanvasEditorWidget::setCursor(Gdk::CursorType cursor) {
	if(glass.get_window() && cursor!=lastCursor) {
		lastCursor= cursor;
		glass.get_window()->set_cursor(Gdk::Cursor(cursor));
	}
}

void WidgetCanvasEditorWidget::unsetCursor() {
	if(glass.get_window()) {
		lastCursor= Gdk::LAST_CURSOR;
		glass.get_window()->set_cursor();
	}
}

void WidgetCanvasEditorWidget::pack(const Point & size) {
	int nx= size.x;
	int ny= size.y;
	int ox, oy;
	containment.getEnclosure()->get_size_request(ox, oy);
	if(ox!=nx || oy!=ny) {
		containment.getEnclosure()->set_size_request(nx, ny);
		if(containment.getWidget())
			containment.getWidget()->queue_resize();
	}
}

void WidgetCanvasEditorWidget::clearPaintList() {
	for(PaintList::iterator it= paintList.begin(); it!=paintList.end(); ++it)
		delete *it;
	paintList.clear();
}

void WidgetCanvasEditorWidget::paint(const Rectangle & R, bool fill, const Gdk::Color & frame, const Gdk::Color & area) {
	if(fill) {
		Paint * paint= new Paint(frame, area);
		paint->set_size_request(R.w, R.h);
		fixed.put(*paint, R.x+GAP, R.y+GAP);
		paint->show();
		paintList.push_back(paint);
	} else {
		paint(Rectangle(R.x, R.y, R.w-1, 1), true, frame, area);
		paint(Rectangle(R.x+R.w-1, R.y, 1, R.h-1), true, frame, area);
		paint(Rectangle(R.x+1, R.y+R.h-1, R.w-1, 1), true, frame, area);
		paint(Rectangle(R.x, R.y+1, 1, R.h-1), true, frame, area);
	}
}

void WidgetCanvasEditorWidget::loadState(PObject state) {
	PState S= PState::cast_dynamic(state);
	if(S) {
		scrollx= S->sx;
		scrolly= S->sy;
	}
}

void WidgetCanvasEditorWidget::setupState(Gtk::Allocation &) {
	scroll.get_hadjustment()->set_value(scrollx);
	scroll.get_vadjustment()->set_value(scrolly);
	scrollc.disconnect();
}

PObject WidgetCanvasEditorWidget::saveState() {
	PState S= MakeRefPtr<State>();
	S->sx= scroll.get_hadjustment()->get_value();
	S->sy= scroll.get_vadjustment()->get_value();
	return S;
}

// WidgetCanvasEditor ----------------------------------------------------------

WidgetCanvasEditor::Location::Location() {
	clear();
}

void WidgetCanvasEditor::Location::clear() {
	widget= PNode();
	where= Nowhere;
}

bool WidgetCanvasEditor::Location::manipulator() {
	return !empty() && !border() && !internal();
}

bool WidgetCanvasEditor::Location::slider() {
	return where==SliderHor || where==SliderVer;
}

bool WidgetCanvasEditor::Location::border() {
	return !empty() && where == Edge;
}

bool WidgetCanvasEditor::Location::internal() {
	return !empty() && where == Nowhere;
}

bool WidgetCanvasEditor::Location::empty() {
	return !widget;
}

Gdk::CursorType WidgetCanvasEditor::Location::cursor() {
	CHECK(!empty());
	switch(where) {
		case North:     return Gdk::TOP_SIDE;
		case NorthEast: return Gdk::TOP_RIGHT_CORNER;
		case East:      return Gdk::RIGHT_SIDE;
		case SouthEast: return Gdk::BOTTOM_RIGHT_CORNER;
		case South:     return Gdk::BOTTOM_SIDE;
		case SouthWest: return Gdk::BOTTOM_LEFT_CORNER;
		case West:      return Gdk::LEFT_SIDE;
		case NorthWest: return Gdk::TOP_LEFT_CORNER;
		case Edge:      return Gdk::FLEUR;
		case SliderHor: return Gdk::SB_UP_ARROW;
		case SliderVer: return Gdk::SB_LEFT_ARROW;
		default:        return Gdk::CIRCLE;
	}
}

WidgetCanvasEditor::WidgetCanvasEditor() {
	GAP= 0;
	ML= 2;
	SG= 6*ML;
	SL= 3*ML/2;
	control= false;
	selectedOnPress= false;
	pasteNormal= true;
	button= -1;
	state= Selecting;
	deleting= false;
}

void WidgetCanvasEditor::created() {
	PWidgetCanvasEditorWidget ew= getEditorWidget();
	GAP= ew->getGAP();
	ew->getGlass()->signal_enter_notify_event().connect(sigc::mem_fun(*this, &WidgetCanvasEditor::onMouseEnter));
	ew->getGlass()->signal_leave_notify_event().connect(sigc::mem_fun(*this, &WidgetCanvasEditor::onMouseLeave));
	ew->getGlass()->signal_motion_notify_event().connect(sigc::mem_fun(*this, &WidgetCanvasEditor::onMouseMove));
	ew->getGlass()->signal_button_press_event().connect(sigc::mem_fun(*this, &WidgetCanvasEditor::onMouseButtonPress));
	ew->getGlass()->signal_button_release_event().connect(sigc::mem_fun(*this, &WidgetCanvasEditor::onMouseButtonRelease));
	getManager()->connect_action_completed(sigc::mem_fun(*this, &WidgetCanvasEditor::onActionCompleted));
}

void WidgetCanvasEditor::load(bool reload) {
	PWidgetCanvasEditorWidget ew= getEditorWidget();
	Model * model= getManager()->getModel();
	PNode master= model->raiseMaster(getSession()->getNode1());
	if(!reload) {
		topMaster= master;
		ew->pack(getPackSize(master));
		ew->setWidget(getWidget(master));
		getWidgetView(master)->revive();
		ew->loadState(getManager()->getStates()->get(topMaster));
		beginSelecting();
	} else {
		CHECK(topMaster == master);
		ew->pack(getPackSize(master));
	}
}

void WidgetCanvasEditor::deleted() {
	getManager()->getStates()->set(topMaster, getEditorWidget()->saveState());
	deleting= true;
	CHECK(state==Selecting || state==Pasting);
	if(state==Pasting)
		read(Cancel);
	CHECK(state==Selecting);
}

PNode WidgetCanvasEditor::getTopMaster() {
	return topMaster;
}

bool WidgetCanvasEditor::startPasting(ustring_cr data, bool normal) {
	CHECK(state==Selecting || state==Pasting);
	pasteData= data;
	pasteNormal= normal;
	if(state==Selecting)
		read(Paste);
	return true;
}

void WidgetCanvasEditor::stopPasting() {
	CHECK(state==Pasting);
	read(Cancel);
}

void WidgetCanvasEditor::onActionCompleted() {
	showSelectionFrames();
}

void WidgetCanvasEditor::read(Command command) {

	switch(state) {

		case Selecting:
			switch(command) {

				case Move:
					selectingMove();
					break;

				case Press:
					testLocation(mouse);
					if(location.manipulator()) {
						state= Adjusting;
					} else if(!location.empty() && button==1) {
						state= Selecting1;
						selectingPress();
					}
					break;

				case Release:
					if(button==3)
						selectingMenu();
					break;

				case Paste:
					state= Pasting;
					beginPasting();
					break;

				default:;

			}
			break;

		case Selecting1:
			switch(command) {

				case Move:
					state= Dragging;
					beginDragging();
					draggingMove();
					break;

				case Release:
					state= Selecting;
					selectingRelease();
					break;

				default:;

			}
			break;

		case Dragging:
			switch(command) {

				case Move:
					draggingMove();
					break;

				case Release:
					endDragging();
					state= Selecting;
					beginSelecting();
					break;

				default:;

			}
			break;

		case Pasting:
			switch(command) {

				case Move:
					pastingMove();
					break;

				case Leave:
					pastingLeave();
					break;

				case Press:
					state= Pasting1;
					pastingPress();
					break;

				case Cancel:
					endPasting();
					state= Selecting;
					pastingFinished();
					beginSelecting();
					break;

				default:;

			}
			break;

		case Pasting1:
			switch(command) {

				case Release:
					if(!control) {
						endPasting();
						state= Selecting;
						pastingFinished();
						beginSelecting();
					} else {
						state= Pasting;
						pastingMove();
					}
					break;

				default:;

			}
			break;

		case Adjusting:
			switch(command) {

				case Move:
					adjustingMove();
					break;

				case Release:
					endAdjusting();
					state= Selecting;
					beginSelecting();
					break;

				default:;

			}
			break;

		default:;

	}

}

void WidgetCanvasEditor::beginSelecting() {
	showSelectionFrames();
	getEditorWidget()->unsetCursor();
}

void WidgetCanvasEditor::selectingMove() {
	testLocation(mouse);
	if(location.empty()) {
		getEditorWidget()->unsetCursor();
		updateStatus("");
	} else {
		if(location.manipulator() || location.border())
			getEditorWidget()->setCursor(location.cursor());
		else
			getEditorWidget()->unsetCursor();
		updateStatus(location.widget, getWidgetRect(location.widget).size());
	}
}

void WidgetCanvasEditor::selectingPress() {
	CHECK(!location.empty());
	selectedOnPress= getManager()->isSelected(location.widget);
	if(!selectedOnPress) {
		if(control) {
			Nodes selection= *getManager()->getSelection();
			selection.push_back(location.widget);
			getManager()->submitSelection(selection);
		} else {
			getManager()->submitSelection(Nodes(1, location.widget));
		}
	}
}

void WidgetCanvasEditor::selectingRelease() {
	if(selectedOnPress) {
		if(control) {
			Nodes selection= *getManager()->getSelection();
			CHECK(FindAndErase(selection, location.widget));
			getManager()->submitSelection(selection);
		} else {
			getManager()->submitSelection(Nodes(1, location.widget));
		}
	}
}

void WidgetCanvasEditor::selectingMenu() {
	testLocation(mouse);
	if(!location.empty()) {
		if(!getManager()->isSelected(location.widget))
			getManager()->submitSelection(Nodes(1, location.widget));
		popupMenu();
	}
}

void WidgetCanvasEditor::beginPasting() {
	getEditorWidget()->setCursor(Gdk::CROSSHAIR);
	frames.clear();
	paintFrames();
}

void WidgetCanvasEditor::pastingMove() {
	FreePlace fplace;
	PNode widget= findWidget(mouse);
	bool found= false;
	bool changed;
	if(widget) {
		Point wp= toLocal(widget, mouse);
		found= getWidgetView(widget)->findFreePlace(wp, &fplace);
	}
	if(found)
		changed= createSingleFrame(Rectangle(
			toGlass(widget, fplace.area.point()),
			fplace.area.size()
		));
	else {
		changed= !frames.empty();
		frames.clear();
	}
	if(changed)
		paintFrames("wce_vc");
}

void WidgetCanvasEditor::pastingLeave() {
	if(!frames.empty()) {
		frames.clear();
		paintFrames();
	}
}

void WidgetCanvasEditor::pastingPress() {
	PNode widget= findWidget(mouse);
	if(widget) {
		Point wp= toLocal(widget, mouse);
		Model * model= getManager()->getModel();

		getManager()->begin(pasteNormal ? umNormal : umPaste);

		Document document;
		Nodes loadedNodes;
		bool loaded= document.load(model, pasteData, loadedNodes);
		bool widgets= true;
		if(loaded) {

			Nodes masters= model->selectMasters(loadedNodes);
			for(Nodes::iterator it=masters.begin(); it!=masters.end(); ++it)
				if(
					!Palette::get()->isWidget((*it)->getType()) ||
					GetEntry<EntityEntry>((*it)->getType())->hasFlag(efTopMaster)
				) {
					widgets= false;
					break;
				}

			if(widgets) {
				FreePlaces fplaces= getWidgetView(widget)->findFreePlaces(wp, masters, wp);
				if(!fplaces.empty()) {

					int i=0;
					for(Nodes::iterator it= masters.begin(); it!=masters.end(); ++it, ++i)
						getWidgetView(widget)->createChild(fplaces[i].place, *it);

					getManager()->commit(masters);
					return;
				}
			}

		}

		getManager()->rollback();

		read(Release);
		button= -1;
		std::string msg;
		if(loaded)
			msg= widgets ? _("Cannot find free place(s)") : _("Non-widget objects could be pasted to root only");
		else
			msg= _("Pasting failed: ") + *document.getError();
		getController()->getContext()->showMessage(msg);

	}

}

void WidgetCanvasEditor::endPasting() {
	pasteNormal= true;
	pasteData.clear();
}

void WidgetCanvasEditor::beginDragging() {
	getEditorWidget()->setCursor(Gdk::PLUS);
	frames.clear();
}

void WidgetCanvasEditor::draggingMove() {
	const Nodes & reduced= *getManager()->getReducedSelection();
	FreePlaces fplaces;
	PNode widget= findWidget(mouse, true);
	Frames F;
	if(widget) {
		Point wp= toLocal(widget, mouse);
		Point wp0= toLocal(widget, press);
		fplaces= getWidgetView(widget)->findFreePlaces(wp, reduced, wp0);
	}
	if(!fplaces.empty()) {
		int i= 0;
		for(Nodes::const_iterator it=reduced.begin(); it!=reduced.end(); ++it, ++i)
			F[*it].R= Rectangle(
				toGlass(widget, fplaces[i].area.point()),
				fplaces[i].area.size()
			);
	} else {
		if(widget)
			F[widget].R= getWidgetRect(widget);
		else
			for(Nodes::const_iterator it=reduced.begin(); it!=reduced.end(); ++it)
				F[*it].R= getWidgetRect(*it);
	}
	if(F != frames) {
		frames= F;
		paintFrames(fplaces.empty() ? "wce_wc" : "wce_vc");
	}
}

void WidgetCanvasEditor::endDragging() {
	frames.clear();
	const Nodes & reduced= *getManager()->getReducedSelection();
	FreePlaces fplaces;
	PNode widget= findWidget(mouse, true);
	if(widget) {
		Point wp= toLocal(widget, mouse);
		Point wp0= toLocal(widget, press);
		fplaces= getWidgetView(widget)->findFreePlaces(wp, reduced, wp0);
	}
	if(!fplaces.empty()) {
		getManager()->begin();
		int i;

		std::vector<Nodes> props;
		props.resize(fplaces.size());
		i= 0;
		for(Nodes::const_iterator it=reduced.begin(); it!=reduced.end(); ++it, ++i)
			if(widget->getType() == getManager()->getModel()->findParent(*it)->getType()) {
				PNode child= getManager()->getModel()->findChildNode(*it);
				Nodes out= *child->getOut();
				for(Nodes::iterator it=out.begin(); it!=out.end(); ++it) {
					Property * prop= getManager()->getController()->findViewProperty(*it);
					if(!prop->hasFlags(sfChildDef)) {
						getManager()->getModel()->cut(*it);
						props[i].push_back(*it);
					}
				}
			}

		i= 0;
		for(Nodes::const_iterator it=reduced.begin(); it!=reduced.end(); ++it, ++i) {
			PNode child= getWidgetView(widget)->createChild(fplaces[i].place, *it);
			for(Nodes::iterator it=props[i].begin(); it!=props[i].end(); ++it)
				getManager()->getModel()->paste(*it, child);
		}

		getManager()->commit();
	}
}

void WidgetCanvasEditor::adjustingMove() {

	Point k= mouse - press;
	Rectangle rect= getWidgetRect(location.widget);
	Frame & frame= frames[location.widget];

	if(location.slider()) {
		if(location.where==SliderHor)
			frame.hk= mapOnSlider(frame.R, frame.H, k.x);
		else
			frame.vk= mapOnSlider(frame.R, frame.V, k.y);
	} else {
		Rectangle dr= getConstrainedShift(location.where);
		rect.grow(k.x*dr.x, k.y*dr.y, k.x*dr.w, k.y*dr.h, 1, 1);
		frame.R= rect;
	}

	paintFrames();

	updateStatus(location.widget, rect.size());
}

void WidgetCanvasEditor::endAdjusting() {
	Frame & frame= frames[location.widget];
	if(location.slider()) {
		PGtkWidgetView view= getWidgetView(location.widget);
		if(location.where==SliderHor)
			view->setSlider(SliderHor, Cap(frame.H.value + frame.hk, 0.0, 1.0));
		else
			view->setSlider(SliderVer, Cap(frame.V.value + frame.vk, 0.0, 1.0));
		getManager()->begin();
		getManager()->getModel()->tagModified(location.widget);
		getManager()->commit();
	} else
		if(frame.R != getWidgetRect(location.widget)) {
			if(location.widget==getSession()->getNode1())
				setDesignSize(frame.R.size());
			else if(isXYChild(location.widget))
				setXYChildPlacement(location.widget, frame.R);
		}
}

void WidgetCanvasEditor::showSelectionFrames() {
	if((state==Selecting || state==Selecting1) && !deleting)
		Glib::signal_idle().connect(sigc::bind_return(sigc::mem_fun(*this, &WidgetCanvasEditor::showSelectionFramesIdle), false));
}

void WidgetCanvasEditor::showSelectionFramesIdle() {
	bool changed= createSelectionFrames();
	if(changed)
		paintFrames();
	if(state==Selecting)
		read(Move);
}

bool WidgetCanvasEditor::createSelectionFrames() {
	Frames F;
	const Nodes & selection= *getManager()->getSelection();
	Model * model= getManager()->getModel();
	bool multiple= selection.size() > 1;
	for(Nodes::const_iterator it= selection.begin(); it!=selection.end(); ++it) {
		PNode node= *it;
		if(isActiveWidget(node) && model->raiseParent(node) == getSession()->getNode1()) {
			Frame & f= F[node];
			f.R= getWidgetRect(node);
			if(!multiple) {
				PGtkWidgetView view= getWidgetView(node);
				f.H= view->getSlider(SliderHor);
				f.V= view->getSlider(SliderVer);
			}
		}
	}
	if(F != frames) {
		frames= F;
		return true;
	}
	return false;
}

bool WidgetCanvasEditor::createSingleFrame(const Rectangle & rect) {
	Frames F;
	F[PNode()].R= rect;
	if(F != frames) {
		frames= F;
		return true;
	}
	return false;
}

void WidgetCanvasEditor::paintFrames(string_cr frameColorName) {

	Gdk::Color frameColor;
	if(frameColorName.empty())
		frameColor= GetBlack(getEditorWidget()->get_style());
	else
		frameColor= getController()->getContext()->getKeyValue<Gdk::Color>(frameColorName);
	Gdk::Color manipColor= GetWhite(getEditorWidget()->get_style());

	Gdk::Color scaleColor= getController()->getContext()->getKeyValue<Gdk::Color>("wce_wc");
	Gdk::Color pointColor= getController()->getContext()->getKeyValue<Gdk::Color>("wce_vc");

	getEditorWidget()->clearPaintList();

	for(Frames::iterator it= frames.begin(); it!=frames.end(); ++it) {
		const Rectangle & R= (*it).second.R;

		Rectangle frame= R;
		frame.grow(-1, -1, 2, 2);
		getEditorWidget()->paint(frame, false, frameColor, Gdk::Color());

		if(state==Selecting || state==Selecting1 || state==Adjusting)
			for(Manipulator M= North; M <= NorthWest; M= nextManip(M)) {
				Rectangle manip= getManipulatorRect(R, M);
				getEditorWidget()->paint(manip, true, frameColor, manipColor);
			}

		if(state==Selecting || state==Selecting1 || state==Adjusting) {
			const Slider & hor= (*it).second.H;
			if(!hor.empty()) {
				paintSliderScale(R, hor, scaleColor, pointColor);
				double v= hor.value + (*it).second.hk;
				Rectangle manip= getSliderRect(R, hor, Cap(v, 0.0, 1.0), SL, SL);
				getEditorWidget()->paint(manip, true, frameColor, manipColor);
			}
			const Slider & ver= (*it).second.V;
			if(!ver.empty()) {
				paintSliderScale(R, ver, scaleColor, pointColor);
				double v= ver.value + (*it).second.vk;
				Rectangle manip= getSliderRect(R, ver, Cap(v, 0.0, 1.0), SL, SL);
				getEditorWidget()->paint(manip, true, frameColor, manipColor);
			}
		}

	}

}

void WidgetCanvasEditor::paintSliderScale(const Rectangle & R, const Slider & S, const Gdk::Color & sc, const Gdk::Color & pc) {
	int L= getSliderScale(R, S);
	int LW, LH;
	if(S.manip == SliderHor) {
		LW= L/2;
		LH= 0;
	} else {
		CHECK(S.manip == SliderVer);
		LW= 0;
		LH= L/2;
	}
	Rectangle rect= getSliderRect(R, S, 0.5, LW, LH);
	getEditorWidget()->paint(rect, true, sc, sc);
	if(S.steps == 0) {
		Rectangle rl= getSliderRect(R, S, 0, ML, ML);
		getEditorWidget()->paint(rl, true, pc, pc);
		Rectangle rr= getSliderRect(R, S, 1, ML, ML);
		getEditorWidget()->paint(rr, true, pc, pc);
	} else
		for(int i=0; i<S.steps; ++i) {
			double V= S.steps > 1 ? double(i)/(S.steps-1) : 0;
			Rectangle rect= getSliderRect(R, S, V, ML, ML);
			getEditorWidget()->paint(rect, true, pc, pc);
		}
}

void WidgetCanvasEditor::testLocation(const Point & p) {
	CHECK(state==Selecting || state==Selecting1);

	location.clear();

	for(Frames::iterator it= frames.begin(); it!=frames.end(); ++it) {
		for(Manipulator M= North; M <= NorthWest; M= nextManip(M))
			if(p.inside(getManipulatorRect((*it).second.R, M))) {
				location.widget= (*it).first;
				location.where= M;
				return;
			}
		const Slider & hor= (*it).second.H;
		if(!hor.empty() && p.inside(getSliderRect((*it).second.R, hor, hor.value, SL, SL))) {
			location.widget= (*it).first;
			location.where= SliderHor;
			return;
		}
		const Slider & ver= (*it).second.V;
		if(!ver.empty() && p.inside(getSliderRect((*it).second.R, ver, ver.value, SL, SL))) {
			location.widget= (*it).first;
			location.where= SliderVer;
			return;
		}
	}

	for(Frames::iterator it= frames.begin(); it!=frames.end(); ++it) {
		Rectangle inner= (*it).second.R;
		Rectangle outer= (*it).second.R;
		inner.grow(2, 2, -4, -4);
		outer.grow(-1, -1, 2, 2);
		if(p.inside(outer) && !p.inside(inner)) {
			location.widget= (*it).first;
			location.where= Edge;
			return;
		}
	}

	location.widget= findWidget(p);

}

PNode WidgetCanvasEditor::findWidget(const Point & p, bool ignoreSelection) {

	if(!isActiveWidget(getSession()->getNode1()))
		return PNode();

	PNode widget;
	PNode child= getSession()->getNode1();

	if(p.inside(getWidgetRect(child))) {

		do {

			if(ignoreSelection && getManager()->isSelected(child))
				break;

			widget= child;
			child= findChild(widget, p);

		} while(child);

	}

	return widget;
}

PNode WidgetCanvasEditor::findChild(PNode widget, const Point & p) {
	Model * model= getManager()->getModel();
	PGtkWidgetView view= getWidgetView(widget);
	if(toLocal(widget, p).inside(view->getChildArea())) {
		Nodes children= model->getChildren(widget, false);
		for(Nodes::const_iterator it= children.begin(); it!=children.end(); ++it) {
			PChild child= getController()->getView<ChildView>(model->findChildNode(*it))->getObject();
			if(view->isTopChild(child) && p.inside(getWidgetRect(*it)))
				return *it;
		}
	}
	return PNode();
}

bool WidgetCanvasEditor::isActiveWidget(PNode node) {

	PNode current= node;

	for(;;) {
		PNode master= getManager()->getModel()->findMaster(current);
		if(!master)
			return true;

		PEntityView view= getController()->getView(master);

		if(!view->isActiveWidget(getWidget(current)))
			return false;

		current= master;
	}

}

bool WidgetCanvasEditor::isInteractive(PNode node) {
	return getWidgetView(node)->isInteractive();
}

Gtk::Widget * WidgetCanvasEditor::getWidget(PNode node) {
	Any value= getController()->getView(node)->getValue();
	Gtk::Widget * widget= dynamic_cast<Gtk::Widget *>(GetPtr(value->getObject()));
	return widget;
}

Rectangle WidgetCanvasEditor::getWidgetRect(PNode node) {
	int width= getWidget(node)->get_allocation().get_width();
	int height= getWidget(node)->get_allocation().get_height();
	Point p= toGlass(node, Point(0, 0));
	Rectangle R= Rectangle(p.x, p.y, width, height);
	if(getSession()->getNode1() == node && topMaster == node) {
		Rectangle S= getEditorWidget()->getContainmentShift();
		R.grow(S.x, S.y, S.w, S.h);
	}
	return R;
}

Rectangle WidgetCanvasEditor::getManipulatorRect(const Rectangle & R, Manipulator manip) {
	Rectangle shift= getManipulatorShift(manip);
	int X= shift.w + 1;
	int Y= shift.h + 1;
	Rectangle rect;
	rect.x= R.x-1-ML + X*(R.w+1)/2;
	rect.y= R.y-1-ML + Y*(R.h+1)/2;
	rect.w= 2*ML+1;
	rect.h= 2*ML+1;
	return rect;
}

Rectangle WidgetCanvasEditor::getManipulatorShift(Manipulator manip) {
	switch(manip) {
		case North:      return Rectangle( 0,  1,  0, -1);
		case NorthEast:  return Rectangle( 0,  1,  1, -1);
		case East:       return Rectangle( 0,  0,  1,  0);
		case SouthEast:  return Rectangle( 0,  0,  1,  1);
		case South:      return Rectangle( 0,  0,  0,  1);
		case SouthWest:  return Rectangle( 1,  0, -1,  1);
		case West:       return Rectangle( 1,  0, -1,  0);
		case NorthWest:  return Rectangle( 1,  1, -1, -1);
		default:         return Rectangle( 0,  0,  0,  0);
	}
}

Rectangle WidgetCanvasEditor::getConstrainedShift(Manipulator manip) {
	Rectangle shift= getManipulatorShift(manip);
	if(getSession()->getNode1() == location.widget && topMaster == location.widget) {
		if(shift.w==-1 || shift.h==-1)
			shift= Rectangle();
	} else if(!isXYChild(location.widget))
		shift= Rectangle();
	return shift;
}

int WidgetCanvasEditor::getSliderScale(const Rectangle & R, const Slider & S) {
	int E= 30;
	int L= (S.manip==SliderHor ? R.w : R.h) + 2;
	if(S.steps > 1) {
		int D= L/(S.steps - 1);
		if(D <= E)
			return L;
		else
			return E*(S.steps - 1);
	} else if(S.steps == 1)
		return 0;
	else
		return L;
}

Rectangle WidgetCanvasEditor::getSliderRect(const Rectangle & R, const Slider & S, double V, int LW, int LH) {
	int sx= 0, sy= 0, W, H;
	double X, Y;
	int L= getSliderScale(R, S);
	if(S.manip == SliderHor) {
		W= L;
		H= R.h;
		X= V;
		Y= 1;
		sy= SG;
	} else {
		CHECK(S.manip == SliderVer);
		W= R.w;
		H= L;
		X= 1;
		Y= V;
		sx= SG;
	}
	Rectangle rect;
	rect.x= R.x-1-LW + (R.w+1)/2.0 + (X-0.5)*W + sx;
	rect.y= R.y-1-LH + (R.h+1)/2.0 + (Y-0.5)*H + sy;
	rect.w= 2*LW+1;
	rect.h= 2*LH+1;
	return rect;
}

double WidgetCanvasEditor::mapOnSlider(const Rectangle & R, const Slider & S, int k) {
	int L= getSliderScale(R, S);
	return L > 0 ? double(k)/L : 0;
}

Point WidgetCanvasEditor::toGlass(PNode node, const Point & p) {
	Point r= Transform(p, getWidget(node), getEditorWidget()->getGlass());
	return r - Point(GAP, GAP);
}

Point WidgetCanvasEditor::toLocal(PNode node, const Point & p) {
	Point r= Transform(p, getEditorWidget()->getGlass(), getWidget(node));
	return r + Point(GAP, GAP);
}

PGtkWidgetView WidgetCanvasEditor::getWidgetView(PNode widget) {
	PEntityView view= getController()->getView(widget);
	return PGtkWidgetView::cast_dynamic(view);
}

bool WidgetCanvasEditor::isXYChild(PNode widget) {
	return getManager()->isXYChild(widget);
}

Point WidgetCanvasEditor::getDesignSize(PNode node) {
	Model * model= getManager()->getModel();
	PNode ds= model->find(node, "design-size");
	return ds ? *ds->getValue()->getPoint() : Point(-1, -1);
}

void WidgetCanvasEditor::setDesignSize(const Point & size) {
	Point CS= getEditorWidget()->getContainmentShift().size();
	Point DS= (size.x<0 || size.y<0) ? Point(-1, -1) : size - CS;
	Model * model= getManager()->getModel();
	getManager()->begin();
	model->setScalar(getSession()->getNode1(), "design-size", CAny::createPoint(DS));
	getManager()->commit();
}

void WidgetCanvasEditor::setXYChildPlacement(PNode widget, const Rectangle & r) {
	Model * model= getManager()->getModel();
	PNode child= model->findChildNode(widget);
	PNode parent= model->findParent(widget);
	Point shift= dynamic_cast<XYContainer *>(PGtkContainerView::cast_dynamic(getWidgetView(parent))->getContainer())->getShift();
	getManager()->begin();
	model->setScalar(widget, "size-request", CAny::createPoint(r.size()));
	model->setScalar(child, "xy", CAny::createPoint(-shift + toLocal(parent, r.point())));
	getManager()->commit();
}

Point WidgetCanvasEditor::getPackSize(PNode widget) {
	Point PS= getDesignSize(widget);
	return (PS.x<0 || PS.y<0) ? Point(-1, -1) : PS;
}

bool WidgetCanvasEditor::onMouseEnter(GdkEventCrossing * event) {
	if(event->window==getEditorWidget()->getGlass()->get_window()->gobj()) {
		read(Enter);
		mouseMove(event->x - GAP, event->y - GAP);
	}
	return true;
}

bool WidgetCanvasEditor::onMouseLeave(GdkEventCrossing * event) {
	if(event->window==getEditorWidget()->getGlass()->get_window()->gobj()) {
		updateStatus("");
		read(Leave);
	}
	return true;
}

bool WidgetCanvasEditor::onMouseMove(GdkEventMotion * event) {
	mouseMove(event->x - GAP, event->y - GAP);
	int x, y;
	Gdk::ModifierType state;
	getEditorWidget()->getGlass()->get_window()->get_pointer(x, y, state);
	return true;
}

bool WidgetCanvasEditor::onMouseButtonPress(GdkEventButton * event) {
	if(event->type==GDK_BUTTON_PRESS)
		buttonPress(event->x - GAP, event->y - GAP, (Gdk::ModifierType)event->state, event->button);
	return true;
}

bool WidgetCanvasEditor::onMouseButtonRelease(GdkEventButton * event) {
	buttonRelease(event->x - GAP, event->y - GAP, (Gdk::ModifierType)event->state, event->button);
	return true;
}

void WidgetCanvasEditor::mouseMove(int x, int y) {
	mouse= Point(x, y);
	read(Move);
}

void WidgetCanvasEditor::buttonPress(int x, int y, Gdk::ModifierType state, int mb) {
	if(button==-1) {
		button= mb;
		press= mouse= Point(x, y);
		control= state & Gdk::CONTROL_MASK;
		read(Press);
	}
}

void WidgetCanvasEditor::buttonRelease(int x, int y, Gdk::ModifierType state, int mb) {
	if(button==mb && button!=-1) {
		mouse= Point(x, y);
		control= state & Gdk::CONTROL_MASK;
		read(Release);
		button= -1;
	}
}

void WidgetCanvasEditor::updateStatus(ustring_cr text) {
	statusChanged(text);
}

void WidgetCanvasEditor::updateStatus(PNode widget, const Point & size) {
	string_cr str= *Palette::get()->getEntry(widget->getType())->getTypeName();
	Point S= size;
	std::string child_str;
	if(getSession()->getNode1() == widget && topMaster == widget) {
		Point child= size - getEditorWidget()->getContainmentShift().size();
		if(child != size) {
			S= child;
			child_str= (boost::format(_(" (frame: %1%x%2%)")) % size.x % size.y).str();
		}
	}
	updateStatus((boost::format(
		_("%1%, %2%, width: %3% height: %4%%5%"))
		% *widget->getName() % str % S.x % S.y % child_str
	).str());
}

}
