/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person 
 *  obtaining a copy of this software and associated documentation 
 *  files (the "Software"), to deal in the Software without 
 *  restriction, including without limitation the rights to use, copy, 
 *  modify, merge, publish, distribute, sublicense, and/or sell copies 
 *  of the Software, and to permit persons to whom the Software is 
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind, 
 *  express or implied, including but not limited to the warranties of 
 *  merchantability, fitness for a particular purpose and 
 *  noninfringement. In no event shall the authors or copyright 
 *  holders be liable for any claim, damages or other liability, 
 *  whether in an action of contract, tort or otherwise, arising from, 
 *  out of or in connection with the software or the use or other 
 *  dealings in the software.
 */

#ifndef CROW_CANVASWID_H
#define CROW_CANVASWID_H

#include "widgets.h"
#include "editors.h"
#include "views.h"

namespace Crow {

class Paint : public Gtk::DrawingArea {

public:
	Paint(const Gdk::Color & f, const Gdk::Color & a) { frame= f; area= a; }

protected:
	bool on_expose_event(GdkEventExpose * event);

private:
	Gdk::Color frame;
	Gdk::Color area;

};

class WidgetCanvasEditorWidget : public EditorWidget {

public:
	WidgetCanvasEditorWidget();
	~WidgetCanvasEditorWidget();

	void setWidget(Gtk::Widget * widget);
	Gtk::Widget * getWidget();
	void unsetWidget();

	void setCursor(Gdk::CursorType cursor);
	void unsetCursor();

	int getGAP() { return GAP; }

	Gtk::EventBox * getGlass() { return &glass; }

	void pack(const Point & size = Point(-1, -1));

	void clearPaintList();
	void paint(const Rectangle & R, bool fill, const Gdk::Color & frame, const Gdk::Color & area);

	void loadState(PObject state);
	void setupState(Gtk::Allocation &);
	PObject saveState();

	Rectangle getContainmentShift() { return containment.getShift(); }

private:
	int GAP;

	Gtk::ScrolledWindow scroll;
	Gtk::EventBox glass;
	Gtk::EventBox grabber;
	Gtk::Fixed fixed;
	Gtk::EventBox border;
	Gtk::Alignment container;
	Containment containment;

	typedef std::list<Paint *> PaintList;
	PaintList paintList;

	Gdk::CursorType lastCursor;

	class State : public Crow::Object {
	public:
		State() { sx= sy= 0; }
		int sx, sy;
	};
	
	typedef Glib::RefPtr<State> PState;

	int scrollx, scrolly;
	sigc::connection scrollc;

};

class WidgetCanvasEditor : public CanvasEditor, public TCreateEditorWidget<WidgetCanvasEditorWidget> {

public:
	WidgetCanvasEditor();

	void created();
	void load(bool reload);
	void deleted();

	PNode getTopMaster();

	bool startPasting(ustring_cr data, bool normal);
	void stopPasting();

protected:
	enum State { Selecting, Selecting1, Dragging, Pasting, Pasting1, Adjusting };
	enum Command { Move, Press, Release, Paste, Cancel, Enter, Leave };

	struct Frame {
		Rectangle R;
		Slider H, V;
		double hk, vk;
		Frame() { hk= vk= 0; }
		bool operator == (const Frame & f) const
			{ return R==f.R && H==f.H && V==f.V && hk==f.hk && vk==f.vk; }
	};

	typedef std::map<PNode, Frame> Frames;

	struct Location {
		Manipulator where;
		PNode widget;
		Location();
		void clear();
		bool manipulator();
		bool slider();
		bool border();
		bool internal();
		bool empty();
		Gdk::CursorType cursor();
	};

	void onActionCompleted();

	void read(Command command);

	void beginSelecting();
	void selectingMove();
	void selectingPress();
	void selectingRelease();
	void selectingMenu();
	void beginPasting();
	void pastingMove();
	void pastingLeave();
	void pastingPress();
	void endPasting();
	void beginDragging();
	void draggingMove();
	void endDragging();
	void adjustingMove();
	void endAdjusting();

	void showSelectionFrames();
	void showSelectionFramesIdle();
	bool createSelectionFrames();
	bool createSingleFrame(const Rectangle &);
	void paintFrames(string_cr frameColorName = std::string());
	void paintSliderScale(const Rectangle & R, const Slider & S, const Gdk::Color & sc, const Gdk::Color & pc);
	void testLocation(const Point & p);
	PNode findWidget(const Point & p, bool ignoreSelection = false);
	PNode findChild(PNode widget, const Point & p);
	bool isActiveWidget(PNode node);
	bool isInteractive(PNode node);

	Gtk::Widget * getWidget(PNode node);
	Rectangle getWidgetRect(PNode node);
	Rectangle getManipulatorRect(const Rectangle & R, Manipulator manip);
	Rectangle getManipulatorShift(Manipulator manip);
	Rectangle getConstrainedShift(Manipulator manip);
	int getSliderScale(const Rectangle & R, const Slider & S);
	Rectangle getSliderRect(const Rectangle & R, const Slider & S, double V, int LW, int LH);
	double mapOnSlider(const Rectangle & R, const Slider & S, int k);
	Point toGlass(PNode widget, const Point & p);
	Point toLocal(PNode widget, const Point & p);
	PGtkWidgetView getWidgetView(PNode widget);
	bool isXYChild(PNode widget);

	Point getDesignSize(PNode node);
	void setDesignSize(const Point & size);
	void setXYChildPlacement(PNode widget, const Rectangle & r);
	Point getPackSize(PNode widget);

	bool onMouseEnter(GdkEventCrossing *);
	bool onMouseLeave(GdkEventCrossing *);
	bool onMouseMove(GdkEventMotion *);
	bool onMouseButtonPress(GdkEventButton *);
	bool onMouseButtonRelease(GdkEventButton *);
	void mouseMove(int x, int y);
	void buttonPress(int x, int y, Gdk::ModifierType state, int mb);
	void buttonRelease(int x, int y, Gdk::ModifierType state, int mb);
	void updateStatus(ustring_cr text);
	void updateStatus(PNode widget, const Point & size);

	Manipulator nextManip(Manipulator M)
		{ return checked_cast<Manipulator>(checked_cast<int>(M)+1); }

private:
	State state;
	bool control;
	int button;
	Point press;
	Point mouse;
	Location location;
	Frames frames;

	int GAP;
	int ML;
	int SL, SG;

	Glib::ustring pasteData;
	bool pasteNormal;
	bool selectedOnPress;

	bool deleting;
	PNode topMaster;

};

}

#endif
