/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "forward.h"
#include "containers.h"

#include "util.h"
#include "widgets.h"

namespace Crow {

// Container -------------------------------------------------------------------

Place::Place() {
	a= b= c= d= 0;
}

bool Place::operator == (const Place & p) const {
	return a==p.a && b==p.b && c==p.c && d==p.d;
}

Placeholder::Placeholder() {
	set_size_request(40, 40);
}

bool Placeholder::on_expose_event(GdkEventExpose * event) {
	bool result= Gtk::DrawingArea::on_expose_event(event);
	DrawGrid(get_window(), get_style());
	return result;
}

Container::Container() {
	container= NULL;
}

bool Container::isPlaceholder(PGtkWidget widget) {
	return Glib::RefPtr<Placeholder>::cast_dynamic(widget);
}

PChild Container::createChild() {
	return MakeRefPtr<Child>();
}

void Container::erase(Gtk::Widget * widget) {
	if(widget)
		ContainerRemove(getContainer(), widget);
}

void Container::unparent(PGtkWidget widget, bool force) {
	if(widget && widget->get_parent() && (widget->get_parent()!=getContainer() || force))
		ContainerRemove(widget->get_parent(), GetPtr(widget));
}

PGtkWidget Container::createPlaceholder() {
	PGtkWidget widget= MakeRefPtr<Placeholder>();
	widget->show();
	return widget;
}

Gtk::PackOptions Container::getPackOptions(bool expand, bool fill) {
	if(!expand)
		return Gtk::PACK_SHRINK;
	else
		return fill ? Gtk::PACK_EXPAND_WIDGET : Gtk::PACK_EXPAND_PADDING;
}

bool Container::getExpand(Gtk::PackOptions packOptions) {
	return packOptions != Gtk::PACK_SHRINK;
}

bool Container::getFill(Gtk::PackOptions packOptions) {
	return packOptions == Gtk::PACK_EXPAND_WIDGET;
}

// XYContainer -----------------------------------------------------------------

XYChild::XYChild() {
	SetData(this, "CrowTypeHint", "CrowXYChild");
}

void XYContainer::set(const Children & children) {

	if(children.empty()) {
		ContainerClear(getContainer());
		return;
	}

	Children old= get(false);

	for(int i=0; i<SIGNED(old.size()); ++i) {
		PChild child= find(old[i]->widget, children);
		if(!child)
			erase(GetPtr(old[i]->widget));
	}

	for(int i=0; i<SIGNED(children.size()); ++i) {

		PChild child= find(children[i]->widget, old);

		if(child) {

			if(child->place != children[i]->place)
				move(children[i]);

		}	else {

			insert(children[i]);

		}

	}

}

Children XYContainer::get(bool placeholders) {
	Children result;
	std::vector<Gtk::Widget *> widgets= getContainer()->get_children();
	result.resize(widgets.size());
	Point shift= getShift();
	for(int i=0; i<SIGNED(result.size()); ++i) {
		result[i]= Container::createChild();
		result[i]->widget= MakeRefPtr(widgets[i]);
		Point xy= getXY(widgets[i]);
		result[i]->place.a= xy.x - shift.x;
		result[i]->place.b= xy.y - shift.y;
	}
	return result;
}

Point XYContainer::getShift() {
	return Point(0, 0);
}

PChild XYContainer::find(PGtkWidget widget, const Children & children) {
	for(int i=0; i<SIGNED(children.size()); ++i)
		if(children[i]->widget == widget)
			return children[i];
	return PChild();
}

Point XYContainer::getXY(Gtk::Widget * widget) {
	Point xy;
	xy.x= getChildProperty<int>(widget, "x");
	xy.y= getChildProperty<int>(widget, "y");
	return xy + getShift();
}

// FixedContainer --------------------------------------------------------------

Point FixedContainer::getShift() {
	int bw= getContainer()->get_border_width();
	return Point(bw, bw);
}

void FixedContainer::insert(PChild child) {
	unparent(child->widget);
	getContainer()->put(*GetPtr(child->widget), child->place.a, child->place.b);
}

void FixedContainer::move(PChild child) {
	getContainer()->move(*GetPtr(child->widget), child->place.a, child->place.b);
}

// LayoutContainer -------------------------------------------------------------

Point LayoutContainer::getShift() {
	int sx= getContainer()->get_hadjustment()->get_value();
	int sy= getContainer()->get_vadjustment()->get_value();
	return Point(-sx, -sy);
}

void LayoutContainer::insert(PChild child) {
	unparent(child->widget);
	getContainer()->put(*GetPtr(child->widget), child->place.a, child->place.b);
}

void LayoutContainer::move(PChild child) {
	getContainer()->move(*GetPtr(child->widget), child->place.a, child->place.b);
}

// RowContainer ----------------------------------------------------------------

RowChild::RowChild() {
	SetData(this, "CrowTypeHint", "CrowRowChild");
}

void RowContainer::update(int item, PChild child, bool placeholders) {
	Gtk::Widget * widget= getWidget(item);

	if(!child) {

		if(!placeholders) {

			setWidget(item, NULL);

		} else {

			if(widget) {
				if(!isPlaceholder(MakeRefPtr(widget))) {
					setWidget(item, GetPtr(createPlaceholder()));
					updateChild(item, child);
				}
			} else {
				setWidget(item, GetPtr(createPlaceholder()));
				updateChild(item, child);
			}

		}

	} else {

		if(widget != GetPtr(child->widget)) {
			unparent(child->widget, true);
			setWidget(item, GetPtr(child->widget));
			updateChild(item, child);
		} else
			updateChild(item, child);

	}

}

// BinContaner -----------------------------------------------------------------

void BinContainer::set(const Children & children, bool placeholders) {
	CHECK(children.size()<=1);
	if(children.empty()) {
		update(0, PChild(), placeholders);
	} else
		update(0, children.front(), placeholders);
}

Children BinContainer::get(bool placeholders) {
	Children result;
	Gtk::Widget * child= getWidget(0);
	if(child && (placeholders || !isPlaceholder(MakeRefPtr(child)))) {
		result.push_back(Container::createChild());
		result.front()->widget= MakeRefPtr(child);
	}
	return result;
}

Gtk::Widget * BinContainer::getWidget(int item) {
	CHECK(item==0);
	return getContainer()->get_child();
}

void BinContainer::setWidget(int item, Gtk::Widget * widget) {
	CHECK(item==0);
	BinRemove(getContainer());
	if(widget)
		getContainer()->add(*widget);
}

// ScrolledWindowContainer -----------------------------------------------------

Gtk::Widget * ScrolledWindowContainer::getWidget(int item) {
	CHECK(item==0);
	Gtk::Widget * widget= getContainer()->get_child();
	if(widget && !GetData(widget, "CrowAutoViewportHint").empty()) {
		Gtk::Viewport * viewport= dynamic_cast<Gtk::Viewport *>(widget);
		return viewport->get_child();
	} else
		return widget;
}

void ScrolledWindowContainer::setWidget(int item, Gtk::Widget * widget) {
	CHECK(item==0);
	Gtk::Widget * child= getContainer()->get_child();
	if(child) {
		child->reference();
		BinRemove(getContainer());
		if(!GetData(child, "CrowAutoViewportHint").empty())
			BinRemove(dynamic_cast<Gtk::Bin *>(child));
		g_signal_handlers_block_matched(getContainer()->get_vadjustment()->gobj(), G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, child->gobj());
		g_signal_handlers_block_matched(getContainer()->get_hadjustment()->gobj(), G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, child->gobj());
		child->unreference();
	}
	if(widget) {
		getContainer()->add(*widget);
		if(getContainer()->get_child() != widget && !dynamic_cast<Custom *>(widget)) {
			Gtk::Viewport * viewport= dynamic_cast<Gtk::Viewport *>(getContainer()->get_child());
			SetData(viewport, "CrowAutoViewportHint", "true");
		}
	}
}

// DialogContainer -------------------------------------------------------------

Gtk::Dialog * DialogContainer::getContainer() {
	return dynamic_cast<Gtk::Dialog *>(Container::getContainer());
}

Gtk::Bin * DialogContainer::getDialogBin() {
	Gtk::Widget * bin= getContainer()->get_vbox()->children().back().get_widget();
	Gtk::Alignment * alignment= dynamic_cast<Gtk::Alignment *>(bin);
	CHECK(alignment);
	return alignment;
}

Gtk::Widget * DialogContainer::getWidget(int item) {
	CHECK(item==0);
	return getDialogBin()->get_child();
}

void DialogContainer::setWidget(int item, Gtk::Widget * widget) {
	CHECK(item==0);
	BinRemove(getDialogBin());
	if(widget)
		getDialogBin()->add(*widget);
}

// InfoBarContainer ------------------------------------------------------------

Gtk::InfoBar * InfoBarContainer::getContainer() {
	return dynamic_cast<Gtk::InfoBar *>(Container::getContainer());
}

Gtk::Bin * InfoBarContainer::getInfoBarBin() {
	Gtk::Container * carea= dynamic_cast<Gtk::Container *>(getContainer()->get_content_area());
	CHECK(carea);
	std::list<Gtk::Widget *> list= carea->get_children();
	CHECK(!list.empty());
	Gtk::Widget * bin= list.back();
	Gtk::Alignment * alignment= dynamic_cast<Gtk::Alignment *>(bin);
	CHECK(alignment);
	return alignment;
}

Gtk::Widget * InfoBarContainer::getWidget(int item) {
	CHECK(item==0);
	return getInfoBarBin()->get_child();
}

void InfoBarContainer::setWidget(int item, Gtk::Widget * widget) {
	CHECK(item==0);
	BinRemove(getInfoBarBin());
	if(widget)
		getInfoBarBin()->add(*widget);
}

// PanedContainer --------------------------------------------------------------

PanedChild::PanedChild() {
	SetData(this, "CrowTypeHint", "CrowPanedChild");
	resize= true;
	shrink= true;
}

void PanedContainer::set(const Children & children, bool placeholders) {
	CHECK(children.size()<=2);
	if(children.empty()) {
		update(0, PChild(), placeholders);
		update(1, PChild(), placeholders);
	} else if(children.size()==2) {
		update(children[0]->place.a, children[0], placeholders);
		update(children[1]->place.a, children[1], placeholders);
	} else {
		PChild child= children.front();
		if(child->place.a==0) {
			update(0, child, placeholders);
			update(1, PChild(), placeholders);
		} else {
			update(0, PChild(), placeholders);
			update(1, child, placeholders);
		}
	}
}

Children PanedContainer::get(bool placeholders) {
	Children result;
	for(int i=0; i<2; ++i) {
		Gtk::Widget * child= getWidget(i);
		if(child && (placeholders || !isPlaceholder(MakeRefPtr(child)))) {
			PChild R= Container::createChild();
			R->widget= MakeRefPtr(child);
			R->place.a= i;
			result.push_back(R);
		}
	}
	return result;
}

PChild PanedContainer::createChild() {
	return MakeRefPtr<PanedChild>();
}

Gtk::Widget * PanedContainer::getWidget(int item) {
	CHECK(item<=1);
	return item==0 ? getContainer()->get_child1() : getContainer()->get_child2();
}

void PanedContainer::setWidget(int item, Gtk::Widget * widget) {
	CHECK(item<=1);
	if(item==0) {
		erase(getContainer()->get_child1());
		if(widget)
			getContainer()->add1(*widget);
	} else {
		erase(getContainer()->get_child2());
		if(widget)
			getContainer()->add2(*widget);
	}
}

void PanedContainer::updateChild(int item, PChild child) {
	Gtk::Widget * widget= getWidget(item);
	PPanedChild pchild= PPanedChild::cast_dynamic(child);
	bool resize= pchild ? pchild->resize : true;
	bool shrink= pchild ? pchild->shrink : true;
	if(getChildProperty<bool>(widget, "resize") != resize)
		setChildProperty<bool>(widget, "resize", resize);
	if(getChildProperty<bool>(widget, "shrink") != shrink)
		setChildProperty<bool>(widget, "shrink", shrink);
}

// IndexContainer --------------------------------------------------------------

void IndexContainer::process(int capacity) {

	for(int i=0; i<capacity; ++i) {

		PGtkWidget widget= input(i)->widget;

		int index= find(widget, i);

		if(index == -1) {

			insert(i);

		} else {

			if(index != i)
				reorder(index, i);

		}

		update(i);

	}

	shrink(capacity);

}

// BoxContainer ----------------------------------------------------------------

BoxChild::BoxChild() {
	SetData(this, "CrowTypeHint", "CrowBoxChild");
	place.b= Gtk::PACK_START;
	options= Gtk::PACK_EXPAND_WIDGET;
	padding= 0;
}

void BoxContainer::set(const Children & children, int capS, int capE) {

	sorted.resize(capS + capE);

	for(int i=0; i<SIGNED(children.size()); ++i) {

		PChild child= children[i];

		if(child->place.b == Gtk::PACK_START) {
			CHECK(child->place.a < capS);
			CHECK(!sorted[child->place.a]);
			sorted[child->place.a]= child;
		} else {
			CHECK(child->place.b == Gtk::PACK_END);
			CHECK(child->place.a < capE);
			CHECK(!sorted[capS + child->place.a]);
			sorted[capS + child->place.a]= child;
		}

	}

	for(int i=0; i<capS+capE; ++i)
		if(!sorted[i]) {
			sorted[i]= createChild();
			if(i<capS) {
				sorted[i]->place.a= i;
				sorted[i]->place.b= Gtk::PACK_START;
			} else {
				sorted[i]->place.a= i-capS;
				sorted[i]->place.b= Gtk::PACK_END;
			}
		}

	process(capS + capE);

	sorted.clear();

}

Children BoxContainer::get(bool placeholders) {

	Children result;
	Gtk::Box_Helpers::BoxList::iterator child= getContainer()->children().begin();
	int i=0;
	result.reserve(getContainer()->children().size());
	int capS= 0;
	int capE= 0;

	for(; child!=getContainer()->children().end(); ++child, ++i) {

		if(!child->get_pack()) {
			CHECK(capE==0);
			++capS;
		} else
			++capE;

		if(!placeholders && isPlaceholder(MakeRefPtr(child->get_widget())))
			continue;

		PChild R= Container::createChild();
		result.push_back(R);

		R->widget= MakeRefPtr(child->get_widget());
		R->place.a= child->get_pack() ? i - capS : i;
		R->place.b= child->get_pack();

	}

	CHECK(capS + capE == SIGNED(getContainer()->children().size()));

	return result;
}

PChild BoxContainer::createChild() {
	return MakeRefPtr<BoxChild>();
}

int BoxContainer::find(PGtkWidget widget, int start) {

	Gtk::Box_Helpers::BoxList::iterator it= getContainer()->children().begin();
	int i= 0;

	for(; it!=getContainer()->children().end(); ++it, ++i)
		if(i >= start) {

			if(widget) {

				if(it->get_widget() == GetPtr(widget))
					return i;

			} else {

				if(isPlaceholder(MakeRefPtr(it->get_widget())))
					return i;

			}

		}

	return -1;
}

void BoxContainer::reorder(int index1, int index2) {
	Gtk::Widget * widget= getContainer()->children()[index1].get_widget();
	getContainer()->reorder_child(*widget, index2);
}

void BoxContainer::shrink(int size) {
	while(SIGNED(getContainer()->children().size()) > size)
		getContainer()->children().pop_back();
}

PChild BoxContainer::input(int i) {
	CHECK(0 <= i && i < SIGNED(sorted.size()));
	return sorted[i];
}

void BoxContainer::insert(int index) {
	Gtk::Box_Helpers::BoxList::iterator it= getContainer()->children().begin();
	for(int i=0; i<index; ++i)
		++it;
	PBoxChild S= PBoxChild::cast_dynamic(sorted[index]);
	PGtkWidget widget= S->widget ? S->widget : createPlaceholder();
	unparent(widget);
	getContainer()->children().insert(it, Gtk::Box_Helpers::Element(
		*GetPtr(widget),
		S->options,
		S->padding,
		(Gtk::PackType)S->place.b
	));
}

void BoxContainer::update(int index) {
	Gtk::Box_Helpers::BoxList::iterator child= getContainer()->children().begin();
	for(int i=0; i<index; ++i)
		++child;
	PBoxChild S= PBoxChild::cast_dynamic(sorted[index]);
	if(S->widget)
		CHECK(GetPtr(S->widget) == child->get_widget());
	else
		CHECK(isPlaceholder(MakeRefPtr(child->get_widget())));
	if((Gtk::PackType)child->get_pack() != (Gtk::PackType)S->place.b)
		setChildProperty(child->get_widget(), "pack-type", S->place.b);
	if(getPackOptions(child->get_expand(), child->get_fill()) != S->options) {
		setChildProperty(child->get_widget(), "expand", getExpand(S->options));
		setChildProperty(child->get_widget(), "fill", getFill(S->options));
	}
	if(child->get_padding() != S->padding)
		setChildProperty(child->get_widget(), "padding", S->padding);
	postUpdate(child, index);
}

// ButtonBoxContainer ----------------------------------------------------------

ButtonBoxChild::ButtonBoxChild() {
	SetData(this, "CrowTypeHint", "CrowButtonBoxChild");
	secondary= false;
}

PChild ButtonBoxContainer::createChild() {
	return MakeRefPtr<ButtonBoxChild>();
}

void ButtonBoxContainer::postUpdate(Gtk::Box_Helpers::BoxList::iterator child, int index) {
	PButtonBoxChild S= PButtonBoxChild::cast_dynamic(sorted[index]);
	if(getContainer()->get_child_secondary(*child->get_widget()) != S->secondary)
		getContainer()->set_child_secondary(*child->get_widget(), S->secondary);
}

// PagesContainer --------------------------------------------------------------

void PagesContainer::set(const Children & children, int capacity) {
	int page= getCurrentPage();
	sorted.resize(capacity);
	for(int i=0; i<SIGNED(children.size()); ++i) {
		PChild child= children[i];
		CHECK(child->place.a < capacity);
		CHECK(!sorted[child->place.a]);
		sorted[child->place.a]= child;
	}
	for(int i=0; i<capacity; ++i)
		if(!sorted[i]) {
			sorted[i]= createChild();
			sorted[i]->place.a= i;
		}
	process(capacity);
	sorted.clear();
	if(0 <= page && page < getPageCount())
		setCurrentPage(page);
	revive();
}

Children PagesContainer::get(bool placeholders) {

	int count= getPageCount();
	Children result;
	result.reserve(count);

	for(int i= 0; i < count; ++i) {
		Gtk::Widget * child= getPageWidget(i);
		if(!placeholders && isPlaceholder(MakeRefPtr(child)))
			continue;
		PChild R= Container::createChild();
		result.push_back(R);
		R->widget= MakeRefPtr(child);
		R->place.a= i;
	}

	return result;
}

int PagesContainer::find(PGtkWidget widget, int start) {

	int count= getPageCount();

	for(int i= start; i < count; ++i) {

		Gtk::Widget * child= getPageWidget(i);

		if(widget) {

			if(child == GetPtr(widget))
				return i;

		} else {

			if(isPlaceholder(MakeRefPtr(child)))
				return i;

		}

	}

	return -1;
}

void PagesContainer::shrink(int size) {
	while(getPageCount() > size) {
		Gtk::Widget * child= getPageWidget(-1);
		ContainerRemove(getContainer(), child);
	}
}

PChild PagesContainer::input(int i) {
	CHECK(0 <= i && i < SIGNED(sorted.size()));
	return sorted[i];
}

// NotebookContainer -----------------------------------------------------------

NotebookChild::NotebookChild() {
	SetData(this, "CrowTypeHint", "CrowNotebookChild");
	packOptions= Gtk::PACK_SHRINK;
}

PChild NotebookContainer::createChild() {
	return MakeRefPtr<NotebookChild>();
}

void NotebookContainer::reorder(int index1, int index2) {
	Gtk::Widget * widget= getContainer()->get_nth_page(index1);
	getContainer()->reorder_child(*widget, index2);
}

void NotebookContainer::insert(int index) {
	PChild S= input(index);
	PGtkWidget widget= S->widget ? S->widget : createPlaceholder();
	unparent(widget);
	getContainer()->insert_page(*GetPtr(widget), index);
}

void NotebookContainer::update(int index) {
	PNotebookChild S= PNotebookChild::cast_dynamic(input(index));
	Gtk::Notebook_Helpers::PageList::iterator child= getContainer()->pages().find(index);

	if(S->widget)
		CHECK(GetPtr(S->widget) == child->get_child());
	else
		CHECK(isPlaceholder(MakeRefPtr(child->get_child())));

	if(S->tabLabelWidget) {
		if(child->get_tab_label() != GetPtr(S->tabLabelWidget))
			getContainer()->set_tab_label(*child->get_child(), *GetPtr(S->tabLabelWidget));
	} else {
		if(S->tabLabel.empty())
			getContainer()->set_tab_label_text(*child->get_child(), "Page " + ToString(index));
		else
			getContainer()->set_tab_label_text(*child->get_child(), S->tabLabel);
	}

	if(S->menuLabelWidget) {
		if(child->get_menu_label() != GetPtr(S->menuLabelWidget))
			getContainer()->set_menu_label(*child->get_child(), *GetPtr(S->menuLabelWidget));
	} else {
		if(S->menuLabel.empty())
			gtk_container_child_set(
				Container::getContainer()->gobj(), child->get_child()->gobj(),
				"menu-label", NULL, NULL
			);
		else
			getContainer()->set_menu_label_text(*child->get_child(), S->menuLabel);
	}

	bool expand= getChildProperty<bool>(child->get_child(), "tab-expand");
	bool fill= getChildProperty<bool>(child->get_child(), "tab-fill");
	if(getPackOptions(expand, fill) != S->packOptions) {
		setChildProperty(child->get_child(), "tab-expand", getExpand(S->packOptions));
		setChildProperty(child->get_child(), "tab-fill", getFill(S->packOptions));
	}

}

// AssistantContainer ----------------------------------------------------------

AssistantChild::AssistantChild() {
	SetData(this, "CrowTypeHint", "CrowAssistantChild");
	pageType= Gtk::ASSISTANT_PAGE_CONTENT;
	complete= false;
}

PChild AssistantContainer::createChild() {
	return MakeRefPtr<AssistantChild>();
}

void AssistantContainer::reorder(int index1, int index2) {
	PGtkWidget widget= MakeRefPtr(getContainer()->get_nth_page(index1));
	ContainerRemove(getContainer(), GetPtr(widget));
	getContainer()->insert_page(*GetPtr(widget), index2);
}

void AssistantContainer::insert(int index) {
	PChild S= input(index);
	PGtkWidget widget= S->widget ? S->widget : createPlaceholder();
	unparent(widget);
	getContainer()->insert_page(*GetPtr(widget), index);
}

void AssistantContainer::update(int index) {
	PAssistantChild S= PAssistantChild::cast_dynamic(input(index));
	Gtk::Widget * child= getContainer()->get_nth_page(index);

	if(S->widget)
		CHECK(GetPtr(S->widget) == child);
	else
		CHECK(isPlaceholder(MakeRefPtr(child)));

	if(getContainer()->get_page_type(*child) != S->pageType)
		getContainer()->set_page_type(*child, S->pageType);

	if(getContainer()->get_page_complete(*child) != S->complete)
		getContainer()->set_page_complete(*child, S->complete);

	if(S->title.empty())
		getContainer()->set_page_title(*child, "Page " + ToString(index));
	else
		getContainer()->set_page_title(*child, S->title);

	updateImage(child, "header-image", S->headerImage);
	updateImage(child, "sidebar-image", S->sidebarImage);
}

void AssistantContainer::updateImage(Gtk::Widget * child, string_cr prop, string_cr newimg) {
	std::string oldimg;
	PGdkPixbuf pixbuf= getChildProperty<PGdkPixbuf>(child, prop);
	if(pixbuf)
		oldimg= GetData(GetPtr(pixbuf), "CrowImageName");
	if(oldimg != newimg) {
		pixbuf= LoadImage(newimg, "gtk-missing-image");
		if(pixbuf)
			SetData(GetPtr(pixbuf), "CrowImageName", newimg);
		setChildProperty(child, prop, pixbuf);
	}
}

// TableContainer --------------------------------------------------------------

TableChild::TableChild() {
	SetData(this, "CrowTypeHint", "CrowTableChild");
	place.c= 1;
	place.d= 1;
	paddingX= 0;
	paddingY= 0;
	xExpand= true;
	xFill= true;
	xShrink= false;
	yExpand= true;
	yFill= true;
	yShrink= false;
}

TableContainer::TableContainer() {
	skip= createChild();
	capX= capY= 0;
}

void TableContainer::set(const Children & children, int capX, int capY, bool placeholders) {
	sorted.resize(capX*capY);
	this->capX= capX;
	this->capY= capY;
	for(int i=0; i<SIGNED(children.size()); ++i) {
		PChild child= children[i];
		put(child->place.a, child->place.b, child);
	}
	for(int i=0; i<SIGNED(children.size()); ++i) {
		PChild child= children[i];
		int x= child->place.a;
		int y= child->place.b;
		int sx=1, sy=1;
		while(sx<child->place.c && mark(x+sx, y, 1))
			++sx;
		while(sy<child->place.d && mark(x, y+sy, sx))
			++sy;
		child->place.c= sx;
		child->place.d= sy;
	}
	for(int x=0; x<capX; ++x)
		for(int y=0; y<capY; ++y)
			if(!input(x, y)) {
				PChild child= createChild();
				child->place.a= x;
				child->place.b= y;
				put(x, y, child);
			}
	process(placeholders);
	sorted.clear();
	this->capX= 0;
	this->capY= 0;
}

Children TableContainer::get(bool placeholders) {
	Children result;
	result.reserve(getContainer()->children().size());
	Gtk::Table_Helpers::TableList::iterator child= getContainer()->children().begin();
	for(; child!=getContainer()->children().end(); ++child)
		if(placeholders || !isPlaceholder(MakeRefPtr(child->get_widget()))) {
			PChild R= Container::createChild();
			R->widget= MakeRefPtr(child->get_widget());
			R->place.a= child->get_left_attach();
			R->place.b= child->get_top_attach();
			R->place.c= child->get_right_attach() - child->get_left_attach();
			R->place.d= child->get_bottom_attach() - child->get_top_attach();
			result.push_back(R);
		}
	std::sort(result.begin(), result.end(), less());
	return result;
}

bool TableContainer::less::operator () (const PChild & ch1, const PChild & ch2) const {
	const Place & p1= ch1->place;
	const Place & p2= ch2->place;
	return p1.b < p2.b || (p1.b==p2.b && p1.a < p2.a);
}

PChild TableContainer::createChild() {
	return MakeRefPtr<TableChild>();
}

bool TableContainer::input(int x, int y) {
	CHECK(x < capX && y < capY);
	return sorted[x+capX*y];
}

void TableContainer::put(int x, int y, PChild child) {
	CHECK(x < capX && y < capY);
	CHECK(!input(x, y));
	sorted[x+capX*y]= child;
}

bool TableContainer::mark(int x, int y, int s) {
	if(y>=capY || x+s-1>=capX)
		return false;
	for(int i=0; i<s; ++i)
		if(input(x+i, y))
			return false;
	for(int i=0; i<s; ++i)
		put(x+i, y, skip);
	return true;
}

void TableContainer::process(bool placeholders) {

	getContainer()->resize(capX, capY);

	Children old= get(true);

	for(int i=0; i<SIGNED(sorted.size()); ++i)
		if(sorted[i]!=skip && (placeholders || sorted[i]->widget)) {

			PTableChild tchild= PTableChild::cast_dynamic(sorted[i]);
			PChild child;
			for(int n=0; n<SIGNED(old.size()); ++n)
				if(
					(tchild->widget && old[n]->widget == tchild->widget) ||
					(!tchild->widget && isPlaceholder(old[n]->widget) && old[n]->place==tchild->place)
				)
					child= old[n];

			if(child) {
				tchild->widget= child->widget;
				child->widget= PGtkWidget();
				if(child->place != tchild->place)
					move(tchild);
				if(!isPlaceholder(tchild->widget))
					update(tchild);
			}	else {
				insert(tchild);
			}
		}

	for(int i=0; i<SIGNED(old.size()); ++i)
		if(old[i]->widget)
			erase(GetPtr(old[i]->widget));

}

void TableContainer::move(PTableChild child) {
	setChildProperty(GetPtr(child->widget), "left-attach", child->place.a);
	setChildProperty(GetPtr(child->widget), "top-attach", child->place.b);
	setChildProperty(GetPtr(child->widget), "right-attach", child->place.a + child->place.c);
	setChildProperty(GetPtr(child->widget), "bottom-attach", child->place.b + child->place.d);
}

void TableContainer::insert(PTableChild child) {
	if(!child->widget)
		child->widget= createPlaceholder();
	unparent(child->widget);
	getContainer()->attach(*GetPtr(child->widget),
		child->place.a, child->place.a + child->place.c,
		child->place.b, child->place.b + child->place.d,
		getAttachOptions(child->xExpand, child->xFill, child->xShrink),
		getAttachOptions(child->yExpand, child->yFill, child->yShrink),
		child->paddingX, child->paddingY
	);
}

void TableContainer::update(PTableChild child) {
	int xpad= getChildProperty<unsigned>(GetPtr(child->widget), "x-padding");
	if(xpad != child->paddingX)
		setChildProperty(GetPtr(child->widget), "x-padding", child->paddingX);
	int ypad= getChildProperty<unsigned>(GetPtr(child->widget), "y-padding");
	if(ypad != child->paddingY)
		setChildProperty(GetPtr(child->widget), "y-padding", child->paddingY);
	Gtk::AttachOptions xOptions= getAttachOptions(child->xExpand, child->xFill, child->xShrink);
	if(getChildProperty<Gtk::AttachOptions>(GetPtr(child->widget), "x-options") != xOptions)
		setChildProperty(GetPtr(child->widget), "x-options", xOptions);
	Gtk::AttachOptions yOptions= getAttachOptions(child->yExpand, child->yFill, child->yShrink);
	if(getChildProperty<Gtk::AttachOptions>(GetPtr(child->widget), "y-options") != yOptions)
		setChildProperty(GetPtr(child->widget), "y-options", yOptions);
}

Gtk::AttachOptions TableContainer::getAttachOptions(bool expand, bool fill, bool shrink) {
	Gtk::AttachOptions result= Gtk::AttachOptions();
	if(expand)
		result |= Gtk::EXPAND;
	if(fill)
		result |= Gtk::FILL;
	if(shrink)
		result |= Gtk::SHRINK;
	return result;
}

}
