/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#ifndef CROW_CONTAINERS_H
#define CROW_CONTAINERS_H

#include "util.h"

namespace Crow {

struct Place {
	int a, b, c, d;
	Place();
	bool operator == (const Place & p) const;
};

struct Child : public Glib::Object {
	PGtkWidget widget;
	Place place;
};

typedef Glib::RefPtr<Child> PChild;
typedef std::vector<PChild> Children;

class Placeholder : public Gtk::DrawingArea {

public:
	Placeholder();

protected:
	bool on_expose_event(GdkEventExpose * event);

};

class Container {

public:
	Container();
	virtual ~Container() { }

	virtual Children get(bool placeholders) = 0;

	Gtk::Container * getContainer() { return container; }
	void setContainer(Gtk::Container * cont) { container= cont; }

	bool isPlaceholder(PGtkWidget widget);

protected:
	virtual PChild createChild();

	template <class T> T getChildProperty(Gtk::Widget * widget, string_cr name);
	template <class T> void setChildProperty(Gtk::Widget * widget, string_cr name, const T & t);

	void erase(Gtk::Widget * widget);
	void unparent(PGtkWidget widget, bool force = false);

	PGtkWidget createPlaceholder();

	Gtk::PackOptions getPackOptions(bool expand, bool fill);
	bool getExpand(Gtk::PackOptions packOptions);
	bool getFill(Gtk::PackOptions packOptions);

private:
	Gtk::Container * container;

};

template <class T> T Container::getChildProperty(Gtk::Widget * widget, string_cr name) {
	Glib::Value<T> value;
	value.init(Glib::Value<T>::value_type());
	gtk_container_child_get_property(
		getContainer()->gobj(), widget->gobj(),
		name.c_str(), value.gobj()
	);
	return value.get();
}

template <class T> void Container::setChildProperty(Gtk::Widget * widget, string_cr name, const T & t) {
	Glib::Value<T> value;
	value.init(Glib::Value<T>::value_type());
	value.set(t);
	gtk_container_child_set_property(
		getContainer()->gobj(), widget->gobj(),
		name.c_str(), value.gobj()
	);
}

// -----------------------------------------------------------------------------

struct XYChild : public Child {
	XYChild();
};

typedef PChild PXYChild;

class XYContainer : public Container {

public:
	void set(const Children & children);
	Children get(bool placeholders);

	virtual Point getShift();

protected:
	virtual void insert(PChild child) = 0;
	virtual void move(PChild child) = 0;

	PChild find(PGtkWidget widget, const Children & children);
	Point getXY(Gtk::Widget * widget);

};

class FixedContainer : public XYContainer {

public:
	Gtk::Fixed * getContainer() { return dynamic_cast<Gtk::Fixed *>(Container::getContainer()); }

	Point getShift();

protected:
	void insert(PChild child);
	void move(PChild child);

};

class LayoutContainer : public XYContainer {

public:
	Gtk::Layout * getContainer() { return dynamic_cast<Gtk::Layout *>(Container::getContainer()); }

	Point getShift();

protected:
	void insert(PChild child);
	void move(PChild child);

};

// -----------------------------------------------------------------------------

struct RowChild : public Child {
	RowChild();
};

typedef PChild PRowChild;

class RowContainer : public Container {

protected:
	void update(int item, PChild child, bool placeholders);

	virtual Gtk::Widget * getWidget(int item) = 0;
	virtual void setWidget(int item, Gtk::Widget * widget) = 0;
	virtual void updateChild(int item, PChild child) { }

};

// -----------------------------------------------------------------------------

class BinContainer : public RowContainer {

public:
	Gtk::Bin * getContainer() { return dynamic_cast<Gtk::Bin *>(Container::getContainer()); }

	void set(const Children & children, bool placeholders);
	Children get(bool placeholders);

protected:
	Gtk::Widget * getWidget(int item);
	void setWidget(int item, Gtk::Widget * widget);

};

class ScrolledWindowContainer : public BinContainer {

public:
	Gtk::ScrolledWindow * getContainer() { return dynamic_cast<Gtk::ScrolledWindow *>(Container::getContainer()); }

protected:
	Gtk::Widget * getWidget(int item);
	void setWidget(int item, Gtk::Widget * widget);

};

class DialogContainer : public BinContainer {

public:
	Gtk::Dialog * getContainer();

protected:
	Gtk::Bin * getDialogBin();
	Gtk::Widget * getWidget(int item);
	void setWidget(int item, Gtk::Widget * widget);

};

class InfoBarContainer : public BinContainer {

public:
	Gtk::InfoBar * getContainer();

protected:
	Gtk::Bin * getInfoBarBin();
	Gtk::Widget * getWidget(int item);
	void setWidget(int item, Gtk::Widget * widget);

};

// -----------------------------------------------------------------------------

struct PanedChild : public RowChild {
	bool resize;
	bool shrink;
	PanedChild();
};

typedef Glib::RefPtr<PanedChild> PPanedChild;

class PanedContainer : public RowContainer {

public:
	Gtk::Paned * getContainer() { return dynamic_cast<Gtk::Paned *>(Container::getContainer()); }

	void set(const Children & children, bool placeholders);
	Children get(bool placeholders);

protected:
	PChild createChild();
	Gtk::Widget * getWidget(int item);
	void setWidget(int item, Gtk::Widget * widget);
	void updateChild(int item, PChild child);

};

// -----------------------------------------------------------------------------

class IndexContainer : public Container {

protected:
	void process(int capacity);

	virtual int find(PGtkWidget widget, int start) = 0;
	virtual void reorder(int index1, int index2) = 0;
	virtual void shrink(int size) = 0;
	virtual PChild input(int index) = 0;
	virtual void insert(int index) = 0;
	virtual void update(int index) = 0;

};

// -----------------------------------------------------------------------------

struct BoxChild : public Child {
	Gtk::PackOptions options;
	int padding;
	BoxChild();
};

typedef Glib::RefPtr<BoxChild> PBoxChild;

class BoxContainer : public IndexContainer {

public:
	Gtk::Box * getContainer() { return dynamic_cast<Gtk::Box *>(Container::getContainer()); }

	void set(const Children & children, int capStart, int capEnd);
	Children get(bool placeholders);

protected:
	Children sorted;

	PChild createChild();

	int find(PGtkWidget widget, int start);
	void reorder(int index1, int index2);
	void shrink(int size);
	PChild input(int i);
	void insert(int index);
	void update(int index);

	virtual void postUpdate(Gtk::Box_Helpers::BoxList::iterator child, int index) { }

};

// -----------------------------------------------------------------------------

struct ButtonBoxChild : public BoxChild {
	bool secondary;
	ButtonBoxChild();
};

typedef Glib::RefPtr<ButtonBoxChild> PButtonBoxChild;

class ButtonBoxContainer : public BoxContainer {

public:
	Gtk::ButtonBox * getContainer() { return dynamic_cast<Gtk::ButtonBox *>(Container::getContainer()); }

protected:
	PChild createChild();

	void postUpdate(Gtk::Box_Helpers::BoxList::iterator child, int index);

};

// -----------------------------------------------------------------------------

class PagesContainer : public IndexContainer {

friend class PagesViewBase;

public:
	void set(const Children & children, int capacity);
	Children get(bool placeholders);

protected:
	int find(PGtkWidget widget, int start);
	void shrink(int size);
	PChild input(int i);

	virtual int getCurrentPage() = 0;
	virtual void setCurrentPage(int page) = 0;
	virtual int getPageCount() = 0;
	virtual Gtk::Widget * getPageWidget(int page) = 0;

	virtual void revive() { }

private:
	Children sorted;

};

// -----------------------------------------------------------------------------

struct NotebookChild : public Child {
	Glib::ustring tabLabel;
	PGtkWidget tabLabelWidget;
	Glib::ustring menuLabel;
	PGtkWidget menuLabelWidget;
	Gtk::PackOptions packOptions;
	NotebookChild();
};

typedef Glib::RefPtr<NotebookChild> PNotebookChild;

class NotebookContainer : public PagesContainer {

public:
	Gtk::Notebook * getContainer() { return dynamic_cast<Gtk::Notebook *>(Container::getContainer()); }

protected:
	PChild createChild();

	void reorder(int index1, int index2);
	void insert(int index);
	void update(int index);

	int getCurrentPage() { return getContainer()->get_current_page(); }
	void setCurrentPage(int page) { getContainer()->set_current_page(page); }
	int getPageCount() { return getContainer()->get_n_pages(); }
	Gtk::Widget * getPageWidget(int page) { return getContainer()->get_nth_page(page); }

};

// -----------------------------------------------------------------------------

struct AssistantChild : public Child {
	Gtk::AssistantPageType pageType;
	bool complete;
	Glib::ustring title;
	std::string headerImage;
	std::string sidebarImage;
	AssistantChild();
};

typedef Glib::RefPtr<AssistantChild> PAssistantChild;

class AssistantContainer : public PagesContainer {

public:
	Gtk::Assistant * getContainer() { return dynamic_cast<Gtk::Assistant *>(Container::getContainer()); }

protected:
	PChild createChild();

	void reorder(int index1, int index2);
	void insert(int index);
	void update(int index);

	int getCurrentPage() { return getContainer()->get_current_page(); }
	void setCurrentPage(int page) { getContainer()->set_current_page(page); }
	int getPageCount() { return getContainer()->get_n_pages(); }
	Gtk::Widget * getPageWidget(int page) { return getContainer()->get_nth_page(page); }

	void revive() { ReviveGtkAssistant(getContainer()); }

	void updateImage(Gtk::Widget * child, string_cr prop, string_cr newimg);

};

// -----------------------------------------------------------------------------

struct TableChild : public Child {
	int paddingX;
	int paddingY;
	bool xExpand;
	bool xFill;
	bool xShrink;
	bool yExpand;
	bool yFill;
	bool yShrink;
	TableChild();
};

typedef Glib::RefPtr<TableChild> PTableChild;

class TableContainer : public Container {

public:
	TableContainer();
	Gtk::Table * getContainer() { return dynamic_cast<Gtk::Table *>(Container::getContainer()); }

	void set(const Children & children, int capX, int capY, bool placeholders);
	Children get(bool placeholders);

protected:

	struct less { bool operator () (const PChild & ch1, const PChild & ch2) const; };

	PChild createChild();

	bool input(int x, int y);
	void put(int x, int y, PChild child);
	bool mark(int x, int y, int s);
	void process(bool placeholders);
	void move(PTableChild child);
	void insert(PTableChild child);
	void update(PTableChild child);

	Gtk::AttachOptions getAttachOptions(bool expand, bool fill, bool shrink);

private:
	Children sorted;
	PChild skip;
	int capX, capY;

};

}

#endif
