/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "forward.h"
#include "crow_impl.h"

#include "views2.h"
#include "palette.h"
#include "designer.h"

namespace Crow {

CROW_API PDesigner CreateDesigner() {
	return MakeRefPtr(new Crow::DesignerImpl());
}

CROW_API int GetCrowVersion() {
	return CROW_VERSION;
}

CROW_API Enter::Enter(const char * msg) {
	guiloader_enter_check(msg);
	Crow::CrowInit::enter(msg ? msg : "");
}

CROW_API Enter::~Enter() {
	Crow::CrowInit::leave();
	guiloader_leave();
}

// CrowInit --------------------------------------------------------------------

CrowInit * CrowInit::get() {
	CHECK(ptr());
	return ptr();
}

void CrowInit::enter(string_cr msg) {
	if(!ptr()) {
		ptr()= new CrowInit();
		ptr()->init();
	} else
		++ptr()->refcounter;
	ptr()->snapshot(msg);
}

void CrowInit::leave() {
	CHECK(ptr() && ptr()->refcounter > 0);
	ptr()->check();
	--ptr()->refcounter;
	if(ptr()->refcounter == 0) {
		delete ptr();
		ptr()= NULL;
	}
}

CrowInit * & CrowInit::ptr() {
	static CrowInit * crowInit= NULL;
	return crowInit;
}

CrowInit::CrowInit() {
	CHECK(GTK_MAJOR_VERSION*1000+GTK_MINOR_VERSION <= gtk_major_version*1000+gtk_minor_version);
	CHECK(GUILOADER_VERSION/1000 <= guiloader_version()/1000);
	refcounter= 1;
	palette= new Palette();
}

CrowInit::~CrowInit() {
	delete palette;
}

void CrowInit::init() {
	InitGettext();
	palette->init();
}

void CrowInit::snapshot(string_cr msg) {
	labels.push_back(msg);
	if(!msg.empty())
		snapshots.push_back(objects);
}

void CrowInit::check() {

	std::string msg= labels.back();
	labels.pop_back();

	if(!msg.empty()) {

		#ifdef CROW_SANDBOX

			const Set & snapshot= snapshots.back();

			CHECK(std::includes(objects.begin(), objects.end(), snapshot.begin(), snapshot.end()));

			std::list<Crow::Object *> leaked;
			std::set_difference(
				objects.begin(), objects.end(),
				snapshot.begin(), snapshot.end(),
				std::back_inserter(leaked)
			);

			int leaks= leaked.size();

			if(leaks > 0) {

				std::cerr << "[Possible leaks in " << msg << ": " << ToString(leaks) << "]" << std::endl;

				while(!leaked.empty()) {
					Crow::Object * obj= leaked.front();
					std::string cname= GetTypeName(typeid(*obj));
					std::cerr << "---------" << cname << ", " << obj->refcounter;

					CAny * any= dynamic_cast<CAny *>(obj);
					if(any && any->getType()) {
						std::cerr << ", "
							<< *palette->getEntry(any->getType())->getTypeName()
							<< "=" << any->toString();
					}

					Node * node= dynamic_cast<Node *>(obj);
					if(node) {
						std::cerr
							<< ", name=" << *node->getName()
							<< ", out=" << ((node->getRole()==nrScalar || node->getRole()==nrLink) ? 0 : node->getOut()->size())
							<< ", in=" << node->getIn()->size();
					}

					std::cerr << std::endl;
					leaked.pop_front();
				}

			}

		#else

			int snapshot= snapshots.back();
			int leaks= objects - snapshot;
			CHECK(leaks >= 0);

		#endif

		if(leaks != 0) {
			Glib::ustring str= "Possible leaks in " + msg + ": " + ToString(leaks);
			#ifndef CROW_SANDBOX
				std::cerr << str << std::endl;
			#endif
			//ShowMessage(str);
		}

		snapshots.pop_back();

	}

}

void CrowInit::objectCreated(Crow::Object * object) {
	#ifdef CROW_SANDBOX
		CHECK(objects.insert(object).second);
	#else
		++objects;
	#endif
}

void CrowInit::objectDeleted(Crow::Object * object) {
	#ifdef CROW_SANDBOX
		CHECK(objects.erase(object)==1);
	#else
		--objects;
	#endif
}

}
