/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#ifndef CROW_EDITORS_H
#define CROW_EDITORS_H

#include "util.h"
#include "widgets.h"

namespace Crow {

class Editor : public Object {

public:
	void init(PSession session);
	void done();

	PEditorWidget getEditorWidget();
	template <class T> Glib::RefPtr<T> getEditorWidget() { return Glib::RefPtr<T>::cast_dynamic(getEditorWidget()); }

	virtual void created() { }
	virtual void load(bool reload) { }
	virtual void applySettings() { }
	virtual void storeSettings() { }
	virtual void deleted() { }

	PSession getSession();

	typedef sigc::signal<void> SignalPopupMenu;
	sigc::connection connect_popup_menu(SignalPopupMenu::slot_type slot) { return signalPopupMenu.connect(slot); }

	Controller * getController();

protected:
	SessionManager * getManager();

	virtual PEditorWidget createEditorWidget() = 0;
	void setEditorWidget(PEditorWidget);

	void popupMenu() { signalPopupMenu.emit(); }

private:
	PSession session;
	PEditorWidget widget;
	SignalPopupMenu signalPopupMenu;

};

template <class T> class TCreateEditorWidget : public virtual Editor {

public:
	Glib::RefPtr<T> getEditorWidget() { return Editor::getEditorWidget<T>(); }

protected:
	PEditorWidget createEditorWidget() { return MakeRefPtr<T>(); }

};

class HierarchyEditor : public TCreateEditorWidget<HierarchyEditorWidget> {

public:
	void created();
	void load(bool reload);
	void deleted();

protected:
	void setRowValues(PPolyelem elem);
	void onManagerSelectionChanged();
	void onTreeViewSelectionChanged();
	bool validator(ustring_cr oldtext, ustring_cr newtext);
	void onRename(PolycellInput * polycell, PNode node);

};

class CanvasEditor : public virtual Editor {

public:
	virtual bool startPasting(ustring_cr data, bool normal) { return false; }
	virtual void stopPasting() { }

	void pastingFinished() { signalPastingFinished.emit(); }
	void statusChanged(ustring_cr str) { signalStatusChanged.emit(str); }

	typedef sigc::signal<void> SignalPastingFinished;
	typedef sigc::signal<void, Glib::ustring> SignalStatusChanged;

	sigc::connection connect_pasting_finished(SignalPastingFinished::slot_type slot) { return signalPastingFinished.connect(slot); }
	sigc::connection connect_status_changed(SignalStatusChanged::slot_type slot) { return signalStatusChanged.connect(slot); }

private:
	SignalPastingFinished signalPastingFinished;
	SignalStatusChanged signalStatusChanged;

};

class NullCanvasEditor : public CanvasEditor {

protected:
	PEditorWidget createEditorWidget();

};

class PropertyEditor : public virtual Editor {

public:
	PPropertyEditorWidget getEditorWidget();

	void created();

protected:
	virtual void onAccepted() { }

	TypeId getEditingFinalType();
	Property * getProperty1();

	Any getScalar();
	void setScalar(Any value);
	void setScalarAndMeta(Any value, ustring_cr meta);

};

class VectorPropertyEditor : public PropertyEditor, public TCreateEditorWidget<PropertyEditorWidgetText> {

public:
	using TCreateEditorWidget<PropertyEditorWidgetText>::getEditorWidget;

	void load(bool reload);

};

class ObjectPropertyEditor : public PropertyEditor, TCreateEditorWidget<PropertyEditorWidgetInput> {

public:
	using TCreateEditorWidget<PropertyEditorWidgetInput>::getEditorWidget;

	void created();
	void load(bool reload);

protected:
	void onPopup();
	void onMenuClear();
	void onMenuSelect(PNode object);
	void onMenuCreate(TypeId type);

};

class PropertyEditorInput : public PropertyEditor, public TCreateEditorWidget<PropertyEditorWidgetInput> {

public:
	using TCreateEditorWidget<PropertyEditorWidgetInput>::getEditorWidget;

	void created();
	void load(bool reload);

protected:
	void onAccepted();

	virtual Glib::ustring toText(Any value);
	virtual Any fromText(ustring_cr text);

private:
	void onPopup();

};

class PropertyEditorChoice : public PropertyEditor, public TCreateEditorWidget<PropertyEditorWidgetChoice> {

public:
	using TCreateEditorWidget<PropertyEditorWidgetChoice>::getEditorWidget;

	void load(bool reload);

};

class BoolPropertyEditor : public PropertyEditorChoice {

public:
	void created();

protected:
	void onAccepted();

};

class EnumPropertyEditor : public PropertyEditorChoice {

public:
	void created();

protected:
	void onAccepted();

};

class FlagsPropertyEditor : public PropertyEditorInput {

public:
	void created();

private:
	void onPopup();

};

class EmitterPropertyEditor : public PropertyEditorInput {

public:
	EmitterPropertyEditor();
	void created();

protected:
	std::string getMySignature();
	bool validator(ustring_cr oldtext, ustring_cr newtext);
	Glib::ustring toText(Any value);
	Any fromText(ustring_cr text);
	void onPopup();
	void onOrderSelect(Gtk::Menu * menu);
	void onMenuSelect(string_cr name);

private:
	bool lock;

};

class ColorPropertyEditor : public PropertyEditor, public TCreateEditorWidget<ColorPropertyEditorWidget> {

public:
	using TCreateEditorWidget<ColorPropertyEditorWidget>::getEditorWidget;

	void load(bool reload);

protected:
	void onAccepted();

};

class StockIDPropertyEditor : public PropertyEditorInput {

public:
	void created();

private:
	void onPopup();

};

class IconNamePropertyEditor : public PropertyEditorInput {

public:
	void created();

private:
	void onPopup();

};

class ElementPropertyEditor : public PropertyEditorInput {

public:
	void created();

private:
	void onPopup();
	void onMenuSelect(string_cr choice);

};

}

#endif
