/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person 
 *  obtaining a copy of this software and associated documentation 
 *  files (the "Software"), to deal in the Software without 
 *  restriction, including without limitation the rights to use, copy, 
 *  modify, merge, publish, distribute, sublicense, and/or sell copies 
 *  of the Software, and to permit persons to whom the Software is 
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind, 
 *  express or implied, including but not limited to the warranties of 
 *  merchantability, fitness for a particular purpose and 
 *  noninfringement. In no event shall the authors or copyright 
 *  holders be liable for any claim, damages or other liability, 
 *  whether in an action of contract, tort or otherwise, arising from, 
 *  out of or in connection with the software or the use or other 
 *  dealings in the software.
 */

#ifndef CROW_MODEL_H
#define CROW_MODEL_H

#include "util.h"

namespace Crow {

enum NodeRole { nrNone, nrScalar, nrVector, nrEntity, nrLink };

class Node : public Object {

friend class Model;

public:
	NodeRole getRole() { return role; }
	TypeId getType() { CHECK(role!=nrVector && role!=nrLink); return type; }
	StateFlags getState() { CHECK(role!=nrVector && role!=nrEntity); return state; }

	const std::string * getName() { return &name; }
	Any getValue();
	const Glib::ustring * getMeta() { return &meta; }

	bool isInactive();
	bool isModified() { return modified; }

	bool isWeakLink() { return state & sfWeakLink; }
	bool isNoUndo() { return state & sfNoUndo; }

	const Nodes * getOut() { CHECK(role!=nrLink && role!=nrScalar); return &out; }
	const Nodes * getIn() { return &in; }

	PNode getOwner();
	PNode getLink();

protected:
	Node(NodeRole r, TypeId t, StateFlags s=sfNone);
	static PNode create(NodeRole r, TypeId t, StateFlags s=sfNone);

	void setModified(bool b) { modified= b; }

	void setName(string_cr n) { name= n; }
	void setValue(Any any);
	void setMeta(ustring_cr m) { meta= m; }

	void pushOut(PNode node);
	void pushIn(PNode node);

	void eraseOut(PNode node);
	void eraseIn(PNode node);

	void setOwner(PNode owner);
	void setLink(PNode link);

private:
	NodeRole role;
	TypeId type;
	StateFlags state;

	bool modified;

	std::string name;
	Any value;
	Glib::ustring meta;
	Nodes in, out;

	PNode owner;
	PNode link;

};

enum UpdateMode { umNone, umNormal, umSeek, umLoad, umClear, umPaste };

enum OperationCode {
	ocNone,
	ocSetOwner,
	ocModified,
	ocSetName,
	ocSetMeta,
	ocSetValue,
	ocPushIn,
	ocPushOut,
	ocEraseIn,
	ocEraseOut,
	ocSetLink
};

struct Operation : public Object {
	OperationCode code;
	PNode node;

	static POperation create(OperationCode oc, PNode node);
	static POperation create(OperationCode oc, PNode node, string_cr oldname, string_cr newname);
	static POperation create(OperationCode oc, PNode node, ustring_cr oldmeta, ustring_cr newmeta);
	static POperation create(OperationCode oc, PNode node, Any oldvalue, Any newvalue);
	static POperation create(OperationCode oc, PNode node, PNode link);
	static POperation create(OperationCode oc, PNode node, PNode oldnode, PNode newnode);

};

struct OperationName : public Operation {
	std::string oldname;
	std::string newname;
};

struct OperationMeta : public Operation {
	Glib::ustring oldmeta;
	Glib::ustring newmeta;
};

struct OperationValue : public Operation {
	Any oldvalue;
	Any newvalue;
};

struct OperationLink : public Operation {
	PNode link;
};

struct OperationNode : public Operation {
	PNode oldnode;
	PNode newnode;
};

typedef Bits ChangeType;
const ChangeType ctNone   = 0;
const ChangeType ctAppend = 1;
const ChangeType ctRemove = 2;
const ChangeType ctEdit   = 4;

class Model {

public:
	Model();
	~Model();

	PNode getRoot();
	bool empty();
	bool emptyHistory();
	UpdateMode getUpdateMode() { return mode; }
	bool testChange(ChangeType ctype);

	bool validName(string_cr name);
	std::string makeUnique(string_cr name);
	std::string suggestName(TypeId type);

	bool isModified();
	void tagModified(PNode node=PNode());

	IdPath getNodePath(PNode node, PNode cut=PNode());
	IdPath getChildPath(PNode node);

	Nodes getNested(PNode node, bool all=true);
	Nodes getChildren(PNode node, bool all=true);
	Nodes getSlaves(PNode node, bool all=true);
	NodeSet growSlaves(const Nodes & nodes);
	Nodes selectMasters(const Nodes & nodes);
	Nodes selectParents(const Nodes & nodes);

	Nodes query(TypeId baseType, bool all=true);
	PNode findObject(string_cr name);
	PNode find(PNode owner, string_cr name);
	PNode findParent(PNode node);
	PNode findMaster(PNode node);
	PNode raiseOwner(PNode node);
	PNode raiseParent(PNode node);
	PNode raiseMaster(PNode node);
	bool hasParent(PNode child, PNode parent);
	bool hasMaster(PNode child, PNode master);
	PNode findChildNode(PNode node);
	PNode getWidgetNode(PNode node);
	Any getScalar(PNode owner, string_cr name);
	bool vectorEmptyOrNullLinks(PNode node);

	PNode createScalar(PNode owner, string_cr name, Any value, StateFlags sf=sfNone);
	PNode createScalar(PNode owner, TypeId type, string_cr name, StateFlags sf=sfNone);
	PNode createVector(PNode owner, string_cr name);
	PNode createEntity(PNode owner, TypeId type, string_cr name);
	PNode createLink(PNode owner, string_cr name, StateFlags sf);

	PNode vectorCreateEntity(PNode owner, TypeId type);
	PNode vectorCreateScalar(PNode owner, Any value, StateFlags sf);
	PNode vectorCreateLink(PNode owner, StateFlags sf);

	void cut(PNode node);
	void paste(PNode node, PNode owner);
	void remove(PNode node);
	void rootRemove(PNode entity);
	void vectorClear(PNode node);
	void vectorRemove(PNode node);
	void vectorMove(PNode node, int shift);
	void ensureOut(PNode node, const std::vector<std::string> & names);

	void clearLink(PNode node);
	void setLink(PNode node, PNode target);
	void setScalar(PNode node, Any value);
	void setScalar(PNode owner, string_cr name, Any value);
	void setName(PNode node, string_cr name);
	void setMeta(PNode node, ustring_cr meta);
	void setChildNode(PNode widget, PNode child);

	void clearState();
	void begin(UpdateMode mode);
	void clean();
	void seek(int ptr);
	void rollback();
	void commit();

	void clear();

	int getPointer() { return hpointer; }
	int getMarkCount() { return hsize+1; }

	void setReadOnly(bool ro) { readOnly= ro; }

	void clearLinksCleared() { gatherLinks= true; }
	bool getLinksCleared();

protected:
	bool isChildrenVector(PNode node);
	void getNested(PNode node, Nodes * nodes, bool all);
	void getChildren(PNode node, Nodes * nodes, bool all);
	void getSlaves(PNode node, Nodes * nodes, bool all);
	void query(PNode node, TypeId baseType, Nodes * nodes, bool all);

	PNode createNode(NodeRole role, TypeId type, StateFlags sf=sfNone);

	void op_setOwner(PNode node, PNode owner);
	void op_modified(PNode node);
	void op_setName(PNode node, string_cr name);
	void op_setMeta(PNode node, ustring_cr meta);
	void op_setValue(PNode node, Any value);
	void op_pushIn(PNode node, PNode push);
	void op_pushOut(PNode node, PNode push);
	void op_eraseIn(PNode node, PNode erase);
	void op_eraseOut(PNode node, PNode erase);
	void op_setLink(PNode node, PNode link);

	void setModified(PNode node);
	void clearModified(PNode node);

	bool redundantLog();
	void undo();
	void redo();
	void appendHistory();
	void clearHistory();
	void undo(POperation op);
	void redo(POperation op);

private:
	PNode root;
	UpdateMode mode;
	ChangeType changeType;

	typedef std::vector<POperation> Log;
	typedef std::list<Log> History;

	History history;
	Log log;

	History::iterator pointer;
	int hpointer;
	int hsize;

	bool dirty;
	bool readOnly;
	
	bool gatherLinks;
	Nodes linksCleared;

};

}

#endif
