/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "forward.h"
#include "palette.h"

#include "crow_impl.h"
#include "editors.h"

namespace Crow {

// Palette ---------------------------------------------------------------------

Palette::Palette() {
	leaks= 0;
	oldLeaks= 0;
	sigclass= G_TYPE_INVALID;
}

Palette::~Palette() {
	for(Entries::iterator it= entries.begin(); it!=entries.end(); ++it)
		delete *it;
	entries.clear();
	for(Editors::iterator it= editors.begin(); it!=editors.end(); ++it)
		delete *it;
	editors.clear();
}

void Palette::addEditor(EditorConstructor * editor, string_cr ename) {
	editors.push_back(editor);
	editor->setName(ename);
}

EditorId Palette::findEditor(string_cr ename) {
	for(Editors::iterator it= editors.begin(); it!=editors.end(); ++it)
		if(*(*it)->getName() == ename)
			return *it;
	return InvalidEditorId;
}

PEditor Palette::createEditor(EditorId eid) {
	CHECK(eid != InvalidEditorId);
	return eid->create();
}

TypeId Palette::findType(string_cr name) {
	for(Entries::iterator it= entries.begin(); it!=entries.end(); ++it)
		if(*(*it)->getTypeName() == name)
			return (*it)->getType();
	return InvalidTypeId;
}

TypeId Palette::findType(GType gtype) {
	if(gtype != G_TYPE_INVALID)
		for(Entries::iterator it= entries.begin(); it!=entries.end(); ++it)
			if((*it)->getGType() == gtype)
				return (*it)->getType();
	return InvalidTypeId;
}

TypeId Palette::findType(PGlibObject object) {
	TypeId type= InvalidTypeId;
	if(object) {
		std::string hint= GetData(GetPtr(object), "CrowTypeHint");
		if(!hint.empty()) {
			type= findType(hint);
		} else {
			GType gtype= GetGType(GetPtr(object));
			type= findType(gtype);
		}
	}
	return type;
}

TypeList Palette::queryTypes(TypeId baseType) {
	TypeList result;
	for(Entries::iterator it= entries.begin(); it!=entries.end(); ++it)
		if(typeIsA((*it)->getType(), baseType))
			result.push_back((*it)->getType());

	struct tmp { static bool less(TypeId a, TypeId b) {
		return *Palette::get()->getEntry(a)->getTypeName() < *Palette::get()->getEntry(b)->getTypeName();
	}};

	result.sort(sigc::ptr_fun(&tmp::less));

	return result;
}

bool Palette::typeIsA(TypeId type, TypeId base) {
	CHECK(base && type);
	if(type==base)
		return true;
	if(isEntity(type)) {
		if(getEntry(type)->getIsA()==base)
			return true;
		GType gtype= getEntry(type)->getGType();
		GType gbase= getEntry(base)->getGType();
		if(gtype!=G_TYPE_INVALID && gbase!=G_TYPE_INVALID)
			return g_type_is_a(gtype, gbase);
	}
	return false;
}

bool Palette::isScalar(TypeId type) {
	return getEntry<ScalarEntry>(type) != NULL;
}

bool Palette::isEntity(TypeId type) {
	return getEntry<EntityEntry>(type) != NULL;
}

bool Palette::isObject(TypeId type) {
	return getEntry<ObjectEntry>(type) != NULL;
}

bool Palette::isWidget(TypeId type) {
	return getEntry<WidgetEntry>(type) != NULL;
}

Glib::ustring Palette::createPasteData(TypeId type) {

	EntityEntry * entry= getEntry<EntityEntry>(type);
	CHECK(entry);
	std::string nname= entry->suggestName();
	string_cr tname= *entry->getTypeName();

	Glib::ustring data=
		"<?xml version='1.0' encoding='UTF-8'?>"
		"<gui namespace='GTK+' version='" + ToString(CROW_VERSION_GUIXML) + "'>"
		"  <entity type='" + tname + "' name='" + nname + "'>"
		"  </entity>"
		"</gui>"
	;

	return data;
}

const std::string * Palette::getTabLabel(PaletteTab tab) {
	TabLabels::iterator it= tabLabels.find(tab);
	CHECK(it != tabLabels.end());
	return &it->second;
}

std::string Palette::findSignal(string_cr name, GType type) {
	Signals::iterator it= signals.lower_bound(name);
	if(it!=signals.end())
		do {
			if(g_type_is_a(type, it->second.first))
				return it->second.second;
			++it;
		} while(it!=signals.end() && it->first==name);
	CHECK(false);
	return std::string();
}

SignalsQuery Palette::querySignals(GType type) {
	SignalsQuery result;
	for(Signals::iterator it= signals.begin(); it!=signals.end(); ++it)
		if(g_type_is_a(type, it->second.first))
			result.push_back(std::make_pair(it->second.first, it->first));

	struct tmp { static bool less(const std::pair<GType, std::string> & a, const std::pair<GType, std::string> & b, GType sub) {
		if(a.first==b.first)
			return a.second < b.second;
		else {
			int da= TypeDist(sub, a.first);
			int db= TypeDist(sub, b.first);
			return da==db ? g_type_is_a(b.first, a.first) : da > db;
		}
	}};

	result.sort(sigc::bind<GType>(sigc::ptr_fun(&tmp::less), type));

	return result;
}

Palette * Palette::get() {
	return CrowInit::get()->getPalette();
}

void Palette::checkLeaks() {
	signalLeaks.emit();
	if(leaks!=oldLeaks)
		std::cerr << "Leaks: " << leaks << std::endl;
	oldLeaks= leaks;
}

PaletteEntry & Palette::takePaletteEntry(PaletteEntry * entry) {
	entries.push_back(entry);
	entry->setType(entry);
	return *entry;
}

PaletteEntry * Palette::getPaletteEntry(TypeId tid) {
	return tid;
}

void Palette::initTabLabels() {
	tabLabels[TAB_NONE]= "";
	tabLabels[TAB_WINDOWS]= _("Windows");
	tabLabels[TAB_UI]= _("UI");
	tabLabels[TAB_STANDARD]= _("Standard");
	tabLabels[TAB_ADDITIONAL]= _("Additional");
	tabLabels[TAB_BINS]= _("Bins");
	tabLabels[TAB_LAYOUTS]= _("Layouts");
	tabLabels[TAB_VIEWS]= _("Views");
	tabLabels[TAB_SELECTORS]= _("Selectors");
	tabLabels[TAB_SPECIAL]= _("Special");
	tabLabels[TAB_DEBUG]= _("Debug");
}

void Palette::init() {

	initTabLabels();

	registerSignals();
	RegisterEditors();

	takeEntry(new TScalarEntry<int>("int", G_TYPE_INT));
	takeEntry(new TScalarEntry<unsigned>("unsigned", G_TYPE_UINT));
	takeEntry(new TScalarEntry<bool>("bool", G_TYPE_BOOLEAN));
	takeEntry(new TScalarEntry<float>("float", G_TYPE_FLOAT));
	takeEntry(new TScalarEntry<double>("double", G_TYPE_DOUBLE));
	takeEntry(new TScalarEntry<Glib::ustring>("string", G_TYPE_STRING));
	takeEntry(new TScalarEntry<UniChar>("char", G_TYPE_UINT));
	takeEntry(new TScalarEntry<Point>("CrowPoint"));
	takeEntry(new TScalarEntry<Emitter>("CrowEmitter"));
	takeEntry(new TScalarEntry<Gdk::Color>("GdkColor", GDK_TYPE_COLOR));
	takeEntry(new TScalarEntry<Gtk::StockID>("GtkStockID", G_TYPE_STRING));
	takeEntry(new TScalarEntry<Border>("GtkBorder", GTK_TYPE_BORDER));

	takeEntry(new AbstractGlibObjectEntry("GlibObject", G_TYPE_OBJECT));

	takeEntry(new TObjectEntry<RelationEntryView>("CrowRelationEntry", G_TYPE_INVALID));
	takeEntry(new EnumEntry("AtkRelationType")).
		addValue("ATK_RELATION_NULL", ATK_RELATION_NULL).
		addValue("ATK_RELATION_CONTROLLER_FOR", ATK_RELATION_CONTROLLER_FOR).
		addValue("ATK_RELATION_LABEL_FOR", ATK_RELATION_LABEL_FOR).
		addValue("ATK_RELATION_MEMBER_OF", ATK_RELATION_MEMBER_OF).
		addValue("ATK_RELATION_FLOWS_TO", ATK_RELATION_FLOWS_TO).
		addValue("ATK_RELATION_SUBWINDOW_OF", ATK_RELATION_SUBWINDOW_OF).
		addValue("ATK_RELATION_EMBEDDED_BY", ATK_RELATION_EMBEDDED_BY).
		addValue("ATK_RELATION_POPUP_FOR", ATK_RELATION_POPUP_FOR).
		addValue("ATK_RELATION_DESCRIPTION_FOR", ATK_RELATION_DESCRIPTION_FOR);

	takeEntry(new AbstractGtkWidgetEntry("GtkWidget", GTK_TYPE_WIDGET));
	takeEntry(new AbstractGtkWidgetEntry("GtkActivatable", GTK_TYPE_ACTIVATABLE));
	takeEntry(new AbstractGlibObjectEntry("CrowChild"));
	takeEntry(new AbstractGtkWidgetEntry("GtkContainer", GTK_TYPE_CONTAINER));

	takeEntry(new FlagsEntry("GdkEventMask")).
		addValue("GDK_EXPOSURE_MASK", Gdk::EXPOSURE_MASK).
		addValue("GDK_POINTER_MOTION_MASK", Gdk::POINTER_MOTION_MASK).
		addValue("GDK_POINTER_MOTION_HINT_MASK", Gdk::POINTER_MOTION_HINT_MASK).
		addValue("GDK_BUTTON_MOTION_MASK", Gdk::BUTTON_MOTION_MASK).
		addValue("GDK_BUTTON1_MOTION_MASK", Gdk::BUTTON1_MOTION_MASK).
		addValue("GDK_BUTTON2_MOTION_MASK", Gdk::BUTTON2_MOTION_MASK).
		addValue("GDK_BUTTON3_MOTION_MASK", Gdk::BUTTON3_MOTION_MASK).
		addValue("GDK_BUTTON_PRESS_MASK", Gdk::BUTTON_PRESS_MASK).
		addValue("GDK_BUTTON_RELEASE_MASK", Gdk::BUTTON_RELEASE_MASK).
		addValue("GDK_KEY_PRESS_MASK", Gdk::KEY_PRESS_MASK).
		addValue("GDK_KEY_RELEASE_MASK", Gdk::KEY_RELEASE_MASK).
		addValue("GDK_ENTER_NOTIFY_MASK", Gdk::ENTER_NOTIFY_MASK).
		addValue("GDK_LEAVE_NOTIFY_MASK", Gdk::LEAVE_NOTIFY_MASK).
		addValue("GDK_FOCUS_CHANGE_MASK", Gdk::FOCUS_CHANGE_MASK).
		addValue("GDK_STRUCTURE_MASK", Gdk::STRUCTURE_MASK).
		addValue("GDK_PROPERTY_CHANGE_MASK", Gdk::PROPERTY_CHANGE_MASK).
		addValue("GDK_VISIBILITY_NOTIFY_MASK", Gdk::VISIBILITY_NOTIFY_MASK).
		addValue("GDK_PROXIMITY_IN_MASK", Gdk::PROXIMITY_IN_MASK).
		addValue("GDK_PROXIMITY_OUT_MASK", Gdk::PROXIMITY_OUT_MASK).
		addValue("GDK_SUBSTRUCTURE_MASK", Gdk::SUBSTRUCTURE_MASK).
		addValue("GDK_SCROLL_MASK", Gdk::SCROLL_MASK);

	takeEntry(new EnumEntry("GdkExtensionMode")).
		addValue("GDK_EXTENSION_EVENTS_NONE", Gdk::EXTENSION_EVENTS_NONE).
		addValue("GDK_EXTENSION_EVENTS_ALL", Gdk::EXTENSION_EVENTS_ALL).
		addValue("GDK_EXTENSION_EVENTS_CURSOR", Gdk::EXTENSION_EVENTS_CURSOR);

	takeEntry(new EnumEntry("CrowTooltipMode")).
		addValue("CROW_TOOLTIP_TEXT", TOOLTIP_TEXT).
		addValue("CROW_TOOLTIP_TEXT_OFF", TOOLTIP_TEXT_OFF).
		addValue("CROW_TOOLTIP_MARKUP", TOOLTIP_MARKUP).
		addValue("CROW_TOOLTIP_MARKUP_OFF", TOOLTIP_MARKUP_OFF);

	RegisterViews1();
	RegisterViews2();
	RegisterViews3();

	registerNodeStates();

}

void Palette::addingSignals(GType gtype) {
	sigclass= gtype;
}

void Palette::addSignal(string_cr name, string_cr signature) {
	signals.insert(std::make_pair(name, std::make_pair(sigclass, signature)));
}

void Palette::addNodeState(string_cr type, string_cr prop, StateFlags sf) {
	TypeId base= findType(type);
	CHECK(base != InvalidTypeId);
	TypeList types= queryTypes(base);
	for(TypeList::iterator it= types.begin(); it!=types.end(); ++it) {
		EntityEntry * entry= getEntry<EntityEntry>(*it);
		CHECK(entry);
		entry->setNodeState(prop, sf);
	}
}

// PaletteEntry ----------------------------------------------------------------

PaletteEntry::PaletteEntry() {
	type= InvalidTypeId;
	gtype= G_TYPE_INVALID;
	isA= InvalidTypeId;
	propertyEditor= InvalidEditorId;
	//priority= 0;
}

void PaletteEntry::setPropertyEditor(string_cr ename) {
	propertyEditor= Palette::get()->findEditor(ename);
}

int PaletteEntry::getPriority() {
	//return priority;
	return 0;
}

void PaletteEntry::setPriority(int pr) {
	//priority= pr;
}

void PaletteEntry::setIsA(string_cr name) {
	isA= FindType(name);
}

std::string PaletteEntry::suggestName() {
	std::string name= typeName;
	if(Begins(name, std::string("Gtk")))
		name= name.substr(3);
	else if(Begins(name, std::string("Crow")))
		name= name.substr(4);
	else
		CHECK(false);

	int count= 0;
	for(int i=0; i < SIGNED(name.length()); ++i)
		if(std::isupper(name[i]))
			++count;
		else if(std::islower(name[i])) {
			if(count > 1)
				--count;
			break;
		}

	for(int i=0; i<count; ++i)
		name[i]= std::tolower(name[i]);

	return name + "1";
}

Any PaletteEntry::createFromGValue(const Glib::ValueBase &) {
	return Any();
}

void PaletteEntry::setGType(const char * gtname) {
	if(gtname) {
		gtype= g_type_from_name(gtname);
		CHECK(gtype != G_TYPE_INVALID);
	} else
		gtype= G_TYPE_INVALID;
}

// EnumEntry -------------------------------------------------------------------

EnumEntry::EnumEntry(string_cr tname) {
	setPropertyEditor("ENUM");
	setTypeName(tname);
	setGType(G_TYPE_INT);
}

Any EnumEntry::createFromGValue(const Glib::ValueBase & val) {
	int v= ExtractCppValue<int>(val);
	return CAny::createEnum(getType(), v);
}

Any EnumEntry::createFromString(ustring_cr str) {
	return CAny::createEnum(getType(), getByName(str));
}

int EnumEntry::getByName(string_cr name) {
	for(Values::iterator it= values.begin(); it!=values.end(); ++it)
		if(it->first == name)
			return it->second;
	CHECK(false);
	return 0;
}

const std::string * EnumEntry::getByValue(int value) {
	for(Values::iterator it= values.begin(); it!=values.end(); ++it)
		if(it->second == value)
			return &it->first;
	CHECK(false);
	return NULL;
}

EnumEntry & EnumEntry::addValue(string_cr name, int value) {
	values.push_back(std::make_pair(name, value));
	return *this;
}

// FlagsEntry -------------------------------------------------------------------

FlagsEntry::FlagsEntry(string_cr tname) {
	setPropertyEditor("FLAGS");
	setTypeName(tname);
	setGType(G_TYPE_UINT);
}

Any FlagsEntry::createFromGValue(const Glib::ValueBase & val) {
	Bits v= ExtractCppValue<Bits>(val);
	return CAny::createFlags(getType(), v);
}

Any FlagsEntry::createFromString(ustring_cr str) {
	std::istringstream iss(str);
	Bits value;
	iss >> std::hex >> value;
	if(iss.fail() || !valid(value))
		RaiseError(_("createFromString() failed"));
	return CAny::createFlags(getType(), value);
}

FlagsEntry & FlagsEntry::addValue(string_cr name, Bits value) {
	values.push_back(std::make_pair(name, value));
	return *this;
}

bool FlagsEntry::valid(Bits value) {
	Bits all= 0;
	for(Values::iterator it= values.begin(); it!=values.end(); ++it)
		all |= it->second;
	return (all & value) == value;
}

// EntityEntry -----------------------------------------------------------------

EntityEntry::EntityEntry() {
	paletteTab= TAB_NONE;
	canvasEditor= InvalidEditorId;
	flags= efNone;
}

EntityEntry & EntityEntry::setCanvasEditor(string_cr ename) {
	canvasEditor= Palette::get()->findEditor(ename);
	return *this;
}

StateFlags EntityEntry::getNodeState(string_cr pname) {
	NSMap::iterator it= nsmap.find(pname);
	return it==nsmap.end() ? sfNone : it->second;
}

void EntityEntry::setNodeState(string_cr property, StateFlags sf) {
	nsmap[property]= sf;
}

PEntityView EntityEntry::createView() {
	return PEntityView();
}

void EntityEntry::prepareView(PEntityView view) {
	view->prepare(getType());
}

// ObjectEntry -----------------------------------------------------------------

Any ObjectEntry::createFromGValue(const Glib::ValueBase & gvalue) {
	PGlibObject object= ExtractCppValue<PGlibObject>(gvalue);
	TypeId typeId= Palette::get()->findType(object);
	CHECK(!object || typeId);
	return CAny::createObject(typeId, object);
}

// AbstractGlibObjectEntry -----------------------------------------------------

AbstractGlibObjectEntry::AbstractGlibObjectEntry(string_cr tname, GType gt) {
	setTypeName(tname);
	setGType(gt);
	setPropertyEditor("GlibObject");
	setFlags(efAbstract);
}

// AbstractGtkWidgetEntry ------------------------------------------------------

AbstractGtkWidgetEntry::AbstractGtkWidgetEntry(string_cr tname, GType gt) {
	setTypeName(tname);
	setGType(gt);
	setPropertyEditor("GlibObject");
	setFlags(efAbstract);
}

// -----------------------------------------------------------------------------

void Palette::registerSignals() {

	addingSignals(G_TYPE_OBJECT);
	addSignal("notify", "void, GParamSpec *");

	addingSignals(GTK_TYPE_OBJECT);
	addSignal("destroy", "void");

	addingSignals(GTK_TYPE_WIDGET);
	addSignal("button-press-event", "gboolean, GdkEventButton *");
	addSignal("button-release-event", "gboolean, GdkEventButton *");
	addSignal("child-notify", "void, GParamSpec *");
	addSignal("client-event", "gboolean, GdkEventClient *");
	addSignal("configure-event", "gboolean, GdkEventConfigure *");
	addSignal("delete-event", "gboolean, GdkEvent *");
	addSignal("direction-changed", "void, GtkTextDirection");
	addSignal("drag-begin", "void, GdkDragContext *");
	addSignal("drag-data-delete", "void, GdkDragContext *");
	addSignal("drag-data-get", "void, GdkDragContext *, GtkSelectionData *, guint, guint");
	addSignal("drag-data-received", "void, GdkDragContext *, gint, gint, GtkSelectionData *, guint, guint");
	addSignal("drag-drop", "gboolean, GdkDragContext *, gint, gint, guint");
	addSignal("drag-end", "void, GdkDragContext *");
	addSignal("drag-leave", "void, GdkDragContext *, guint");
	addSignal("drag-motion", "gboolean, GdkDragContext *, gint, gint, guint");
	addSignal("enter-notify-event", "gboolean, GdkEventCrossing *");
	addSignal("event", "gboolean, GdkEvent *");
	addSignal("event-after", "void, GdkEvent *");
	addSignal("expose-event", "gboolean, GdkEventExpose *");
	addSignal("focus", "gboolean, GtkDirectionType");
	addSignal("focus-in-event", "gboolean, GdkEventFocus *");
	addSignal("focus-out-event", "gboolean, GdkEventFocus *");
	addSignal("grab-focus", "void");
	addSignal("grab-notify", "void, gboolean");
	addSignal("hide", "void");
	addSignal("hierarchy-changed", "void, GtkWidget *");
	addSignal("key-press-event", "gboolean, GdkEventKey *");
	addSignal("key-release-event", "gboolean, GdkEventKey *");
	addSignal("leave-notify-event", "gboolean, GdkEventCrossing *");
	addSignal("map", "void");
	addSignal("map-event", "gboolean, GdkEvent *");
	addSignal("mnemonic-activate", "gboolean, gboolean");
	addSignal("motion-notify-event", "gboolean, GdkEventMotion *");
	addSignal("no-expose-event", "gboolean, GdkEventNoExpose *");
	addSignal("parent-set", "void, GtkObject *");
	addSignal("property-notify-event", "gboolean, GdkEventProperty *");
	addSignal("proximity-in-event", "gboolean, GdkEventProximity *");
	addSignal("proximity-out-event", "gboolean, GdkEventProximity *");
	addSignal("realize", "void");
	addSignal("screen-changed", "void, GdkScreen *");
	addSignal("scroll-event", "gboolean, GdkEventScroll *");
	addSignal("selection-clear-event", "gboolean, GdkEventSelection *");
	addSignal("selection-get", "void, GtkSelectionData *, guint, guint");
	addSignal("selection-notify-event", "gboolean, GdkEventSelection *");
	addSignal("selection-received", "void, GtkSelectionData *, guint");
	addSignal("selection-request-event", "gboolean, GdkEventSelection *");
	addSignal("show", "void");
	addSignal("size-allocate", "void, GtkAllocation *");
	addSignal("size-request", "void, GtkRequisition *");
	addSignal("state-changed", "void, GtkStateType");
	addSignal("style-set", "void, GtkStyle *");
	addSignal("unmap", "void");
	addSignal("unmap-event", "gboolean, GdkEvent *");
	addSignal("unrealize", "void");
	addSignal("visibility-notify-event", "gboolean, GdkEventVisibility *");
	addSignal("window-state-event", "gboolean, GdkEventWindowState *");
	addSignal("composited-changed", "void");
	addSignal("query-tooltip", "gboolean, gint, gint, gboolean, GtkTooltip *");
	addSignal("damage-event", "gboolean, GdkEvent *");
	addSignal("drag-failed", "gboolean, GdkDragContext *, GtkDragResult");
	addSignal("popup-menu", "gboolean");

	addingSignals(GTK_TYPE_CONTAINER);
	addSignal("add", "void, GtkWidget *");
	addSignal("check-resize", "void");
	addSignal("remove", "void, GtkWidget *");
	addSignal("set-focus-child", "void, GtkWidget *");

	addingSignals(GTK_TYPE_HANDLE_BOX);
	addSignal("child-attached", "void, GtkWidget *");
	addSignal("child-detached", "void, GtkWidget *");

	addingSignals(GTK_TYPE_VIEWPORT);
	addSignal("set-scroll-adjustments", "void, GtkAdjustment *, GtkAdjustment *");

	addingSignals(GTK_TYPE_NOTEBOOK);
	addSignal("switch-page", "void, GtkNotebookPage *, guint");
	addSignal("page-added", "void, GtkWidget *, guint");
	addSignal("page-removed", "void, GtkWidget *, guint");
	addSignal("page-reordered", "void, GtkWidget *, guint");

	addingSignals(GTK_TYPE_ADJUSTMENT);
	addSignal("changed", "void");
	addSignal("value-changed", "void");

	addingSignals(GTK_TYPE_WINDOW);
	addSignal("frame-event", "gboolean, GdkEvent *");
	addSignal("set-focus", "void, GtkWidget *");

	addingSignals(GTK_TYPE_ASSISTANT);
	addSignal("apply", "void");
	addSignal("cancel", "void");
	addSignal("close", "void");
	addSignal("prepare", "GtkWidget *");

	addingSignals(GTK_TYPE_DIALOG);
	addSignal("response", "void, gint");

	addingSignals(GTK_TYPE_INFO_BAR);
	addSignal("response", "void, gint");

	addingSignals(GTK_TYPE_FILE_CHOOSER);
	addSignal("current-folder-changed", "void");
	addSignal("file-activated", "void");
	addSignal("selection-changed", "void");
	addSignal("update-preview", "void");
	addSignal("confirm-overwrite", "GtkFileChooserConfirmation");

	addingSignals(GTK_TYPE_FILE_CHOOSER_BUTTON);
	addSignal("file-set", "void");

	addingSignals(GTK_TYPE_STATUSBAR);
	addSignal("text-popped", "void, guint, gchar *");
	addSignal("text-pushed", "void, guint, gchar *");

	addingSignals(GTK_TYPE_ACTION_GROUP);
	addSignal("connect-proxy", "void, GtkAction *, GtkWidget *");
	addSignal("disconnect-proxy", "void, GtkAction *, GtkWidget *");
	addSignal("post-activate", "void, GtkAction *");
	addSignal("pre-activate", "void, GtkAction *");

	addingSignals(GTK_TYPE_ACTION);
	addSignal("activate", "void");

	addingSignals(GTK_TYPE_TOGGLE_ACTION);
	addSignal("toggled", "void");

	addingSignals(GTK_TYPE_RADIO_ACTION);
	addSignal("changed", "void, GtkRadioAction *");

	addingSignals(GTK_TYPE_UI_MANAGER);
	addSignal("actions-changed", "void");
	addSignal("add-widget", "void, GtkWidget *");
	addSignal("connect-proxy", "void, GtkAction *, GtkWidget *");
	addSignal("disconnect-proxy", "void, GtkAction *, GtkWidget *");
	addSignal("post-activate", "void, GtkAction *");
	addSignal("pre-activate", "void, GtkAction *");

	addingSignals(GTK_TYPE_MENU_SHELL);
	addSignal("deactivate", "void");
	addSignal("selection-done", "void");
	addSignal("move-selected", "gboolean, gint");

	addingSignals(GTK_TYPE_TOOLBAR);
	addSignal("orientation-changed", "void, GtkOrientation");
	addSignal("popup-context-menu", "gboolean, gint, gint, gint");
	addSignal("style-changed", "void, GtkToolbarStyle");

	addingSignals(GTK_TYPE_LABEL);
	addSignal("populate-popup", "void, GtkMenu *");

	addingSignals(GTK_TYPE_BUTTON);
	addSignal("activate", "void");
	addSignal("clicked", "void");
	addSignal("enter", "void");
	addSignal("leave", "void");
	addSignal("pressed", "void");
	addSignal("released", "void");

	addingSignals(GTK_TYPE_TOGGLE_BUTTON);
	addSignal("toggled", "void");

	addingSignals(GTK_TYPE_RADIO_BUTTON);
	addSignal("group-changed", "void");

	addingSignals(GTK_TYPE_ENTRY_COMPLETION);
	addSignal("action-activated", "void, gint");
	addSignal("insert-prefix", "gboolean, gchar *");
	addSignal("match-selected", "gboolean, GtkTreeModel *, GtkTreeIter *");
	addSignal("cursor-on-match", "gboolean, GtkTreeModel *, GtkTreeIter *");

	addingSignals(GTK_TYPE_EDITABLE);
	addSignal("changed", "void");
	addSignal("delete-text", "void, gint, gint");
	addSignal("insert-text", "void, gchar *, gint, gint *");

	addingSignals(GTK_TYPE_ENTRY);
	addSignal("activate", "void");
	addSignal("insert-at-cursor", "void, gchar *");
	addSignal("populate-popup", "void, GtkMenu *");
	addSignal("icon-press", "void, GtkEntryIconPosition, GdkEvent *");
	addSignal("icon-release", "void, GtkEntryIconPosition, GdkEvent *");

	addingSignals(GTK_TYPE_COMBO_BOX);
	addSignal("changed", "void");

	addingSignals(GTK_TYPE_SPIN_BUTTON);
	addSignal("input", "gint, gpointer");
	addSignal("output", "gboolean");
	addSignal("value-changed", "void");
	addSignal("wrapped", "void");

	addingSignals(GTK_TYPE_RANGE);
	addSignal("adjust-bounds", "void, gdouble");
	addSignal("change-value", "gboolean, GtkScrollType, gdouble");
	addSignal("move-slider", "void, GtkScrollType");
	addSignal("value-changed", "void");

	addingSignals(GTK_TYPE_SCALE);
	addSignal("format-value", "gchar *, gdouble");

	addingSignals(GTK_TYPE_TEXT_VIEW);
	addSignal("insert-at-cursor", "void, gchar *");
	addSignal("populate-popup", "void, GtkMenu *");
	addSignal("set-anchor", "void");
	addSignal("set-scroll-adjustments", "void, GtkAdjustment *, GtkAdjustment *");

	addingSignals(GTK_TYPE_ICON_VIEW);
	addSignal("item-activated", "void, GtkTreePath *");
	addSignal("selection-changed", "void");
	addSignal("set-scroll-adjustments", "void, GtkAdjustment *, GtkAdjustment *");

	addingSignals(GTK_TYPE_TREE_VIEW);
	addSignal("columns-changed", "void");
	addSignal("cursor-changed", "void");
	addSignal("row-activated", "void, GtkTreePath *, GtkTreeViewColumn *");
	addSignal("row-collapsed", "void, GtkTreeIter *, GtkTreePath *");
	addSignal("row-expanded", "void, GtkTreeIter *, GtkTreePath *");
	addSignal("set-scroll-adjustments", "void, GtkAdjustment *, GtkAdjustment *");
	addSignal("test-collapse-row", "gboolean, GtkTreeIter *, GtkTreePath *");
	addSignal("test-expand-row", "gboolean, GtkTreeIter *, GtkTreePath *");

	addingSignals(GTK_TYPE_COLOR_BUTTON);
	addSignal("color-set", "void");

	addingSignals(GTK_TYPE_COLOR_SELECTION);
	addSignal("color-changed", "void");

	addingSignals(GTK_TYPE_FONT_BUTTON);
	addSignal("font-set", "void");

	addingSignals(GTK_TYPE_LAYOUT);
	addSignal("set-scroll-adjustments", "void, GtkAdjustment *, GtkAdjustment *");

	addingSignals(GTK_TYPE_CALENDAR);
	addSignal("day-selected", "void");
	addSignal("day-selected-double-click", "void");
	addSignal("month-changed", "void");
	addSignal("next-month", "void");
	addSignal("next-year", "void");
	addSignal("prev-month", "void");
	addSignal("prev-year", "void");

	addingSignals(GTK_TYPE_STATUS_ICON);
	addSignal("activate", "void");
	addSignal("popup-menu", "void, guint, guint");
	addSignal("size-changed", "gboolean, gint");
	addSignal("button-press-event", "gboolean, GdkEventButton *");
	addSignal("button-release-event", "gboolean, GdkEventButton *");
	addSignal("scroll-event", "gboolean, GdkEventScroll *");
	addSignal("query-tooltip", "gboolean, gint, gint, gboolean, GtkTooltip *");

	addingSignals(GTK_TYPE_RECENT_CHOOSER);
	addSignal("item-activated", "void");
	addSignal("selection-changed", "void");

	addingSignals(G_TYPE_INVALID);
}

void Palette::registerNodeStates() {
	addNodeState("CrowRelationEntry", "target",          sfWeakLink);
	addNodeState("CrowUIElement",     "manager",         sfWeakLink);
	addNodeState("GtkContainer",      "focus-chain",     sfWeakLink);
	addNodeState("GtkWindow",         "transient-for",   sfWeakLink);
	addNodeState("GtkWindow",         "default",         sfWeakLink);
	addNodeState("GtkRadioAction",    "group",           sfWeakLink);
	addNodeState("GtkLabel",          "mnemonic-widget", sfWeakLink);
	addNodeState("GtkRadioButton",    "group",           sfWeakLink);
	addNodeState("GtkSizeGroup",      "widgets",         sfWeakLink);
	addNodeState("GtkTreeView",       "search-entry",    sfWeakLink);
	addNodeState("GtkAction",         "widgets",         sfWeakLink);
	addNodeState("GtkPaned",          "position",        sfNoUndo);
	addNodeState("GtkNotebook",       "page",            sfNoUndo);
	addNodeState("GtkAssistant",      "page",            sfNoUndo);
}

}
