/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "forward.h"
#include "polytree.h"

namespace Crow {

// PolycellProxy ---------------------------------------------------------------

PolycellProxy::~PolycellProxy() {
	CHECK(!polycell);
}

PPolycellProxy PolycellProxy::create(PPolycell pc) {
	PPolycellProxy proxy= MakeRefPtr<PolycellProxy>();
	proxy->polycell= pc;
	return proxy;
}

PPolycell PolycellProxy::getPolycell() {
	return polycell;
}

void PolycellProxy::clear() {
	polycell= PPolycell();
}

// Polycell --------------------------------------------------------------------

Polycell::Polycell() {
	allowEditing= false;
	editing= false;
	autoStop= false;
	manualStop= false;
	column= -1;
	fuzzy= false;
}

Polycell::~Polycell() {
	CHECK(!element);
}

void Polycell::init(PPolyelem elem, int col) {
	element= elem;
	column= col;
}

void Polycell::done() {
	CHECK(!isEditing());
	element= PPolyelem();
	column= -1;
}

void Polycell::setFuzzy(bool fuzzy) {
	this->fuzzy= fuzzy;
	invalidate();
}

PPolyelem Polycell::getElement() {
	return element;
}

Gtk::TreeModel::Path Polycell::getPath() {
	return *element->getTreePath();
}

Polytree * Polycell::getTreeView() {
	return element->getTreeView();
}

void Polycell::startEditing() {
	CHECK(isEditable());
	if(!isEditing())
		getTreeView()->startEditing(getPath(), getModelColumn());
}

void Polycell::accept(bool stop) {
	bool ok= onAccept();
	if((autoStop && ok) || stop)
		stopEditing();
	if(ok)
		accepted();
}

void Polycell::reject(bool stop) {
	onReject();
	if(autoStop || stop)
		stopEditing();
}

void Polycell::start_editing_vfunc(GdkEvent *) {
	editingStarted();
}

void Polycell::on_remove_widget() {
	if(!manualStop) {
		onReject();
		editingStopped();
	} else
		manualStop= false;
}

void Polycell::stopEditing() {
	editingStopped();
	editing_done();
	manualStop= true;
	remove_widget();
}

void Polycell::editingStarted() {
	editing= true;
	getTreeView()->editingStarted(this);
	onEditingStarted();
}

void Polycell::editingStopped() {
	editing= false;
	getTreeView()->editingStopped(this);
	onEditingStopped();
}

PolycellWidget * Polycell::asWidget() {
	return dynamic_cast<PolycellWidget *>(this);
}

void Polycell::invalidate() {
	if(element)
		getTreeView()->invalidate(getPath(), getModelColumn());
}

// PolycellRenderer ------------------------------------------------------------

PolycellRenderer::PolycellRenderer(Polytree * tree, int mc) :
	Glib::ObjectBase(typeid(PolycellRenderer)),
	propertyPolycell(*this, "polycell")
{
	property_mode()= Gtk::CELL_RENDERER_MODE_EDITABLE;
	property_ypad()= 2;
	property_xpad()= 2;
	treeView= tree;
	column= mc;
	showIndents= false;

	depthIndent= 10;
	signSize= 9;
	signIndent= depthIndent + (signSize-1)/2 + 2;

}

bool PolycellRenderer::pointInIndent(int depth, const Rectangle & back, int x, int y) {
	int indent= depthIndent*depth;
	return x < back.x + indent;
}

bool PolycellRenderer::pointInSign(int depth, const Rectangle & back, int x, int y) {
	return Point(x, y).inside(getSignRect(back, depth));
}

void PolycellRenderer::get_size_vfunc(
	Gtk::Widget & widget,
	const Gdk::Rectangle * cell_area,
	int * x_offset, int * y_offset,
	int * width, int * height
) const {

	if(x_offset)
		*x_offset= 0;

	if(y_offset)
		*y_offset= 0;

	if(width || height) {

		PPangoLayout layout= treeView->create_pango_layout("");
		int textw, texth;
		layout->get_pixel_size(textw, texth);

		if(width)
			*width= 0;

		if(height) {
			*height= texth + property_ypad()*2 + 1;

			int vertical_separator;
			treeView->get_style_property("vertical-separator", vertical_separator);
			*height -= vertical_separator;

		}

	}

}

void PolycellRenderer::render_vfunc(
	const PGdkDrawable & drawable,
	Gtk::Widget & widget,
	const Gdk::Rectangle & background_area,
	const Gdk::Rectangle & cell_area,
	const Gdk::Rectangle & expose_area,
	Gtk::CellRendererState flags
) {

	PGdkWindow window= PGdkWindow::cast_dynamic(drawable);
	PGdkGC gc= Gdk::GC::create(window);
	Polycell * polycell= getCurrent();

	Rectangle clip= background_area;

	Rectangle cell= clip;
	cell.grow(
		property_xpad(), property_ypad(),
		-2*int(property_xpad())-1, -2*int(property_ypad())-1
	);

	int indent= 0;

	if(showIndents) {

		Gtk::TreeModel::Path path= polycell->getPath();
		int depth= path.size()-1;
		bool showsign= !(*treeView->get_model()->get_iter(path)).children().empty();
		bool expanded= treeView->row_expanded(path);

		if(depth > 0) {
			treeView->get_style()->paint_flat_box(
				window, Gtk::STATE_NORMAL, Gtk::SHADOW_NONE, clip, *treeView, "cell_even",
				clip.x, clip.y,
				depthIndent*depth, clip.h
			);
			gc->set_foreground(getGreyColor());
			window->draw_line(
				gc,
				clip.x + depthIndent*depth - 1, clip.y,
				clip.x + depthIndent*depth - 1, clip.y + clip.h - 1
			);
		}

		indent= depthIndent*depth;

		if(showsign)
			cell.grow(indent+signIndent, 0, -indent-signIndent, 0);
		else
			cell.grow(indent, 0, -indent, 0);

		if(showsign)
			drawSign(window, gc, expanded, depth, clip);

	}

	if(!cell.empty()) {
		if(!polycell->isFuzzy())
			polycell->cellRender(&widget, window, cell, flags);
		else
			polycell->fuzzyRender(&widget, window, cell, flags);
	}

	if(indent > 0) {
		Gtk::TreeModel::Path nextPath= polycell->getPath();
		bool found= findNextRowPath(&nextPath);
		if(found) {
			int nextIndent= (nextPath.size()-1)*depthIndent;
			indent= std::min(indent, nextIndent);
		}
	}

	int hlx= clip.x + indent - 1;
	int hll= clip.w - indent;

	gc->set_foreground(getGreyColor());
	if(hll > 0)
		window->draw_line(gc, hlx, clip.y + clip.h - 1, hlx + hll - 1, clip.y + clip.h - 1);
	window->draw_line(gc, clip.x + clip.w-1, clip.y, clip.x + clip.w-1, clip.y + clip.h-1);

}

Gtk::CellEditable * PolycellRenderer::start_editing_vfunc(
	GdkEvent * event,
 	Gtk::Widget & widget,
 	ustring_cr path,
	const Gdk::Rectangle & background_area,
	const Gdk::Rectangle & cell_area,
	Gtk::CellRendererState state
) {

	//std::cout << "Height: " << background_area.get_height() << std::endl;

	Polycell * polycell= getCurrent();

	if(!polycell->isEditable() || !treeView->isEditable())
		return NULL;

	Rectangle clip= background_area;

	Rectangle placement= clip;
	placement.grow(0, 0, -1, -1);

	PolycellWidget * polycellWidget= polycell->asWidget();

	polycellWidget->setPlacement(
		placement,
		treeView->get_hadjustment()->get_value(),
		treeView->get_vadjustment()->get_value()
	);
	polycellWidget->show();

	return polycellWidget;
}

Polycell * PolycellRenderer::getCurrent() {
	PPolycellProxy polycellProxy= property_polycell();
	Polycell * polycell= GetPtr(polycellProxy->getPolycell());
	return polycell;
}

bool PolycellRenderer::findNextRowPath(Gtk::TreePath * path) {

	Gtk::TreePath pp= *path;

	if(treeView->row_expanded(pp)) {

		pp.down();

		*path= pp;
		return true;

	} else {

		for(;;) {

			pp.next();

			GtkTreeIter iter;
			bool ok= gtk_tree_model_get_iter(treeView->get_model()->gobj(), &iter, pp.gobj());

			if(ok) {

				*path= pp;
				return true;

			} else {

				pp.up();

				if(pp.empty())
					return false;

			}

		}

	}

}

void PolycellRenderer::drawSign(
	PGdkWindow window,
	PGdkGC gc,
	bool expanded,
	int depth,
	const Rectangle & clip
) {

	int cx= clip.x + depthIndent*(depth+1) - 1;
	int cy= clip.y + clip.h/2;
	Rectangle rect= getSignRect(clip, depth);
	int shs= (signSize-1)/2;

	gc->set_foreground(getWhiteColor());
	window->draw_rectangle(gc, true, rect.x, rect.y, rect.w, rect.h);

	gc->set_foreground(getGreyColor());
	window->draw_rectangle(gc, false, rect.x, rect.y, rect.w-1, rect.h-1);

	gc->set_foreground(getBlackColor());
	window->draw_line(gc, cx - shs + 2, cy, cx + shs - 2, cy);

	if(expanded) {
		gc->set_foreground(getGreyColor());
		window->draw_line(gc, cx, cy + shs + 1, cx, clip.y + clip.h - 1);
	} else
		window->draw_line(gc, cx, cy - shs + 2, cx, cy + shs - 2);

}

Rectangle PolycellRenderer::getSignRect(const Rectangle & clip, int depth) {

	int cx= clip.x + depthIndent*(depth+1) - 1;
	int cy= clip.y + clip.h/2;

	Rectangle rect;

	int shs= (signSize-1)/2;

	rect.x= cx - shs;
	rect.w= signSize;
	rect.y= cy - shs;
	rect.h= signSize;

	return rect;
}

Gdk::Color PolycellRenderer::getBlackColor() {
	return GetBlack(treeView->get_style());
}

Gdk::Color PolycellRenderer::getWhiteColor() {
	return GetWhite(treeView->get_style());
}

Gdk::Color PolycellRenderer::getGreyColor() {
	return GetLight(treeView->get_style());
}

// Polyelem --------------------------------------------------------------------

PPolyelem Polyelem::create(const IdPath & path, PObject object) {
	return MakeRefPtr(new Polyelem(path, object));
}

PObject Polyelem::getObject() {
	return object;
}

void Polyelem::setObject(PObject obj) {
	object= obj;
}

Gtk::TreeRow Polyelem::getRow() {
	return *treeView->getModel()->get_iter(treePath);
}

PPolycell Polyelem::getValue(int column) {
	PPolycellProxy ppp;
	getRow().get_value(column, ppp);
	return ppp ? ppp->getPolycell() : PPolycell();
}

void Polyelem::setValue(int column, PPolycell polycell) {
	clearValue(column);
	getRow().set_value(column, PolycellProxy::create(polycell));
	polycell->init(MakeRefPtr(this), column);
}

void Polyelem::clearValue(int column) {
	PPolycellProxy ppp;
	getRow().get_value(column, ppp);
	if(ppp) {
		if(ppp->getPolycell()->isEditing())
			ppp->getPolycell()->reject(true);
		ppp->getPolycell()->done();
		ppp->clear();
		getRow().set_value(column, PPolycellProxy());
	}
}

void Polyelem::clear() {
	for(int i=0; i<treeView->getModel()->get_n_columns(); ++i)
		clearValue(i);
}

Polyelem::Polyelem(
	const IdPath & p,
	PObject obj
) {
	path= p;
	parent= GetParentPath(path);
	object= obj;
	operation= opInsert;
	treeView= NULL;
}

// Polytree --------------------------------------------------------------------

Polytree::Polytree() {
	editable= true;
	immediateEditing= true;
	lock= false;
	editingPolycell= NULL;
	keepEditing= -1;
	property_show_expanders()= false;
	get_selection()->signal_changed().connect(sigc::mem_fun(*this, &Polytree::onSelectionChanged));
}

PGtkTreeStore Polytree::getModel() {
	return PGtkTreeStore::cast_dynamic(get_model());
}

void Polytree::createModel(int ncol) {
	std::vector<PolycellColumn> pcol(ncol);
	Gtk::TreeModel::ColumnRecord columns;
	for(int i=0; i<ncol; ++i)
		columns.add(pcol[i]);
	set_model(Gtk::TreeStore::create(columns));
}

void Polytree::addColumn(int mcol, ustring_cr name) {
	PolycellRenderer * renderer= Gtk::manage(new PolycellRenderer(this, mcol));
	Gtk::TreeViewColumn * column= Gtk::manage(new Gtk::TreeViewColumn(name));
	column->set_resizable(true);
	append_column(*column);
	column->pack_start(*renderer);
	column->add_attribute(*renderer, "polycell", mcol);
	if(get_columns().size() == 1)
		renderer->setShowIndents(true);
}

int Polytree::getColumnWidth(int ncol) {
	return get_column(ncol)->get_width();
}

void Polytree::setColumnWidth(int ncol, int width) {
	SetColumnWidth(get_column(ncol), width);
}

Gtk::TreeViewColumn * Polytree::findViewColumn(int modelColumn) {
	std::vector<Gtk::TreeViewColumn *> columns= get_columns();
	for(int i=0; i<SIGNED(columns.size()); ++i) {
		PolycellRenderer * renderer= dynamic_cast<PolycellRenderer *>(columns[i]->get_first_cell_renderer());
		if(modelColumn == renderer->getModelColumn())
			return columns[i];
	}
	return NULL;
}

void Polytree::prepare() {
	for(Elements::iterator it= elements.begin(); it!=elements.end(); ++it)
		it->second->setOperation(Polyelem::opRemove);

	CHECK(keepSelected.empty());
	keepSelected.reserve(selection.size());
	for(int i=0; i<SIGNED(selection.size()); ++i)
		keepSelected.push_back(*findElement(selection[i])->getPath());

	if(isEditing()) {
		CHECK(keepEditing==-1 && keepSelected.size()==1);
		Polycell * polycell= getEditingPolycell();
		CHECK(keepSelected[0] == *polycell->getElement()->getPath());
		keepEditing= polycell->getModelColumn();
	}

}

void Polytree::update(const IdPath & path, PObject object) {
	Elements::iterator it= elements.find(path);
	if(it != elements.end()) {
		PPolyelem E= it->second;
		CHECK(*E->getPath()==path);
		if(E->getObject() == object)
			E->setOperation(Polyelem::opUpdate);
		else {
			E->setOperation(Polyelem::opReplace);
			E->setObject(object);
		}
	} else
		elements[path]= Polyelem::create(path, object);
}

void Polytree::submit() {

	//dump("Polytree::submit()");
	CHECK(!lock);
	lock= true;
	Elements::iterator result= process(getModel()->children(), elements.begin(), IdPath());
	lock= false;
	CHECK(result == elements.end());

	for(Elements::iterator it= elements.begin(); it!=elements.end(); )
		if(it->second->getOperation()==Polyelem::opRemove)
			it= EraseAndNext(elements, it);
		else
			++it;

	Selection newsel;
	newsel.reserve(keepSelected.size());
	for(int i=0; i<SIGNED(keepSelected.size()); ++i) {
		PPolyelem elem= findElement(keepSelected[i]);
		if(elem)
			newsel.push_back(elem->getObject());
	}

	select(newsel, false, true);

	if(keepEditing != -1) {
		CHECK(keepSelected.size()==1);
		PPolyelem elem= findElement(keepSelected[0]);
		if(elem) {
			PPolycell polycell= elem->getValue(keepEditing);
			CHECK(polycell && polycell->getElement() == elem);
			if(polycell->isEditable())
				polycell->startEditing();
		}
		keepEditing= -1;
	}

	keepSelected.clear();

	showSelection();

	if(isEditing() && !getEditingPolycell()->isEditable())
		rejectEditing();

	signalSubmited.emit();
}

void Polytree::setSelection(const Selection & osel, bool fire) {
	select(osel, fire, false);
	showSelection();
}

std::vector<Gtk::TreePath> Polytree::getTreePathSelection() {
	std::vector<Gtk::TreePath> psel;
	psel.reserve(selection.size());
	for(int i=0; i<SIGNED(selection.size()); ++i)
		psel.push_back(*findElement(selection[i])->getTreePath());
	return psel;
}

void Polytree::setTreePathSelection(const std::vector<Gtk::TreePath> & psel, bool fire) {
	std::vector<PObject> osel;
	osel.reserve(psel.size());
	for(int i=0; i<SIGNED(psel.size()); ++i)
		osel.push_back(findElement(psel[i])->getObject());
	setSelection(osel, fire);
}

PObject Polytree::saveState() {

	PPolystate state= MakeRefPtr<Polystate>();

	for(Elements::iterator it= elements.begin(); it!=elements.end(); ++it)
		if(row_expanded(*it->second->getTreePath()))
			state->expanded.push_back(*it->second->getPath());

	Gtk::TreeViewColumn * column;
	int cell_x, cell_y;
	Gtk::TreePath path;
	bool ok= get_bin_window() && get_path_at_pos(1, 1, path, column, cell_x, cell_y);
	if(ok)
		state->scroll= *findElement(path)->getPath();

	state->selected.reserve(selection.size());
	for(int i=0; i<SIGNED(selection.size()); ++i)
		state->selected.push_back(*findElement(selection[i])->getPath());

	return state;
}

void Polytree::loadState(PObject obj) {
	PPolystate state= PPolystate::cast_dynamic(obj);

	Polycell * wasEditing= getEditingPolycell();

	for(std::list<IdPath>::iterator it= state->expanded.begin(); it!=state->expanded.end(); ++it) {
		PPolyelem elem= findElement(*it);
		if(elem)
			expand_to_path(*elem->getTreePath());
	}

	if(!state->scroll.empty()) {
		PPolyelem elem= findElement(state->scroll);
		if(elem)
			scroll_to_row(*elem->getTreePath(), 0);
	}

	if(wasEditing) {
		wasEditing->startEditing();
		showSelection();
	}

	if(getSelection()->empty() && !state->selected.empty()) {
		Selection newsel;
		newsel.reserve(state->selected.size());
		for(int i=0; i<SIGNED(state->selected.size()); ++i) {
			PPolyelem elem= findElement(state->selected[i]);
			if(elem)
				newsel.push_back(elem->getObject());
		}
		select(newsel, true);
	}

}

void Polytree::expandRow(PObject obj) {
	PPolyelem elem= findElement(obj);
	CHECK(elem);
	expand_row(*elem->getTreePath(), false);
}

void Polytree::expandAll() {
	for(Elements::iterator it= elements.begin(); it!=elements.end(); ++it)
		expand_row(*it->second->getTreePath(), false);
}

PObject Polytree::getByTreePath(const Gtk::TreePath & path) {
	PPolyelem elem= findElement(path);
	CHECK(elem);
	return elem->getObject();
}

void Polytree::startEditing(const Gtk::TreePath & path, int modelColumn) {
	Gtk::TreeViewColumn * viewColumn= findViewColumn(modelColumn);
	set_cursor(path, *viewColumn, true);
}

void Polytree::editingStarted(Polycell * polycell) {
	CHECK(editingPolycell==NULL);
	setTreePathSelection(std::vector<Gtk::TreePath>(1, polycell->getPath()), true);
	editingPolycell= polycell;
}

void Polytree::editingStopped(Polycell * polycell) {
	CHECK(editingPolycell==polycell);
	editingPolycell= NULL;
}

void Polytree::invalidate(const Gtk::TreePath & path, int modelColumn) {
	Gtk::TreeViewColumn * viewColumn= findViewColumn(modelColumn);
	Gdk::Rectangle rect;
	get_background_area(path, *viewColumn, rect);
	if(get_bin_window())
		get_bin_window()->invalidate_rect(rect, false);
}

bool Polytree::on_button_press_event(GdkEventButton * event) {

	if(event->window == get_bin_window()->gobj()) {

		int x= event->x;
		int y= event->y;
		int cell_x, cell_y;

		Gtk::TreeModel::Path path;
		Gtk::TreeViewColumn * column;

		bool ok= get_path_at_pos(x, y, path, column, cell_x, cell_y);
		if(ok) {

			PolycellRenderer * renderer= dynamic_cast<PolycellRenderer *>(column->get_first_cell_renderer());
			int mc= renderer->getModelColumn();
			PPolycell polycell= findElement(path)->getValue(mc);

			if(renderer->getShowIndents()) {

				Gdk::Rectangle back;
				get_background_area(path, *column, back);

				int X= cell_x + back.get_x();
				int Y= cell_y + back.get_y();

				if(renderer->pointInIndent(path.size()-1, back, X, Y))
					return true;

				if(renderer->pointInSign(path.size()-1, back, X, Y)) {
					bool expanded= row_expanded(path);
					if(expanded)
						collapse_row(path);
					else
						expand_row(path, false);
					return true;
				}

			}

			if(event->button == 1) {

				if(isEditing())
					acceptEditing();

				if(getImmediateEditing() && polycell->isEditable()) {
					set_cursor(path, *column, true);
				} else {
					return Gtk::TreeView::on_button_press_event(event);
				}

			} else if(event->button == 3) {

				if(!get_selection()->is_selected(path))
					setTreePathSelection(std::vector<Gtk::TreePath>(1, path), true);
				signalContextMenu.emit();

			}

			return true;

		}

	}

	return Gtk::TreeView::on_button_press_event(event);
}

bool Polytree::on_button_release_event(GdkEventButton * event) {
	return Gtk::TreeView::on_button_release_event(event);
}

void Polytree::on_size_allocate(Gtk::Allocation & allocation) {
	if(
		allocation.get_width() != get_allocation().get_width() ||
		allocation.get_height() != get_allocation().get_height()
	) {
		if(isEditing())
			rejectEditing();
	}
	Gtk::TreeView::on_size_allocate(allocation);
}

void Polytree::on_row_expanded(const Gtk::TreeModel::iterator & iter, const Gtk::TreeModel::Path & path) {
	if(isEditing() && getEditingPath() > path)
		rejectEditing();
	Gtk::TreeView::on_row_expanded(iter, path);
}

void Polytree::on_row_collapsed(const Gtk::TreeModel::iterator & iter, const Gtk::TreeModel::Path & path) {
	if(isEditing() && getEditingPath() > path)
		rejectEditing();
	Gtk::TreeView::on_row_collapsed(iter, path);
}

void Polytree::onSelectionChanged() {
	if(!lock) {
		std::vector<Gtk::TreePath> psel= get_selection()->get_selected_rows();
		Selection newsel;
		newsel.reserve(psel.size());
		for(int i=0; i<SIGNED(psel.size()); ++i) {
			PPolyelem elem= findElement(psel[i]);
			newsel.push_back(elem->getObject());
		}
		std::sort(newsel.begin(), newsel.end());
		if(newsel != selection) {
			selection= newsel;
			signalSelectionChanged.emit();
		}
	}
}

Polytree::Elements::iterator Polytree::process(
	const Gtk::TreeNodeChildren & tnc,
	Elements::iterator element,
	const IdPath & parent
) {

	bool wasEmpty= tnc.empty();

	Gtk::TreeModel::iterator rowit= tnc.begin();
	Elements::iterator it= element;

	while(it!=elements.end() && *it->second->getParent()==parent) {

		PPolyelem E= it->second;

		++it;

		switch(E->getOperation()) {

			case Polyelem::opInsert:
				rowit= getModel()->insert(rowit);
				initElem(E, rowit);
				setRowValues(E);
				it= process(rowit->children(), it, *E->getPath());
				++rowit;
				break;

			case Polyelem::opReplace:
				collapse_row(get_model()->get_path(rowit));
				initElem(E, rowit);
				setRowValues(E);
				it= process(rowit->children(), it, *E->getPath());
				++rowit;
				break;

			case Polyelem::opUpdate:
				initElem(E, rowit);
				it= process(rowit->children(), it, *E->getPath());
				++rowit;
				break;

			case Polyelem::opRemove:
				initElem(E, rowit);
				it= process(rowit->children(), it, *E->getPath());
				E->clear();
				rowit= getModel()->erase(rowit);
				break;

			default:
				CHECK(false);

		}

	}

	if(wasEmpty && tnc.empty())
		CHECK(element==it);

	return it;
}

void Polytree::initElem(PPolyelem E, const Gtk::TreeIter & rowit) {
	E->setTreeView(this);
	E->setTreePath(Gtk::TreePath(rowit));
}

void Polytree::acceptEditing() {
	CHECK(isEditing());
	editingPolycell->accept(true);
}

void Polytree::rejectEditing() {
	CHECK(isEditing());
	editingPolycell->reject(true);
}

Gtk::TreePath Polytree::getEditingPath() {
	CHECK(isEditing());
	return editingPolycell->getPath();
}

void Polytree::select(const Selection & osel, bool fire, bool force) {
	std::vector<Gtk::TreePath> psel;
	psel.reserve(osel.size());
	Selection newsel;
	newsel.reserve(osel.size());
	for(int i=0; i<SIGNED(osel.size()); ++i) {
		PPolyelem elem= findElement(osel[i]);
		CHECK(elem);
		newsel.push_back(osel[i]);
		psel.push_back(*elem->getTreePath());
	}

	std::sort(newsel.begin(), newsel.end());
	bool changed= (newsel != selection) || force;
	if(changed) {
		selection= newsel;
		selectPathes(psel);

		if(isEditing()) {
			PPolyelem elem= getEditingPolycell()->getElement();
			if(selection.size()!=1 || selection[0] != elem->getObject())
				rejectEditing();
		}

	}

	if(changed && fire)
		signalSelectionChanged.emit();
}

void Polytree::selectPathes(const std::vector<Gtk::TreePath> & psel) {
	CHECK(!lock);
	lock= true;
	get_selection()->unselect_all();
	if(!psel.empty())
		for(int i=0; i<SIGNED(psel.size()); ++i) {
			expandToPath(psel[i]);
			get_selection()->select(psel[i]);
		}
	lock= false;
}

void Polytree::expandToPath(const Gtk::TreePath & path) {
	Gtk::TreePath expand;
	for(int i=0; i<SIGNED(path.size())-1; ++i) {
		expand.push_back(path[i]);
		expand_row(expand, false);
	}
}

bool Polytree::rowVisible(const Gtk::TreePath & path) {
	Gdk::Rectangle vis, row;
	get_visible_rect(vis);
	int count= get_columns().size();
	CHECK(count > 1);
	get_background_area(path, *get_column(1), row);
	for(int i=2; i<count; ++i) {
		Gdk::Rectangle cell;
		get_background_area(path, *get_column(i), cell);
		row.join(cell);
	}
	bool intersects;
	vis.intersect(row, intersects);
	return intersects;
}

void Polytree::showSelection() {
	if(!is_realized())
		return;
	std::vector<Gtk::TreePath> psel= getTreePathSelection();
	if(psel.empty()) {
		if(!getModel()->children().empty()) {
			Gtk::TreeRow row= getModel()->children()[0];
			scroll_to_row(getModel()->get_path(row));
		}
	} else {
		for(int i=0; i<SIGNED(psel.size()); ++i)
			if(rowVisible(psel[i]))
				return;
		Gtk::TreePath min= psel.front();
		for(int i=1; i<SIGNED(psel.size()); ++i)
			if(psel[i] < min)
				min= psel[i];
		scroll_to_row(min);
	}
}

PPolyelem Polytree::findElement(const Gtk::TreePath & treePath) {
	for(Elements::iterator it= elements.begin(); it!=elements.end(); ++it)
		if(*it->second->getTreePath()==treePath)
			return it->second;
	return PPolyelem();
}

PPolyelem Polytree::findElement(PObject object) {
	for(Elements::iterator it= elements.begin(); it!=elements.end(); ++it)
		if(it->second->getObject() == object)
			return it->second;
	return PPolyelem();
}

PPolyelem Polytree::findElement(const IdPath & path) {
	Elements::iterator it= elements.find(path);
	if(it!=elements.end()) {
		CHECK(*it->second->getPath()==path);
		return it->second;
	}
	return PPolyelem();
}

void Polytree::dump(string_cr str) {
	std::cout << "--------------------" << str << std::endl;
	for(Elements::iterator it= elements.begin(); it!=elements.end(); ++it) {
		IdPath path= it->first;
		PPolyelem elem= it->second;
		CHECK(path==*elem->getPath());
		std::cout << IdPathToString(path) << std::endl;
	}
}

// PolycellText ----------------------------------------------------------------

CustomAlignment::CustomAlignment() :
	propertyEditingCanceled(*this, "editing-canceled", false)
{
}

PolycellText::PolycellText() :
	Glib::ObjectBase(typeid(PolycellText))
{
	ellipsize= Pango::ELLIPSIZE_END;
}

void PolycellText::setText(ustring_cr t) {
	text= t;
	setFuzzy(false);
}

PPangoLayout PolycellText::createLayout(ustring_cr str, Gtk::Widget * widget, int width) {
	PPangoLayout layout= widget->create_pango_layout(str);
	layout->set_ellipsize(ellipsize);
	layout->set_single_paragraph_mode(true);
	layout->set_width(width==-1 ? -1 : width*Pango::SCALE);
	return layout;
}

PPangoLayout PolycellText::getLayout(Gtk::Widget * widget, int width) {
	PPangoLayout layout= createLayout(text, widget, width);
	if(attributes.get_iter())
		layout->set_attributes(attributes);
	return layout;
}

void PolycellText::textRender(
	PPangoLayout layout,
	Gtk::Widget * widget,
	PGdkWindow window,
	const Gdk::Rectangle & rect,
	Gtk::CellRendererState flags
) {

	Gtk::StateType state;

	if(flags & Gtk::CELL_RENDERER_SELECTED) {

		if(widget->has_focus())
			state = Gtk::STATE_SELECTED;
		else
			state = Gtk::STATE_ACTIVE;

	} else if(
		(flags & Gtk::CELL_RENDERER_PRELIT) &&
		widget->get_state() == Gtk::STATE_PRELIGHT
	) {

		state = Gtk::STATE_PRELIGHT;

	} else {

		if(widget->get_state() == Gtk::STATE_INSENSITIVE)
			state = Gtk::STATE_INSENSITIVE;
		else
			state = Gtk::STATE_NORMAL;

	}

	widget->get_style()->paint_layout(
		window, state, true, rect, *widget, "polycellrenderer",
		rect.get_x(), rect.get_y(),
		layout
	);

}

void PolycellText::cellRender(
	Gtk::Widget * widget,
	PGdkWindow window,
	const Gdk::Rectangle & rect,
	Gtk::CellRendererState flags
) {
	PPangoLayout layout= getLayout(widget, rect.get_width());
	textRender(layout, widget, window, rect, flags);
}

void PolycellText::fuzzyRender(
	Gtk::Widget * widget,
	PGdkWindow window,
	const Gdk::Rectangle & rect,
	Gtk::CellRendererState flags
) {
	PPangoLayout layout= createLayout("...", widget, rect.get_width());
	Pango::AttrList attributes;
	ChangeAttribute(&attributes, Pango::Attribute::create_attr_weight(Pango::WEIGHT_BOLD));
	layout->set_attributes(attributes);
	layout->set_alignment(Pango::ALIGN_CENTER);
	textRender(layout, widget, window, rect, flags);
}

// PolycellWidget --------------------------------------------------------------

PolycellWidget::PolycellWidget() {
	hShift= vShift= 0;
}

void PolycellWidget::setPlacement(const Rectangle & rect, int hshift, int vshift) {
	placement= rect;
	hShift= hshift;
	vShift= vshift;
	set_size_request(rect.w, rect.h);
}

void PolycellWidget::on_size_allocate(Gtk::Allocation & allocation) {
	Gdk::Rectangle rect= placement;
	int hShiftNew= getElement()->getTreeView()->get_hadjustment()->get_value();
	int vShiftNew= getElement()->getTreeView()->get_vadjustment()->get_value();
	int hdiff= hShiftNew - hShift;
	int vdiff= vShiftNew - vShift;
	rect.set_x(rect.get_x() - hdiff);
	rect.set_y(rect.get_y() - vdiff);
	Gtk::Alignment::on_size_allocate(rect);
}

// PolycellInput ---------------------------------------------------------------

PolycellInput::PolycellInput() :
	Glib::ObjectBase(typeid(PolycellInput))
{

	validator= sigc::mem_fun(*this, &PolycellInput::validatorDefault);

	add(hBox);
	hBox.show();

	hBox.pack_start(entry, Gtk::PACK_EXPAND_WIDGET);
	entry.set_has_frame(false);
	entry.signal_activate().connect(sigc::mem_fun(*this, &PolycellInput::onEntryActivate));
	entry.signal_key_press_event().connect(sigc::mem_fun(*this, &PolycellInput::onEntryKeyPressEvent));
	entry.show();

	Gtk::Arrow * arrow= Gtk::manage(new Gtk::Arrow(Gtk::ARROW_LEFT, Gtk::SHADOW_NONE));
	button.add(*arrow);
	arrow->show();

	button.set_focus_on_click(false);
	hBox.pack_start(button, Gtk::PACK_SHRINK);
	button.signal_clicked().connect(sigc::mem_fun(*this, &PolycellInput::onButtonClicked));

}

void PolycellInput::setText(ustring_cr txt) {
	PolycellWidget::setText(txt);
	if(isEditing()) {
		entry.set_text(*getText());
		if(entry.get_editable())
			entry.select_region(0, -1);
	}
}

void PolycellInput::setEntryEditable(bool ee) {
	entry.set_editable(ee);
	entry.property_can_focus()= ee;
	button.set_focus_on_click(!ee);
}

Glib::ustring PolycellInput::getEntryText() {
	return entry.get_text();
}

bool PolycellInput::isPopupEnabled() {
	return button.get_visible();
}

void PolycellInput::setPopupEnabled(bool enabled) {
	button.property_visible()= enabled;
}

void PolycellInput::setPlacement(const Rectangle & rect, int hshift, int vshift) {
	PolycellWidget::setPlacement(rect, hshift, vshift);
	entry.set_size_request(-1, rect.h);
}

void PolycellInput::onEditingStarted() {
	entry.set_text(*getText());
	if(entry.get_editable()) {
		entry.select_region(0, -1);
		entry.grab_focus();
	} else
		button.grab_focus();
}

bool PolycellInput::onAccept() {
	ustring_cr oldtext= *PolycellText::getText();
	Glib::ustring newtext= entry.get_text();
	if(validator(oldtext, newtext)) {
		PolycellWidget::setText(newtext);
		return true;
	} else
		return false;
}

void PolycellInput::onReject() {
	entry.set_text(*getText());
}

void PolycellInput::onEntryActivate() {
	accept();
}

bool PolycellInput::onEntryKeyPressEvent(GdkEventKey * event) {
	if(event->keyval == GDK_Escape) {
		reject();
		return true;
	}
	return false;
}

// PolycellChoice --------------------------------------------------------------

PolycellChoice::PolycellChoice() :
	Glib::ObjectBase(typeid(PolycellChoice))
{
	lock= false;
	choice= -1;
	combo.property_active().signal_changed().connect(sigc::mem_fun(*this, &PolycellChoice::onComboChanged));
	add(combo);
	combo.show();
}

void PolycellChoice::setText(ustring_cr txt) {
	std::vector<Glib::ustring>::iterator it= std::find(choices.begin(), choices.end(), txt);
	CHECK(it!=choices.end());
	setChoice(checked_cast<int>(it - choices.begin()));
}

void PolycellChoice::setChoice(int c) {
	choice= c;
	if(isEditing()) {
		CHECK(!lock);
		lock= true;
		combo.set_active(choice);
		lock= false;
	}
	PolycellWidget::setText(choices[choice]);
}

void PolycellChoice::addChoice(ustring_cr text) {
	combo.append_text(text);
	choices.push_back(text);
}

void PolycellChoice::onEditingStarted() {
	CHECK(!lock);
	lock= true;
	combo.set_active(choice);
	lock= false;
	combo.grab_focus();
}

bool PolycellChoice::onAccept() {
	choice= combo.get_active_row_number();
	PolycellWidget::setText(choices[choice]);
	return true;
}

void PolycellChoice::onComboChanged() {
	if(!lock)
		accept();
}

}
