/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "forward.h"
#include "util.h"

#include "palette.h"
#include "crow_impl.h"
#include "model.h"

namespace Crow {

void CheckFailed(const char * expr, const char * file, int line) {
	std::ostringstream oss;
	oss << boost::format(_("Check failed: %s, file: %s, line: %d")) % expr % file % line;
	FatalError(oss.str());
}

void FatalError(string_cr message) {
	g_critical(message.c_str());
	#ifdef CROW_SANDBOX
		assert(0);
	#endif
	Gtk::MessageDialog dialog(
		_("Fatal Error") + std::string(" [" G_LOG_DOMAIN "]"), false,
		Gtk::MESSAGE_ERROR, Gtk::BUTTONS_CLOSE, true
	);
	dialog.set_secondary_text(message);
	dialog.set_keep_above(true);
	dialog.run();
	abort();
}

std::string GetTypeName(const std::type_info & ti) {
	std::string result;
	#ifdef __GNUC__
		int status;
		char * buf= abi::__cxa_demangle(ti.name(), NULL, NULL, &status);
		result= buf;
		free(buf);
	#else
		result= ti.name();
	#endif
	return result;
}

TypeId FindType(const std::string typeName) {
	return Palette::get()->findType(typeName);
}

PaletteEntry * GetEntry(TypeId type) {
	return Palette::get()->getEntry(type);
}

void DumpParents(const char * name) {
	GType type= g_type_from_name(name);
	GType parent= g_type_parent(type);
	if(parent!=0) {
		const char * pname= g_type_name(parent);
		std::cout << pname << std::endl;
		DumpParents(pname);
	}
}

void NoOp() {
}

IdPath GetParentPath(const IdPath & path) {
	CHECK(!path.empty());
	return IdPath(path.begin(), path.end()-1);
}

std::string GetData(Glib::Object * object, string_cr key) {
	char * data= static_cast<char *>(object->get_data(Glib::QueryQuark(key)));
	return data ? data : "";
}

void SetData(Glib::Object * object, string_cr key, string_cr value) {
	char * str= g_strdup(value.c_str());
	object->set_data(Glib::Quark(key), str, &g_free);
}

GType GetGType(Glib::Object * object) {
	if(object==NULL)
		return G_TYPE_INVALID;
	GType gtype= G_OBJECT_TYPE(object->gobj());
	std::string name= g_type_name(gtype);
	if(Begins(name, std::string("gtkmm__")))
		return g_type_parent(gtype);
	else
		return gtype;
}

Point Transform(const Point & p, Gtk::Widget * s, Gtk::Widget * d) {
	CHECK(s && d);

	if(!s->is_realized() || !d->is_realized()) {
		g_warning(_("bad Crow::Transform() argument: unrealized widget"));
		if(!s->is_realized())
			gtk_widget_realize(s->gobj());
		if(!d->is_realized())
			gtk_widget_realize(d->gobj());
	}

	CHECK(s->is_realized() && d->is_realized());

	Point S= Point(s->get_allocation().get_x(), s->get_allocation().get_y());
	Point D= Point(d->get_allocation().get_x(), d->get_allocation().get_y());

	Point SW, DW;
	if(!s->has_no_window() && s->get_parent())
		s->get_window()->get_position(SW.x, SW.y);
	if(!d->has_no_window() && d->get_parent())
		d->get_window()->get_position(DW.x, DW.y);

	Point S0, D0;
	s->get_window()->get_origin(S0.x, S0.y);
	d->get_window()->get_origin(D0.x, D0.y);

	return p + S - SW + S0 - D0 + DW - D;
}

void SplitNumber(string_cr name, std::string * str, std::string * num) {
	num->clear();
	int i;
	for(i= name.size()-1; i>=0; --i)
		if(isdigit(name[i]))
			*num += name[i];
		else
			break;
	while(num->size() > 1)
		if((*num)[num->size()-1]=='0' || num->size() > 9) {
			num->resize(num->size()-1);
			++i;
		} else
			break;
	*str= name.substr(0, i+1);
	std::reverse(num->begin(), num->end());
}

void SplitNumber(string_cr name, std::string * str, int * num) {
	std::string str_num;
	SplitNumber(name, str, &str_num);
	*num= str_num.empty() ? -1 : FromString<int>(str_num);
}

bool LessSplit(string_cr name1, string_cr name2) {
	std::string str1, str2;
	int num1, num2;
	SplitNumber(name1, &str1, &num1);
	SplitNumber(name2, &str2, &num2);
	return str1==str2 ? num1 < num2 : str1 < str2;
}

std::string RightPart(string_cr str, char ch) {
	int pos= str.rfind(ch, str.length()-1);
	std::string name;
	if(pos == SIGNED(std::string::npos))
		return str;
	else
		return str.substr(pos+1, str.length()-pos-1);
}

StringList Split(string_cr str, const char * dlm) {
	StringList result;
	typedef boost::tokenizer<boost::char_separator<char> > Tok;
	boost::char_separator<char> sep(dlm, "", boost::keep_empty_tokens);
	Tok tok(str, sep);
	for(Tok::iterator it= tok.begin(); it!=tok.end(); ++it)
		result.push_back(*it);
	return result;
}

Glib::ustring XmlEncode(ustring_cr str, bool attr) {
	std::ostringstream result;
	for(Glib::ustring::const_iterator it= str.begin(); it!=str.end(); ++it) {
		gunichar ch= *it;
		if(Glib::Unicode::isprint(ch)) {
			switch(ch) {
				case '<':
					result << "&lt;";
					continue;
				case '>':
					result << "&gt;";
					continue;
				case '&':
					result << "&amp;";
					continue;
				case '"':
					if(!attr)
						break;
					result << "&quot;";
					continue;
				case '\'':
					if(!attr)
						break;
					result << "&apos;";
					continue;
			}
			result << Glib::ustring(1, ch);
		} else {
			std::ostringstream oss;
			oss << "&#x" << std::hex << checked_cast<guint32>(ch) << ";";
			result << oss.str();
		}
	}
	return result.str();
}

bool ValidName(string_cr name) {
	if(name.empty())
		return false;
	if(name[0] != '_' && !isalpha(name[0]))
		return false;
	for(int i=1; i<SIGNED(name.size()); ++i)
		if(!isalpha(name[i]) && !isdigit(name[i]) && name[i] != '_')
			return false;
	return true;
}

void DrawGrid(PGdkWindow window, PGtkStyle style, bool squares) {
	PGdkGC gc= Gdk::GC::create(window);
	int W, H;
	window->get_size(W, H);
	gc->set_foreground(GetDark(style));
	window->draw_rectangle(gc, false, 1, 1, W-3, H-3);
	if(squares)
		gc->set_foreground(GetLight(style));
	int GS= 9;
	int GW= W-4;
	int GH= H-4;
	int xx= GW/GS;
	int yy= GH/GS;
	int GX= 2 + (GW-xx*GS)/2;
	int GY= 2 + (GH-yy*GS)/2;
	if(!squares) {
		for(int x=0; x<xx; ++x)
			for(int y=0; y<yy; ++y)
				window->draw_point(gc, GX+x*GS+GS/2, GY+y*GS+GS/2);
	} else {
		for(int x=0; x<xx-1; ++x)
			for(int y=0; y<yy-1; ++y)
				if((x+y) % 2 == 0)
					window->draw_rectangle(gc, true, GX+x*GS+GS/2, GY+y*GS+GS/2, GS, GS);
	}
}

void RaiseError(ustring_cr msg) {
	throw RuntimeError(msg);
}

int GetRefCount(const Glib::ObjectBase * object) {
	return object->gobj()->ref_count;
}

void ContainerClear(Gtk::Container * container) {
	std::vector<Gtk::Widget *> widgets= container->get_children();
	for(int i=0; i<SIGNED(widgets.size()); ++i)
		ContainerRemove(container, widgets[i]);
}

void ContainerRemove(Gtk::Container * container, Gtk::Widget * widget) {
	gtk_container_remove(container->gobj(), widget->gobj());
}

void BinRemove(Gtk::Bin * bin) {
	Gtk::Widget * widget= bin->get_child();
	if(widget)
		gtk_container_remove(bin->Gtk::Container::gobj(), widget->gobj());
}

NodeSet ToNodeSet(const Nodes & nodes) {
	NodeSet result(nodes.begin(), nodes.end());
	CHECK(result.size()==nodes.size());
	return result;
}

Gdk::Color GetWhite(PGtkStyle style) {
	return style->get_light(Gtk::STATE_ACTIVE);
}

Gdk::Color GetBlack(PGtkStyle style) {
	return style->get_dark(Gtk::STATE_SELECTED);
}

Gdk::Color GetLight(PGtkStyle style) {
	return style->get_mid(Gtk::STATE_NORMAL);
}

Gdk::Color GetDark(PGtkStyle style) {
	return style->get_dark(Gtk::STATE_NORMAL);
}

std::string IdPathToString(const IdPath & path, bool dots) {
	std::string str;
	if(!dots)
		str += '/';
	for(int i=0; i<SIGNED(path.size()); ++i) {
		str += path[i].toString();
		if(i < SIGNED(path.size())-1)
			str += dots ? '.' : '/';
	}
	return str;
}

void SortByName(Nodes * nodes) {
	struct tmp { static bool less(PNode a, PNode b) {
		return LessSplit(*a->getName(), *b->getName());
	}};
	nodes->sort(sigc::ptr_fun(&tmp::less));
}

void SetColumnWidth(Gtk::TreeViewColumn * column, int width) {
	column->set_sizing(Gtk::TREE_VIEW_COLUMN_FIXED);
	column->set_fixed_width(width);
	column->set_resizable(true);
}

Glib::ustring Trim(ustring_cr str) {
	const char * spaces= " \t\r\n";
	int p0= str.find_first_not_of(spaces);
	if(p0 != SIGNED(Glib::ustring::npos)) {
		int p1= str.find_last_not_of(spaces);
		return str.substr(p0, p1-p0+1);
	}
	return Glib::ustring();
}

bool Equal(const PObject & v1, const PObject & v2) {
	return (v1 && v2) ? v1->equals(v2) : (!v1 && !v2);
}

GQuark ReferencedQuark() {
	static GQuark referenced_quark= 0;
	if(referenced_quark == 0)
		referenced_quark= g_quark_from_string("CrowReferenced");
	return referenced_quark;
}

void ReferenceGlibObject(Glib::Object * obj) {
	CHECK(obj);
	GQuark rq= ReferencedQuark();
	if(GPOINTER_TO_INT(g_object_get_qdata(obj->gobj(), rq))) {
		obj->reference();
	} else {
		g_object_set_qdata(obj->gobj(), rq, GINT_TO_POINTER(TRUE));
		Gtk::Object * gtkobj= dynamic_cast<Gtk::Object *>(obj);
		if(gtkobj) {
			Gtk::Window * gtkwin= dynamic_cast<Gtk::Window *>(gtkobj);
			if(!gtkwin) {
				gtkobj->set_manage();
				g_object_ref_sink(obj->gobj());
			} else {
				obj->reference();
			}
		}
	}
}

void ReleaseWindow(Gtk::Window * window) {
	if(GetRefCount(window) == 1)
		delete window;
	else
		window->unreference();
}

int TypeDist(GType sub, GType type) {
	CHECK(g_type_is_a(sub, type));
	int dist= 0;
	GType gt= sub;
	for(;;) {
		gt= g_type_parent(gt);
		if(gt == G_TYPE_INVALID || !g_type_is_a(gt, type))
			return dist;
		++dist;
	}
}

Glib::ustring Substitute(ustring_cr str, ustring_cr key, ustring_cr val) {
	Glib::ustring R= str;
	int p= 0;
	while(p < SIGNED(R.size())) {
		p= R.find(key, p);
		if(p == SIGNED(Glib::ustring::npos))
			break;
		R.replace(p, key.size(), val);
		p += val.size();
	}
	return R;
}

Glib::ustring Underscores(ustring_cr str) {
	return Substitute(str, "_", "__");
}

Bits Addition(Bits * v, Bits f) {
	*v |= f;
	return *v;
}

Bits Subtraction(Bits * v, Bits f) {
	*v= *v & ~f;
	return *v;
}

bool Includes(Bits v, Bits f) {
	return (v & f) == f;
}

bool Intersect(Bits v, Bits f) {
	return (v & f) != 0;
}

gint64 Round(double v) {
	return (gint64)floor(v+0.5);
}

double GetSliderValue(Gtk::Adjustment * adj) {
	double D= adj->get_upper() - adj->get_page_size() - adj->get_lower();
	if(fabs(D) > Epsilon)
		return Cap((adj->get_value() - adj->get_lower()) / D, 0.0, 1.0);
	else
		return 0.0;
}

void SetSliderValue(Gtk::Adjustment * adj, double value) {
	double D= adj->get_upper() - adj->get_page_size() - adj->get_lower();
	if(fabs(D) > Epsilon)
		adj->set_value(value*D + adj->get_lower());
}

void ReviveGtkAssistant(Gtk::Assistant * self) {
	int page= self->get_current_page();
	int count= self->get_n_pages();
	Gtk::Label * label= Gtk::manage(new Gtk::Label("ERROR"));
	int dummy= self->append_page(*label);
	self->set_current_page(dummy);
	if(count > 0)
		self->set_current_page(page>=0 ? page : 0);
	ContainerRemove(self, label);
}

PGdkPixbuf LoadImage(string_cr name, string_cr fallback) {
	PGdkPixbuf pixbuf;
	if(!name.empty()) {
		Glib::RefPtr<Gtk::IconTheme> theme= Gtk::IconTheme::get_default();
		CHECK(theme);
		Gtk::IconInfo info= theme->lookup_icon(name, -1, Gtk::IconLookupFlags());
		if(info)
			try {
				pixbuf= Gdk::Pixbuf::create_from_file(info.get_filename());
			} catch(const Glib::Error &) {
			}
		if(!pixbuf && !fallback.empty())
			try {
				pixbuf= theme->load_icon(fallback, 16, Gtk::ICON_LOOKUP_FORCE_SIZE);
			} catch(const Glib::Error &) {
			}
	}
	return pixbuf;
}

void ResetAppearance(Gtk::Widget * widget) {
	if(Gtk::Button * button= dynamic_cast<Gtk::Button *>(widget)) {
		button->set_use_stock(false);
		button->property_image()= NULL;
		button->set_label("");
		button->set_use_underline(false);
	}
}

void InitGettext() {
	#if ENABLE_NLS
		static bool done= false;
		if(!done) {
			bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
			bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
			done= true;
		}
	#endif
}

template <> Glib::ustring ToString<bool>(const bool & value) {
	return value ? "true" : "false";
}

template <> bool FromString<bool>(ustring_cr str) {
	if(str=="true")
		return true;
	if(str=="false")
		return false;
	RaiseError(_("FromString<bool>() failed"));
	return false;
}

template <> Glib::ustring ToString<Gdk::Color>(const Gdk::Color & value) {
	guint32 pixel=
		((value.get_red() >> 8) << 16) |
		((value.get_green() >> 8) << 8) |
		(value.get_blue() >> 8);
	std::ostringstream oss;
	oss << "#";
	oss << std::setw(6) << std::setfill('0');
	oss << std::hex << std::uppercase;
	oss << pixel;
	return oss.str();
}

template <> Gdk::Color FromString<Gdk::Color>(ustring_cr str) {
	Gdk::Color color;
	if(!color.set(str))
		RaiseError(_("FromString<Gdk::Color>() failed"));
	return color;
}

template <> Glib::ustring ToString<Gtk::StockID>(const Gtk::StockID & value) {
	return value.get_string();
}

template <> Gtk::StockID FromString<Gtk::StockID>(ustring_cr str) {
	return Gtk::StockID(str);
}

// RuntimeError ----------------------------------------------------------------

RuntimeError::RuntimeError(ustring_cr msg) :
	Glib::Error(g_quark_from_static_string(""), 0, msg)
{
}

// Object ----------------------------------------------------------------------

Object::Object() {
	refcounter= 0;
	CrowInit::get()->objectCreated(this);
}

Object::~Object() {
	CHECK(refcounter==0);
	CrowInit::get()->objectDeleted(this);
}

void Object::reference() {
	++refcounter;
}

void Object::unreference() {
	CHECK(refcounter > 0);
	--refcounter;
	if(refcounter==0)
		delete this;
}

bool Object::equals(PObject po) const {
	return this == GetPtr(po);
}

// UniChar ---------------------------------------------------------------------

template <> Glib::ustring ToString<UniChar>(const UniChar & value) {
	gunichar ch= value;
	return Glib::ustring(1, ch);
}

template <> UniChar FromString<UniChar>(ustring_cr str) {
	if(str.size()!=1)
		RaiseError(_("FromString<UniChar>() failed"));
	gunichar ch= str[0];
	return UniChar(ch);
}

template <> UniChar ExtractCppValue<UniChar>(const Glib::ValueBase & val) {
	gunichar ch= ExtractCppValue<gunichar>(val);
	return UniChar(ch);
}

template <> Glib::ValueBase MakeGValue<UniChar>(const UniChar & val) {
	gunichar ch= val;
	return MakeGValue<gunichar>(ch);
}

// Translatable ----------------------------------------------------------------

Translatable::Translatable(ustring_cr meta) {
	trans= false;
	valid= false;
	if(meta.empty()) {
		valid= true;
	} else {
		StringList vstr= Split(meta, "|");
		int size= vstr.size();
		StringList::iterator it= vstr.begin();
		if(1 <= size && size <= 3) {
			trans= *it++ == "translate";
			if(size >= 2) {
				prefix= *it++;
				if(size >= 3)
					comments= *it++;
			}
			testValid();
		}
	}
	CHECK(valid);
}

Translatable::Translatable(bool tra, ustring_cr pre, ustring_cr com) {
	trans= tra;
	prefix= pre;
	comments= com;
	testValid();
}

void Translatable::testValid() {
	valid=
		prefix.find('|') == Glib::ustring::npos &&
		comments.find("*/") == Glib::ustring::npos;
}

Glib::ustring Translatable::getMeta() const {
	CHECK(valid);
	std::ostringstream oss;
	if(trans)
		oss << "translate";
	if(!prefix.empty() || !comments.empty())
		oss << '|';
	if(!prefix.empty())
		oss << prefix;
	if(!comments.empty())
		oss << '|' << comments;
	return oss.str();
}

// Point -----------------------------------------------------------------------

Point::Point() {
	x= 0;
	y= 0;
}

Point::Point(int x, int y) {
	this->x= x;
	this->y= y;
}

bool Point::inside(const Rectangle & r) const {
	return
		r.x <= x && x <= r.x + r.w - 1 &&
		r.y <= y && y <= r.y + r.h - 1
	;
}

Point Point::operator + (const Point & p) const {
	return Point(x+p.x, y+p.y);
}

Point Point::operator - (const Point & p) const {
	return Point(x-p.x, y-p.y);
}

bool Point::operator == (const Point & p) const {
	return x==p.x && y==p.y;
}

Point Point::operator - () const {
	return Point(-x, -y);
}

template <> Glib::ustring ToString<Point>(const Point & value) {
	return ToString(value.x) + "," + ToString(value.y);
}

template <> Point FromString<Point>(ustring_cr str) {
	StringList split= Split(str, ",");
	if(split.size()==1)
		return Point(FromString<int>(split.front()), 0);
	if(split.size()!=2)
		RaiseError(_("FromString<Point>() failed"));
	if(Trim(split.front()).empty())
		return Point(0, FromString<int>(split.back()));
	else
		return Point(FromString<int>(split.front()), FromString<int>(split.back()));
}

// Rectangle -------------------------------------------------------------------

Rectangle::Rectangle() {
	x= y= 0;
	w= h= 0;
}

Rectangle::Rectangle(int x, int y, int w, int h) {
	this->x= x;
	this->y= y;
	this->w= w;
	this->h= h;
}

Rectangle::Rectangle(const Point & point, const Point & size) {
	x= point.x;
	y= point.y;
	w= size.x;
	h= size.y;
}

Rectangle::Rectangle(const Gdk::Rectangle & rect) {
	x= rect.get_x();
	y= rect.get_y();
	w= rect.get_width();
	h= rect.get_height();
}

Rectangle::operator Gdk::Rectangle () const {
	return Gdk::Rectangle(x, y, w, h);
}

bool Rectangle::operator == (const Rectangle & r) const {
	return x==r.x && y==r.y && w==r.w && h==r.h;
}

void Rectangle::grow(int dx, int dy, int dw, int dh) {
	x += dx;
	y += dy;
	w += dw;
	h += dh;
}

void Rectangle::grow(int dx, int dy, int dw, int dh, int minw, int minh) {
	int L= x;
	int T= y;
	int R= x+w;
	int B= y+h;
	growSegment(&L, &R, dx, dx+dw, minw);
	growSegment(&T, &B, dy, dy+dh, minh);
	x= L;
	y= T;
	w= R-L;
	h= B-T;
}

void Rectangle::growSegment(int * a, int * b, int da, int db, int min) {
	if(*b+db-*a-da < min) {
		int k1= min+*a-*b;
		int k2= db-da;
		int da1= (da*k1)/k2;
		int db1= (db*k1)/k2;
		CHECK(da1*k2 == da*k1);
		CHECK(db1*k2 == db*k1);
		*a += da1;
		*b += db1;
	} else {
		*a += da;
		*b += db;
	}
}

bool Rectangle::empty() const {
	return w <= 0 || h <= 0;
}

// Border ----------------------------------------------------------------------

Border::Border() {
	left= right= top= bottom= 0;
}

bool Border::operator == (const Border & b) const {
	return
		left == b.left &&
		right == b.right &&
		top == b.top &&
		bottom == b.bottom;
}

Border::operator Gtk::Border () const {
	Gtk::Border border;
	border.left= left;
	border.right= right;
	border.top= top;
	border.bottom= bottom;
	return border;
}

template <> Glib::ustring ToString<Border>(const Border & value) {
	return
		ToString(value.left) + "," +
		ToString(value.right) + "," +
		ToString(value.top) + "," +
		ToString(value.bottom);
}

template <> Border FromString<Border>(ustring_cr str) {
	StringList split= Split(str, ",");
	if(split.size()!=4)
		RaiseError(_("FromString<Border>() failed"));
	Border result;
	StringList::iterator it= split.begin();
	result.left= FromString<int>(*it++);
	result.right= FromString<int>(*it++);
	result.top= FromString<int>(*it++);
	result.bottom= FromString<int>(*it++);
	return result;
}

// Emitter ---------------------------------------------------------------------

Emitter::Emitter(string_cr sname, string_cr mname, bool after) {
	signame= sname;
	mapname= mname;
	this->after= after;
}

bool Emitter::operator == (const Emitter & e) const {
	return signame==e.signame && mapname==e.mapname && after==e.after;
}

template <> Glib::ustring ToString<Emitter>(const Emitter & value) {
	return value.signame + "," + value.mapname + "," + ToString(value.after);
}

template <> Emitter FromString<Emitter>(ustring_cr str) {
	StringList split= Split(str, ",");
	if(split.size()!=3)
		RaiseError(_("FromString<Emitter>() failed"));
	StringList::iterator it= split.begin();
	Emitter result;
	result.signame= *it++;
	result.mapname= *it++;
	result.after= FromString<bool>(*it++);
	return result;
}

// Parser ----------------------------------------------------------------------

Parser::Parser() :
	glibParser(this)
{
}

void Parser::parse(Parser * parser, ustring_cr data) {
	Glib::Markup::ParseContext context(parser->glibParser);
	context.parse(data);
	context.end_parse();
}

void Parser::GlibParser::raiseError(const Glib::Exception & x) {
	throw Glib::MarkupError(Glib::MarkupError::INVALID_CONTENT, x.what());
}

void Parser::GlibParser::on_start_element(Glib::Markup::ParseContext & context, ustring_cr name, const AttributeMap & properties) {
	try {
		parser->onStartElement(name, properties);
	} catch(const Glib::Exception & x) {
		raiseError(x);
	}
}

void Parser::GlibParser::on_end_element(Glib::Markup::ParseContext & context, ustring_cr name) {
	try {
		parser->onEndElement(name);
	} catch(const Glib::Exception & x) {
		raiseError(x);
	}
}

void Parser::GlibParser::on_text(Glib::Markup::ParseContext & context, ustring_cr text) {
	try {
		parser->onText(text);
	} catch(const Glib::Exception & x) {
		raiseError(x);
	}
}

void Parser::GlibParser::on_passthrough(Glib::Markup::ParseContext & context, ustring_cr text) {
	try {
		parser->onPassthrough(text);
	} catch(const Glib::Exception & x) {
		raiseError(x);
	}
}

// UiElementExtract ------------------------------------------------------------

UiElementExtract::emap UiElementExtract::extract(ustring_cr str) {
	depth= 0;
	result.clear();
	try {
		Parser::parse(this, str);
	} catch(const Glib::Error & x) {
		FatalError(_("UI parsing failed: ") + x.what());
	}
	return result;
}

void UiElementExtract::onStartElement(ustring_cr element, const Glib::Markup::Parser::AttributeMap & attributes) {

	++depth;

	if(depth == 2) {

		Glib::Markup::Parser::AttributeMap::const_iterator iname= attributes.find("name");
		Glib::Markup::Parser::AttributeMap::const_iterator iaction= attributes.find("action");
		Glib::ustring name, action;

		if(iname!=attributes.end())
			name= iname->second;
		if(iaction!=attributes.end())
			action= iaction->second;

		Glib::ustring realname;

		if(!name.empty())
			realname= name;
		else if(!action.empty())
			realname= action;
		else
			realname= element;

		result.insert(std::make_pair(realname, element));

	}

}

}
