/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "forward.h"
#include "widgets.h"

#include "palette.h"
#include "session.h"
#include "editors.h"

namespace Crow {

// HierarchyWidget -------------------------------------------------------------

HierarchyWidget::HierarchyWidget() {
	set_shadow_type(Gtk::SHADOW_IN);
}

HierarchyEditorWidget * HierarchyWidget::getEditorWidget() {
	return dynamic_cast<HierarchyEditorWidget *>(get_child());
}

void HierarchyWidget::setEditorWidget(EditorWidget * ew) {
	add(*ew);
	ew->show();
}

void HierarchyWidget::unsetEditorWidget() {
	BinRemove(this);
}

int HierarchyWidget::getSplit() {
	return getEditorWidget()->getSplit();
}

void HierarchyWidget::setSplit(int split) {
	getEditorWidget()->setSplit(split);
}

// PaletteWidget ---------------------------------------------------------------

PaletteWidget::PaletteWidget() {

	pushed= NULL;

	set_shadow_type(Gtk::SHADOW_NONE);

	scroll.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_ALWAYS);
	add(scroll);
	scroll.show();

	scroll.add(vBox);
	vBox.show();

	create();
}

void PaletteWidget::setShowDW(bool show_dw) {
	unpush();
	debugTab->property_visible()= show_dw;
}

void PaletteWidget::unpush() {
	if(pushed) {
		Gtk::ToggleButton * button= pushed;
		pushed= NULL;
		button->set_active(false);
	}
}

void PaletteWidget::create() {

	typedef std::list<EntityEntry *> evec;
	typedef std::vector<evec> tmap;

	tmap tabs(TAB_COUNT);

	const TypeList & entries= *Palette::get()->getEntries();

	for(TypeList::const_iterator it= entries.begin(); it!=entries.end(); ++it) {
		EntityEntry * entry= GetEntry<EntityEntry>(*it);
		if(entry) {
			PaletteTab tab= entry->getPaletteTab();
			if(tab!=TAB_NONE)
				tabs[tab].push_back(entry);
		}
	}

	for(int i=0; i<SIGNED(tabs.size()); ++i)
		if(!tabs[i].empty()) {
			string_cr label= *Palette::get()->getTabLabel(static_cast<PaletteTab>(i));
			const evec & vec= tabs[i];

			Gtk::Expander * expander= Gtk::manage(new Gtk::Expander());
			expander->set_label(label);

			Gtk::VBox * vb= manage(new Gtk::VBox());
			expander->add(*vb);
			vb->show();

			for(evec::const_iterator j= vec.begin(); j!=vec.end(); ++j) {
				Gtk::Widget * widget= createPaletteButton(*j);
				vb->pack_start(*widget);
				widget->show();
			}

			vBox.pack_start(*expander, Gtk::PACK_SHRINK);

			if(i != TAB_DEBUG)
				expander->show();
			else
				debugTab= expander;
		}

}

Gtk::Widget * PaletteWidget::createPaletteButton(EntityEntry * entry) {

	Gtk::HBox * box= Gtk::manage(new Gtk::HBox());

	bool root= !Palette::get()->isWidget(entry->getType()) || entry->hasFlag(efTopMaster);

	Gtk::Button * tbutton= Gtk::manage(root ? new Gtk::Button() : new Gtk::ToggleButton());
	tbutton->set_relief(Gtk::RELIEF_NONE);
	tbutton->set_alignment(0.0f, 0.5f);
	tbutton->property_can_focus()= false;
	tbutton->set_label(*entry->getTypeName());
	tbutton->signal_clicked().connect(
		sigc::bind<TypeId, Gtk::ToggleButton *>(
			sigc::mem_fun(*this, &PaletteWidget::onEntryClicked),
			entry->getType(), root ? NULL : dynamic_cast<Gtk::ToggleButton *>(tbutton)
		)
	);
	box->pack_end(*tbutton);
	tbutton->show();

	Gtk::Button * abutton= Gtk::manage(new Gtk::Button());
	abutton->set_relief(Gtk::RELIEF_NONE);
	abutton->property_can_focus()= false;
	abutton->signal_clicked().connect(
		sigc::bind<TypeId, Gtk::ToggleButton *>(
			sigc::mem_fun(*this, &PaletteWidget::onEntryClicked),
			entry->getType(), NULL
		)
	);
	abutton->set_sensitive(!root);
	box->pack_start(*abutton, Gtk::PACK_SHRINK);
	abutton->show();

	Gtk::HBox * bbox= Gtk::manage(new Gtk::HBox());
	abutton->add(*bbox);
	bbox->show();

	Gtk::Arrow * arrow= Gtk::manage(new Gtk::Arrow(Gtk::ARROW_RIGHT, Gtk::SHADOW_NONE));
	arrow->set_size_request(14, 14);
	bbox->pack_start(*arrow, Gtk::PACK_EXPAND_PADDING);
	arrow->show();

	return box;
}

void PaletteWidget::onEntryClicked(TypeId type, Gtk::ToggleButton * button) {
	if(button && !button->get_active()) {
		if(pushed!=NULL) {
			pushed= NULL;
			signalPaletteEvent.emit(type, peCancel);
		}
	} else {
		unpush();
		pushed= button;
		signalPaletteEvent.emit(type, pushed==NULL ? peRoot : peCanvas);
	}
}

// CanvasWidget ----------------------------------------------------------------

CanvasWidget::CanvasWidget() {
	set_shadow_type(Gtk::SHADOW_IN);
}

void CanvasWidget::setEditorWidget(EditorWidget * ew) {
	set_shadow_type(ew->isFrameNeeded() ? Gtk::SHADOW_IN : Gtk::SHADOW_NONE);
	add(*ew);
	ew->show();
}

void CanvasWidget::unsetEditorWidget() {
	BinRemove(this);
	set_shadow_type(Gtk::SHADOW_IN);
}

// PropertyTree ----------------------------------------------------------------

PropertyTree::PropertyTree() {
	createModel(2);
	addColumn(0, _("Name"));
	addColumn(1, _("Value"));
}

int PropertyTree::getSplit() {
	return getColumnWidth(0);
}

void PropertyTree::setSplit(int split) {
	setColumnWidth(0, split);
}

// ExplorerWidget --------------------------------------------------------------

ExplorerWidget::ExplorerWidget() {
	set_shadow_type(Gtk::SHADOW_NONE);

	add(notebook);
	notebook.show();

	addPage(_("Properties"));
	addPage(_("Signals"));
	addPage(_("Packing"));

	notebook.set_current_page(0);

	prev_page= 0;
	notebook.signal_switch_page().connect(sigc::mem_fun(*this, &ExplorerWidget::onSwitchPage));

}

PropertyTree * ExplorerWidget::getTreeView(ExplorerPage page) {
	CHECK(0 <= page && page <= 2);
	Gtk::Bin * bin= dynamic_cast<Gtk::Bin *>(notebook.get_nth_page(page));
	CHECK(bin);
	PropertyTree * tree= dynamic_cast<PropertyTree *>(bin->get_child());
	CHECK(tree);
	return tree;
}

int ExplorerWidget::getSplit() {
	return getCurrentTree()->getSplit();
}

void ExplorerWidget::setSplit(int split) {
	getCurrentTree()->setSplit(split);
}

PSession ExplorerWidget::getSelectedRow() {
	const std::vector<PObject> & selection= *getCurrentTree()->getSelection();
	if(selection.size()==1) {
		PSession session= PSession::cast_dynamic(selection.front());
		return session;
	}
	return PSession();
}

void ExplorerWidget::setSelectedRow(PSession session) {
	PropertyTree * tree= getCurrentTree();
	std::vector<PObject> selection(1, session);
	tree->setSelection(selection, true);
}

ExplorerWidget::ExplorerPage ExplorerWidget::getCurrentPage() {
	return (ExplorerPage)notebook.get_current_page();
}

PropertyTree * ExplorerWidget::getCurrentTree() {
	return getTreeView(getCurrentPage());
}

void ExplorerWidget::addPage(string_cr tab) {

	Gtk::ScrolledWindow * page= Gtk::manage(new Gtk::ScrolledWindow());
	page->set_policy(Gtk::POLICY_NEVER, Gtk::POLICY_ALWAYS);
	notebook.append_page(*page, tab);
	page->show();

	PropertyTree * tree= Gtk::manage(new PropertyTree());
	page->add(*tree);
	tree->show();

	tree->connect_selection_changed(sigc::bind<ExplorerPage>(
		sigc::mem_fun(*this, &ExplorerWidget::onSelectionChanged),
		ExplorerPage(notebook.pages().size()-1)
	));

}

void ExplorerWidget::onSelectionChanged(ExplorerPage page) {
	if(notebook.get_current_page() == checked_cast<int>(page))
		signalSelectionChanged.emit();
}

void ExplorerWidget::onSwitchPage(GtkNotebookPage *, guint) {
	if(prev_page != notebook.get_current_page()) {
		int split= getTreeView((ExplorerPage)prev_page)->getSplit();
		getCurrentTree()->setSplit(split);
		prev_page= notebook.get_current_page();
		signalSelectionChanged.emit();
	}
}

// SignalSelectDialog ----------------------------------------------------------

SignalSelectDialog::SignalSelectDialog() :
	Gtk::Dialog(_("Signal Select"))
{
	lock= false;

	property_default_height()= 300;

	cancelButton= add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_NONE);
	cancelButton->property_can_default()= true;
	okButton= add_button(Gtk::Stock::OK, Gtk::RESPONSE_OK);
	okButton->property_can_default()= true;

	set_default(*cancelButton);
	set_focus(*cancelButton);
	okButton->set_sensitive(false);

	scroll.set_policy(Gtk::POLICY_NEVER, Gtk::POLICY_ALWAYS);
	get_vbox()->add(scroll);
	scroll.show();

	scroll.add(vBox);
	vBox.show();

}

void SignalSelectDialog::setChoices(const SignalsQuery & signals) {

	vBox.children().clear();

	GType gtype= G_TYPE_INVALID;
	std::vector<std::string> names;

	names.reserve(signals.size());

	for(SignalsQuery::const_iterator it= signals.begin(); it!=signals.end(); ++it) {

		if(it->first != gtype) {
			addTreeView(gtype, names);
			gtype= it->first;
			names.clear();
		}

		names.push_back(it->second);
	}

	addTreeView(gtype, names);

}

void SignalSelectDialog::addTreeView(GType gtype, const std::vector<std::string> & names) {

	if(names.empty())
		return;

	Gtk::TreeView * tree= Gtk::manage(new Gtk::TreeView());
	tree->property_can_focus()= false;
	vBox.pack_end(*tree);
	tree->show();

	Gtk::TreeModelColumn<Glib::ustring> column;
	Gtk::TreeModel::ColumnRecord columns;
	columns.add(column);
	PGtkListStore model= Gtk::ListStore::create(columns);

	for(int i=0; i<SIGNED(names.size()); ++i) {
		Gtk::TreeModel::iterator iter= model->append();
		iter->set_value(0, names[i]);
	}

	tree->set_model(model);
	tree->append_column(g_type_name(gtype), column);

	tree->get_selection()->signal_changed().connect(sigc::bind<Gtk::TreeView *>(
		sigc::mem_fun(*this, &SignalSelectDialog::onSelectionChanged), tree
	));

}

void SignalSelectDialog::onSelectionChanged(Gtk::TreeView * tree) {

	if(lock)
		return;

	lock= true;

	bool empty= tree->get_selection()->count_selected_rows() == 0;
	if(!empty) {

		std::vector<Gtk::Widget *> trees= vBox.get_children();
		for(int i=0; i<SIGNED(trees.size()); ++i) {
			Gtk::TreeView * widget= dynamic_cast<Gtk::TreeView *>(trees[i]);
			if(widget != tree)
				widget->get_selection()->unselect_all();
		}

		Gtk::TreeRow row= *tree->get_selection()->get_selected();
		Glib::ustring data;
		row.get_value(0, data);
		selection= data;

	} else
		selection.clear();

	okButton->set_sensitive(!empty);

	lock= false;

}

// TextInputDialog -------------------------------------------------------------

TextInputDialog::TextInputDialog() :
	Gtk::Dialog(_("Text Input"))
{

	set_default_size(400, 400);

	add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_NONE);
	add_button(Gtk::Stock::OK, Gtk::RESPONSE_OK);

	get_vbox()->add(paned);
	paned.show();

	scrollText.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
	scrollText.set_shadow_type(Gtk::SHADOW_IN);
	paned.pack1(scrollText, true, false);
	scrollText.show();

	text.set_wrap_mode(Gtk::WRAP_WORD);
	scrollText.add(text);
	text.show();

	paned.pack2(frame, false, false);
	frame.show();

	check.set_label(_("Enable translation"));
	check.set_active(true);
	frame.set_label_widget(check);
	check.show();

	align.set_padding(8, 8, 8, 8);
	frame.add(align);
	align.show();

	box1.set_spacing(8);
	align.add(box1);
	box1.show();

	box2.set_spacing(8);
	box1.pack_start(box2, Gtk::PACK_SHRINK);
	box2.show();

	box1.pack_start(box3);
	box3.show();

	label1.property_xalign()= 0;
	label1.set_label(_("Context prefix:"));
	box2.pack_start(label1, Gtk::PACK_SHRINK);
	label1.show();

	box2.pack_start(entry);
	entry.show();

	label2.property_xalign()= 0;
	label2.set_label(_("Comments for translators:"));
	box3.pack_start(label2, Gtk::PACK_SHRINK);
	label2.show();

	scrollComments.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
	scrollComments.set_shadow_type(Gtk::SHADOW_IN);
	box3.pack_start(scrollComments);
	scrollComments.show();

	scrollComments.add(comments);
	comments.show();

}

Glib::ustring TextInputDialog::getText() {
	return text.get_buffer()->get_text();
}

void TextInputDialog::setText(ustring_cr str) {
	text.get_buffer()->set_text(str);
	text.get_buffer()->place_cursor(text.get_buffer()->begin());
}

Translatable TextInputDialog::getMeta() {
	return Translatable(check.get_active(), entry.get_text(), comments.get_buffer()->get_text());
}

void TextInputDialog::setMeta(const Translatable & trans) {
	check.set_active(trans.getTranslate());
	entry.set_text(*trans.getPrefix());
	comments.get_buffer()->set_text(*trans.getComments());
	comments.get_buffer()->place_cursor(comments.get_buffer()->begin());
}

// IconSelectDialog ------------------------------------------------------------

IconSelectDialog::IconSelectDialog() :
	Gtk::Dialog(_("Icon Select"))
{
	property_default_height()= 300;

	cancelButton= add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_NONE);
	cancelButton->property_can_default()= true;
	okButton= add_button(Gtk::Stock::OK, Gtk::RESPONSE_OK);
	okButton->property_can_default()= true;

	set_default(*cancelButton);
	set_focus(*cancelButton);
	okButton->set_sensitive(false);

	scroll.set_policy(Gtk::POLICY_NEVER, Gtk::POLICY_ALWAYS);
	get_vbox()->add(scroll);
	scroll.show();

	treeView.property_can_focus()= false;
	treeView.set_headers_visible(false);
	scroll.add(treeView);
	treeView.show();

	treeView.get_selection()->signal_changed().connect(sigc::mem_fun(*this, &IconSelectDialog::onSelectionChanged));

}

void IconSelectDialog::setChoices(const std::vector<std::pair<PGdkPixbuf, Glib::ustring> > & choices) {

	Gtk::TreeModelColumn<PGdkPixbuf> column1;
	Gtk::TreeModelColumn<Glib::ustring> column2;
	Gtk::TreeModel::ColumnRecord columns;
	columns.add(column1);
	columns.add(column2);
	PGtkListStore model= Gtk::ListStore::create(columns);

	for(int i=0; i<SIGNED(choices.size()); ++i) {
		Gtk::TreeModel::iterator iter= model->append();
		iter->set_value(0, choices[i].first);
		iter->set_value(1, choices[i].second);
	}

	model->set_sort_column(column2, Gtk::SORT_ASCENDING);

	treeView.remove_all_columns();
	treeView.set_model(model);
	treeView.append_column("", column1);
	treeView.append_column("", column2);

}

Glib::ustring IconSelectDialog::getChoice() {
	Gtk::TreeRow row= *treeView.get_selection()->get_selected();
	Glib::ustring name;
	row.get_value(1, name);
	return name;
}

void IconSelectDialog::setChoice(ustring_cr name) {
	Gtk::TreeModel::Children rows= treeView.get_model()->children();
	for(Gtk::TreeModel::Children::iterator it= rows.begin(); it!=rows.end(); ++it) {
		Glib::ustring data;
		it->get_value(1, data);
		if(data==name) {
			Gtk::TreePath path= treeView.get_model()->get_path(*it);
			treeView.get_selection()->select(path);
			treeView.scroll_to_row(path);
			break;
		}
	}
}

void IconSelectDialog::onSelectionChanged() {
	bool empty= treeView.get_selection()->count_selected_rows() == 0;
	okButton->set_sensitive(!empty);
}

// FlagsDialog -----------------------------------------------------------------

FlagsDialog::FlagsDialog() {

	property_default_height()= 300;

	Gtk::Button * cancelButton= add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_NONE);
	add_button(Gtk::Stock::OK, Gtk::RESPONSE_OK);

	set_focus(*cancelButton);

	scroll.set_policy(Gtk::POLICY_NEVER, Gtk::POLICY_ALWAYS);
	get_vbox()->add(scroll);
	scroll.show();

	treeView.property_can_focus()= false;
	treeView.set_headers_visible(false);
	treeView.get_selection()->set_mode(Gtk::SELECTION_NONE);
	scroll.add(treeView);
	treeView.show();
}

void FlagsDialog::setOptions(const Values & options) {

	Gtk::TreeModelColumn<Bits> column0;
	Gtk::TreeModelColumn<bool> column1;
	Gtk::TreeModelColumn<Glib::ustring> column2;
	Gtk::TreeModel::ColumnRecord columns;
	columns.add(column0);
	columns.add(column1);
	columns.add(column2);
	PGtkListStore model= Gtk::ListStore::create(columns);

	for(Values::const_iterator it= options.begin(); it!=options.end(); ++it) {
		Gtk::TreeModel::iterator iter= model->append();
		iter->set_value(0, it->second);
		iter->set_value(1, false);
		iter->set_value(2, it->first);
	}

	treeView.remove_all_columns();
	treeView.set_model(model);
	treeView.append_column_editable("", column1);
	treeView.append_column("", column2);

}

Bits FlagsDialog::getValue() {
	Bits value= 0;
	Gtk::TreeModel::Children rows= treeView.get_model()->children();
	for(Gtk::TreeModel::Children::iterator it= rows.begin(); it!=rows.end(); ++it) {
		bool set;
		it->get_value(1, set);
		if(set) {
			Bits flag;
			it->get_value(0, flag);
			value |= flag;
		}
	}
	return value;
}

void FlagsDialog::setValue(Bits value) {
	Gtk::TreeModel::Children rows= treeView.get_model()->children();
	for(Gtk::TreeModel::Children::iterator it= rows.begin(); it!=rows.end(); ++it) {
		Bits flag;
		it->get_value(0, flag);
		bool set= (flag & value) != 0;
		it->set_value(1, set);
	}
}

// TranslatablesDialog ---------------------------------------------------------

TranslatablesDialog::TranslatablesDialog() :
	Gtk::Dialog(_("Translatable strings"))
{

	property_default_height()= 300;
	property_default_width()= 525;

	invertButton.set_use_underline(true);
	invertButton.set_label(_("_All/None"));
	invertButton.property_can_default()= true;
	invertButton.set_image(*Gtk::manage(new Gtk::Image(Gtk::Stock::SELECT_ALL, Gtk::ICON_SIZE_BUTTON)));
	get_action_area()->pack_start(invertButton);
	get_action_area()->set_child_secondary(invertButton);
	invertButton.show();
	invertButton.signal_clicked().connect(sigc::mem_fun(*this, &TranslatablesDialog::onInvert));

	Gtk::Button * cancelButton= add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_NONE);
	add_button(Gtk::Stock::OK, Gtk::RESPONSE_OK);

	set_focus(*cancelButton);

	scroll.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_ALWAYS);
	get_vbox()->add(scroll);
	scroll.show();

	scroll.add(treeView);
	treeView.show();
	treeView.set_rules_hint();
	treeView.set_grid_lines(Gtk::TREE_VIEW_GRID_LINES_VERTICAL);
}

void TranslatablesDialog::setStrings(const Strings & strings) {

	set= strings;

	Gtk::TreeModelColumn<Glib::ustring> column0;
	Gtk::TreeModelColumn<Glib::ustring> column1;
	Gtk::TreeModelColumn<bool> column2;
	Gtk::TreeModelColumn<Glib::ustring> column3;
	Gtk::TreeModelColumn<Glib::ustring> column4;
	Gtk::TreeModel::ColumnRecord columns;
	columns.add(column0);
	columns.add(column1);
	columns.add(column2);
	columns.add(column3);
	columns.add(column4);
	PGtkListStore model= Gtk::ListStore::create(columns);

	for(int i=0; i<SIGNED(strings.size()); ++i) {
		Gtk::TreeModel::iterator iter= model->append();
		iter->set_value(0, IdPathToString(strings[i].path, true));
		iter->set_value(1, strings[i].string);
		iter->set_value(2, strings[i].meta.getTranslate());
		iter->set_value(3, *strings[i].meta.getPrefix());
		iter->set_value(4, *strings[i].meta.getComments());
	}

	treeView.remove_all_columns();
	treeView.set_model(model);
	treeView.append_column(_("Property"), column0);
	treeView.append_column_editable(_("String"), column1);
	treeView.append_column_editable(_("Translate"), column2);
	treeView.append_column_editable(_("Prefix"), column3);
	treeView.append_column_editable(_("Comments"), column4);

	setupColumn(0, 100);
	setupColumn(1, 150);
	setupColumn(3, 75);
	setupColumn(4, 100);

}

TranslatablesDialog::Strings TranslatablesDialog::getStrings() {
	Strings result= set;
	Gtk::TreeModel::Children rows= treeView.get_model()->children();
	for(int i= 0; i<SIGNED(rows.size()); ++i) {
		Gtk::TreeRow row= rows[i];
		Glib::ustring string;
		bool translate;
		Glib::ustring prefix;
		Glib::ustring comments;
		row.get_value(1, string);
		row.get_value(2, translate);
		row.get_value(3, prefix);
		row.get_value(4, comments);
		result[i].string= string;
		result[i].meta= Translatable(translate, prefix, comments);
	}
	return result;
}

void TranslatablesDialog::setupColumn(int column, int width) {
	SetColumnWidth(treeView.get_column(column), width);
	Gtk::CellRendererText * renderer= dynamic_cast<Gtk::CellRendererText *>(treeView.get_column_cell_renderer(column));
	renderer->property_single_paragraph_mode()= true;
	renderer->set_property("ellipsize", Pango::ELLIPSIZE_END);
}

void TranslatablesDialog::onInvert() {
	Gtk::TreeModel::Children rows= treeView.get_model()->children();
	std::vector<bool> flags(rows.size(), true);
	bool value= false;
	for(int i= 0; i<SIGNED(rows.size()); ++i) {
		Gtk::TreeRow row= rows[i];
		Glib::ustring string;
		bool translate;
		row.get_value(1, string);
		row.get_value(2, translate);
		if(Trim(string).empty() || string==set[i].defval) {
			flags[i]= false;
			row.set_value(2, false);
		} else if(!translate)
			value= true;
	}
	for(int i= 0; i<SIGNED(rows.size()); ++i)
		if(flags[i])
			rows[i].set_value(2, value);
}

// Enclosure -------------------------------------------------------------------

void Enclosure::setWidget(Gtk::Widget * W) {
	if(getWidget() != W) {
		unsetWidget();
		if(W) {
			window= dynamic_cast<Gtk::Window *>(W);
			if(window) {
				window->reference();
				encloseWindow(GTK_WIDGET(window->gobj()));
				add(*window);
				window->show();
			} else
				add(*W);
		}
	}
}

void Enclosure::unsetWidget() {
	if(window) {
		BinRemove(this);
		unlooseWindow(GTK_WIDGET(window->gobj()));
		ReleaseWindow(window);
		window= NULL;
	} else {
		if(get_child())
			BinRemove(this);
	}
}

Gtk::Widget * Enclosure::getWidget() {
	if(window)
		return window;
	else
		return get_child();
}

void Enclosure::onRealize(GtkWidget * widget) {
	GtkWindow * window= GTK_WINDOW(widget);
	gint attributes_mask= 0;

	GdkWindowAttr attributes;
	memset(&attributes, 0, sizeof(attributes));

	GTK_WIDGET_SET_FLAGS(widget, GTK_REALIZED);

	attributes.window_type= GDK_WINDOW_CHILD;
	attributes.wclass= GDK_INPUT_OUTPUT;
	attributes.visual= gtk_widget_get_visual(widget);
	attributes.colormap= gtk_widget_get_colormap(widget);
	attributes.x= widget->allocation.x;
	attributes.y= widget->allocation.y;
	attributes.width= widget->allocation.width;
	attributes.height= widget->allocation.height;
	attributes.event_mask= gtk_widget_get_events(widget) | GDK_EXPOSURE_MASK;
	attributes_mask= GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

	if(GDK_IS_WINDOW(widget->window))
		gdk_window_destroy(widget->window);

	widget->window= gdk_window_new(
		gtk_widget_get_parent_window(widget),
		&attributes, attributes_mask
	);

	widget->style= gtk_style_attach(widget->style, widget->window);
	gtk_style_set_background(widget->style, widget->window, GTK_STATE_NORMAL);

	gdk_window_set_user_data(widget->window, window);
}

void Enclosure::onSizeAllocate(GtkWidget * widget, GtkAllocation * allocation) {
	if(GTK_WIDGET_REALIZED(widget))
		gdk_window_move_resize(
			widget->window,
			widget->allocation.x,
			widget->allocation.y,
			widget->allocation.width,
			widget->allocation.height
		);
}

void Enclosure::onSizeRequest(GtkWidget * widget, GtkRequisition * requisition) {
	if(GtkWidget * parent= gtk_widget_get_parent(widget)) {
		int W, H;
		gtk_widget_get_size_request(parent, &W, &H);
		if(W != -1 && H != -1) {
			requisition->width= W;
			requisition->height= H;
			return;
		}
	}
	GTK_WIDGET_GET_CLASS(widget)->size_request(widget, requisition);
}

void Enclosure::encloseWindow(GtkWidget * widget) {
	GtkWindow * window= GTK_WINDOW(widget);
	const char * key= "CrowEnclosedWindow";

	int level= GPOINTER_TO_INT(g_object_get_data(G_OBJECT(window), key));

	if(level < 2) {

		if(GTK_WIDGET_REALIZED(widget))
			gtk_widget_unrealize (widget);

		GTK_WIDGET_UNSET_FLAGS(widget, GTK_TOPLEVEL);

		if(level == 0) {
			gtk_container_set_resize_mode(GTK_CONTAINER (window), GTK_RESIZE_PARENT);
			g_signal_connect(window, "realize", G_CALLBACK(onRealize), NULL);
			g_signal_connect(window, "size-allocate", G_CALLBACK(onSizeAllocate), NULL);
			g_signal_connect(window, "size-request", G_CALLBACK(onSizeRequest), NULL);
		}

		g_object_set_data(G_OBJECT(window), key, GINT_TO_POINTER(2));
	}
}

void Enclosure::unlooseWindow(GtkWidget * widget) {
	GtkWindow * window= GTK_WINDOW(widget);
	const char * key= "CrowEnclosedWindow";

	int level= GPOINTER_TO_INT(g_object_get_data(G_OBJECT(window), key));

	if(level == 2) {

		if(GTK_WIDGET_REALIZED(widget))
			gtk_widget_unrealize (widget);

		GTK_WIDGET_SET_FLAGS(widget, GTK_TOPLEVEL);

		g_object_set_data(G_OBJECT(window), key, GINT_TO_POINTER(1));
	}
}

// DesignWindow ----------------------------------------------------------------

DesignWindow::DesignWindow() {
	bordering= true;
	set_redraw_on_allocate(true);
	BS= 2;
	IS= 16;
	BI= 3;
	int tw, th;
	createLayout("", -1)->get_pixel_size(tw, th);
	CS= th;
	HS= std::max(CS+2, IS+2);
	updatePadding();
	RW= 4+4*(2+IS)+4 +3*IS;
	RH= 4+HS+2*BS;
	loadIcon();
}

void DesignWindow::setBordering(bool b) {
	if(bordering != b) {
		bordering= b;
		updatePadding();
		queue_resize();
	}
}

void DesignWindow::setTitle(ustring_cr t) {
	if(title != t) {
		title= t;
		onTitleChanged();
	}
}

void DesignWindow::setIconName(ustring_cr in) {
	if(iconName != in) {
		iconName= in;
		onIconNameChanged();
	}
}

bool DesignWindow::on_expose_event(GdkEventExpose * event) {
	bool result= Gtk::Alignment::on_expose_event(event);

	if(bordering) {

		PGdkWindow window= get_window();
		PGdkGC gc= Gdk::GC::create(window);

		Rectangle A= get_allocation();

		gc->set_foreground(GetBlack(get_style()));
		window->draw_rectangle(gc, false, A.x, A.y, A.w-1, A.h-1);
		gc->set_foreground(GetWhite(get_style()));
		window->draw_line(gc, A.x+1, A.y+1, A.x+A.w-2, A.y+1);
		window->draw_line(gc, A.x+1, A.y+1, A.x+1, A.y+A.h-2);
		gc->set_foreground(GetDark(get_style()));
		window->draw_line(gc, A.x+A.w-2, A.y+A.h-2, A.x+A.w-2, A.y+2);
		window->draw_line(gc, A.x+A.w-2, A.y+A.h-2, A.x+2, A.y+A.h-2);

		drawIcon(gc, getButtonRect(A, 0));

		Rectangle cap= getCaptionRect(A);

		if(cap.w > 4) {
			gc->set_foreground(GetBlack(get_style()));
			window->draw_rectangle(gc, false, cap.x, cap.y, cap.w-1, cap.h-1);

			get_style()->paint_flat_box(
				window, Gtk::STATE_INSENSITIVE, Gtk::SHADOW_NONE, A, *this, "base",
				cap.x+1, cap.y+1, cap.w-2, cap.h-2
			);

			PPangoLayout layout= createLayout(title, cap.w-4);

			get_style()->paint_layout(
				window, Gtk::STATE_INSENSITIVE, true, cap, *this, "base",
				cap.x+2, cap.y+1, layout
			);

			gc->set_foreground(GetBlack(get_style()));
			drawMinimize(gc, getButtonRect(A, -3));
			drawMaximize(gc, getButtonRect(A, -2));
			drawClose(gc, getButtonRect(A, -1));

		}

	}

	return result;
}

void DesignWindow::on_size_request(Gtk::Requisition * requisition) {
	Gtk::Alignment::on_size_request(requisition);
	if(bordering) {
		requisition->width= std::max(RW, requisition->width);
		requisition->height= std::max(RH, requisition->height);
	}
}

PPangoLayout DesignWindow::createLayout(ustring_cr text, int width) {
	PPangoLayout layout= create_pango_layout(text);
	Pango::AttrList attributes;
	ChangeAttribute(&attributes, Pango::Attribute::create_attr_weight(Pango::WEIGHT_BOLD));
	layout->set_attributes(attributes);
	layout->set_ellipsize(Pango::ELLIPSIZE_END);
	layout->set_single_paragraph_mode(true);
	layout->set_width(width==-1 ? -1 : width*Pango::SCALE);
	return layout;
}

Rectangle DesignWindow::getCaptionRect(const Rectangle & A) {
	int BX= A.w-1-(2+3*(IS+2));
	return Rectangle(A.x+IS+5, A.y+2, BX-(IS+5), HS);
}

Rectangle DesignWindow::getButtonRect(const Rectangle & A, int pos) {
	int X;
	if(pos < 0)
		X= A.w-1-2+pos*(IS+2) +2;
	else
		X= 3;
	return Rectangle(A.x+X, A.y+3, IS, IS);
}

void DesignWindow::drawIcon(PGdkGC gc, const Rectangle & R) {
	if(icon)
		get_window()->draw_pixbuf(gc, icon, 0, 0, R.x, R.y, IS, IS, Gdk::RGB_DITHER_NONE, 0, 0);
}

void DesignWindow::drawMinimize(PGdkGC gc, const Rectangle & R) {
	get_window()->draw_rectangle(gc, false, R.x+BI, R.y+IS-1-BI-1, IS-2*BI-1, 1);
}

void DesignWindow::drawMaximize(PGdkGC gc, const Rectangle & R) {
	get_window()->draw_rectangle(gc, false, R.x+BI, R.y+BI, IS-2*BI-1, IS-2*BI-1);
	get_window()->draw_line(gc, R.x+BI+1, R.y+BI+1, R.x+IS-1-BI-1, R.y+BI+1);
}

void DesignWindow::drawClose(PGdkGC gc, const Rectangle & R) {
	get_window()->draw_line(gc, R.x+BI, R.y+BI, R.x+IS-1-BI, R.y+IS-1-BI);
	get_window()->draw_line(gc, R.x+BI, R.y+BI+1, R.x+IS-1-BI-1, R.y+IS-1-BI);
	get_window()->draw_line(gc, R.x+BI+1, R.y+BI, R.x+IS-1-BI, R.y+IS-1-BI-1);
	get_window()->draw_line(gc, R.x+IS-1-BI, R.y+BI, R.x+BI, R.y+IS-1-BI);
	get_window()->draw_line(gc, R.x+IS-1-BI, R.y+BI+1, R.x+BI+1, R.y+IS-1-BI);
	get_window()->draw_line(gc, R.x+IS-1-BI-1, R.y+BI, R.x+BI, R.y+IS-1-BI-1);
}

void DesignWindow::onTitleChanged() {
	if(get_window())
		get_window()->invalidate_rect(getCaptionRect(get_allocation()), false);
}

void DesignWindow::onIconNameChanged() {
	loadIcon();
	if(get_window())
		get_window()->invalidate_rect(getButtonRect(get_allocation(), 0), false);
}

void DesignWindow::loadIcon() {
	icon= PGdkPixbuf();
	if(!iconName.empty())
		try {
			icon= Gtk::IconTheme::get_default()->load_icon(iconName, IS, Gtk::ICON_LOOKUP_FORCE_SIZE);
		} catch(const Glib::Error &) {
		}
	if(!icon)
		try {
			icon= Gtk::IconTheme::get_default()->load_icon("gtk-missing-image", IS, Gtk::ICON_LOOKUP_FORCE_SIZE);
		} catch(const Glib::Error &) {
		}
}

void DesignWindow::updatePadding() {
	if(bordering)
		set_padding(2+HS+BS, 2+BS, 2+BS, 2+BS);
	else
		set_padding(0, 0, 0, 0);
}

// Containment -----------------------------------------------------------------

Containment::Containment() {
	add(enclosure);
	enclosure.show();
}

void Containment::setWidget(Gtk::Widget * W) {
	Gtk::Window * window= dynamic_cast<Gtk::Window *>(W);
	setBordering(window != NULL);
	enclosure.setWidget(W);
	if(window) {
		onTitleChanged();
		onIconNameChanged();
		s1= Glib::SignalProxyProperty(window, "title").connect(sigc::mem_fun(*this, &Containment::onTitleChanged));
		s2= Glib::SignalProxyProperty(window, "icon-name").connect(sigc::mem_fun(*this, &Containment::onIconNameChanged));
	}
}

void Containment::unsetWidget() {
	s1.disconnect();
	s2.disconnect();
	onTitleChanged();
	onIconNameChanged();
	enclosure.unsetWidget();
	setBordering(false);
}

Rectangle Containment::getShift() {
	guint t, b, l, r;
	get_padding(t, b, l, r);
	return Rectangle(-SIGNED(l), -SIGNED(t), l+r, t+b);
}

void Containment::onTitleChanged() {
	Glib::ustring title;
	if(Gtk::Window * window= dynamic_cast<Gtk::Window *>(getWidget()))
		title= window->get_title();
	setTitle(title);
}

void Containment::onIconNameChanged() {
	Glib::ustring iconName;
	if(Gtk::Window * window= dynamic_cast<Gtk::Window *>(getWidget())) {
		const char * retval= gtk_window_get_icon_name(window->gobj());
		iconName= retval ? retval : "";
	}
	setIconName(iconName);
}

// Button ----------------------------------------------------------------------

Button::Button() {
	std::cout << "Button::Button() - " << GetRefCount(this) << std::endl;
	SetData(this, "CrowTypeHint", "CrowButton");
}

Button::~Button() {
	std::cout << "Button::~Button() - " << GetRefCount(this) << std::endl;
}

// Custom ----------------------------------------------------------------------

Custom::Custom() {
	SetData(this, "CrowTypeHint", "CrowCustom");
}

bool Custom::on_expose_event(GdkEventExpose * event) {
	bool result= Gtk::DrawingArea::on_expose_event(event);
	DrawGrid(get_window(), get_style(), true);
	return result;
}

void Custom::on_size_request(Gtk::Requisition * requisition) {
	Gtk::DrawingArea::on_size_request(requisition);
	requisition->width= 40;
	requisition->height= 40;
}

// ArrowButton -----------------------------------------------------------------

ArrowButton::ArrowButton() :
	arrow(Gtk::ARROW_LEFT, Gtk::SHADOW_NONE)
{
	add(arrow);
	arrow.show();
}

// EditorWidget ----------------------------------------------------------------

EditorWidget::EditorWidget() :
	Glib::ObjectBase(typeid(EditorWidget))
{
	frameNeeded= false;
}

PEditor EditorWidget::getEditor() {
	return editor;
}

void EditorWidget::setEditor(PEditor e) {
	editor= e;
}

void EditorWidget::unsetEditor() {
	editor= PEditor();
}

// HierarchyEditorWidget -------------------------------------------------------

HierarchyEditorWidget::HierarchyEditorWidget() :
	Glib::ObjectBase(typeid(HierarchyEditorWidget))
{

	scrolledWindow.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_ALWAYS);
	add(scrolledWindow);
	scrolledWindow.show();

	treeView.get_selection()->set_mode(Gtk::SELECTION_MULTIPLE);
	treeView.setImmediateEditing(false);
	scrolledWindow.add(treeView);
	treeView.show();

	treeView.createModel(2);
	treeView.addColumn(0, _("Class"));
	treeView.addColumn(1, _("Name"));

	treeView.get_column_cell_renderer(0)->property_ypad()= 0;
	treeView.get_column_cell_renderer(1)->property_ypad()= 0;

}

int HierarchyEditorWidget::getSplit() {
	return treeView.getColumnWidth(0);
}

void HierarchyEditorWidget::setSplit(int split) {
	treeView.setColumnWidth(0, split);
}

void HierarchyEditorWidget::collapse() {
	treeView.collapse_all();
}

// PropertyEditorWidget --------------------------------------------------------

PropertyEditorWidget::PropertyEditorWidget() {
	setAllowEditing(true);
}

bool PropertyEditorWidget::isEditable() {
	return getAllowEditing() && getEditor() && getEditor()->getSession()->canEdit();
}

// PropertyEditorWidgetText ----------------------------------------------------

PropertyEditorWidgetText::PropertyEditorWidgetText() :
	Glib::ObjectBase(typeid(PropertyEditorWidgetText))
{
	setAllowEditing(false);
}

// PropertyEditorWidgetInput ---------------------------------------------------

PropertyEditorWidgetInput::PropertyEditorWidgetInput() :
	Glib::ObjectBase(typeid(PropertyEditorWidgetInput))
{
}

// PropertyEditorWidgetChoice --------------------------------------------------

PropertyEditorWidgetChoice::PropertyEditorWidgetChoice() :
	Glib::ObjectBase(typeid(PropertyEditorWidgetChoice))
{
}

// ColorPropertyEditorWidget ---------------------------------------------------

ColorPropertyEditorWidget::ColorPropertyEditorWidget() :
	Glib::ObjectBase(typeid(ColorPropertyEditorWidget))
{
	add(hBox);
	hBox.show();

	hBox.pack_start(drawing, Gtk::PACK_EXPAND_WIDGET);
	drawing.signal_expose_event().connect(sigc::mem_fun(*this, &ColorPropertyEditorWidget::onDrawingExpose));
	drawing.show();

	hBox.pack_start(button, Gtk::PACK_SHRINK);
	button.signal_clicked().connect(sigc::mem_fun(*this, &ColorPropertyEditorWidget::onButtonClicked));
	button.show();
}

void ColorPropertyEditorWidget::setColor(const Gdk::Color & color) {
	this->color= color;
	setFuzzy(false);
	if(isEditing())
		drawing.queue_draw();
}

bool ColorPropertyEditorWidget::onDrawingExpose(GdkEventExpose * event) {
	Gtk::CellRenderer * renderer= getTreeView()->findViewColumn(getModelColumn())->get_first_cell_renderer();
	Rectangle R= drawing.get_allocation();
	R.x= R.y= 0;
	getTreeView()->get_style()->paint_flat_box(
		drawing.get_window(), Gtk::STATE_NORMAL, Gtk::SHADOW_NONE, R, *getTreeView(), "cell_odd",
		R.x, R.y, R.w, R.h
	);
	R.grow(
		renderer->property_xpad(), renderer->property_ypad(),
		-2*int(renderer->property_xpad()), -2*int(renderer->property_ypad())
	);
	paint(drawing.get_window(), R);
	return true;
}

void ColorPropertyEditorWidget::onButtonClicked() {
	Gtk::ColorSelectionDialog dialog;
	dialog.get_colorsel()->set_current_color(color);
	if(getEditor()->getController()->getContext()->runDialog(&dialog)==Gtk::RESPONSE_OK) {
		color= dialog.get_colorsel()->get_current_color();
		accept();
	}
}

void ColorPropertyEditorWidget::onEditingStarted() {
	button.grab_focus();
}

void ColorPropertyEditorWidget::cellRender(
	Gtk::Widget * widget,
	PGdkWindow window,
	const Gdk::Rectangle & rect,
	Gtk::CellRendererState flags
) {
	paint(window, rect);
}

void ColorPropertyEditorWidget::paint(PGdkWindow window, const Gdk::Rectangle & rect) {
	PGdkGC gc= Gdk::GC::create(window);
	Rectangle R= rect;
	R.grow(2, 2, -4, -4);
	int PX= 10;
	int PY= 3;
	R.grow(PX, PY, -2*PX, -2*PY);
	if(!R.empty()) {
		gc->set_foreground(getTreeView()->get_style()->get_dark(Gtk::STATE_NORMAL));
		window->draw_rectangle(gc, false, R.x-2, R.y-2, R.w+3, R.h+3);
		gc->set_foreground(getTreeView()->get_style()->get_light(Gtk::STATE_NORMAL));
		window->draw_rectangle(gc, false, R.x-1, R.y-1, R.w+1, R.h+1);
		gc->set_rgb_fg_color(color);
		window->draw_rectangle(gc, true, R.x, R.y, R.w, R.h);
	}
}

}
