/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#ifndef CROW_WIDGETS_H
#define CROW_WIDGETS_H

#include "polytree.h"

namespace Crow {

class HierarchyWidget : public Gtk::Frame {

public:
	HierarchyWidget();

	HierarchyEditorWidget * getEditorWidget();
	void setEditorWidget(EditorWidget *);
	void unsetEditorWidget();

	int getSplit();
	void setSplit(int split);

};

enum PaletteEvent { peRoot, peCanvas, peCancel };

class PaletteWidget : public Gtk::Frame {

public:
	PaletteWidget();

	void setShowDW(bool show_dw);
	void unpush();

	typedef sigc::signal<void, TypeId, PaletteEvent> SignalPaletteEvent;
	sigc::connection connect_palette_event(SignalPaletteEvent::slot_type slot) { return signalPaletteEvent.connect(slot); }

protected:
	void create();
	Gtk::Widget * createPaletteButton(EntityEntry * entry);
	void onEntryClicked(TypeId type, Gtk::ToggleButton * tbutton);

private:
	Gtk::ScrolledWindow scroll;
	Gtk::VBox vBox;
	SignalPaletteEvent signalPaletteEvent;
	Gtk::ToggleButton * pushed;
	Gtk::Widget * debugTab;

};

class CanvasWidget : public Gtk::Frame {

public:
	CanvasWidget();

	void setEditorWidget(EditorWidget *);
	void unsetEditorWidget();

};

class PropertyTree : public Polytree {

public:
	PropertyTree();

	int getSplit();
	void setSplit(int split);

};

class ExplorerWidget : public Gtk::Frame {

public:
	ExplorerWidget();

	enum ExplorerPage { epProperties, epSignals, epPacking };

	PropertyTree * getTreeView(ExplorerPage page);

	int getSplit();
	void setSplit(int split);

	PSession getSelectedRow();
	void setSelectedRow(PSession session);

	ExplorerPage getCurrentPage();
	PropertyTree * getCurrentTree();

	typedef sigc::signal<void> SignalSelectionChanged;
	sigc::connection connect_selection_changed(SignalSelectionChanged::slot_type slot) { return signalSelectionChanged.connect(slot); }

protected:
	void addPage(string_cr tab);

	void onSelectionChanged(ExplorerPage page);
	void onSwitchPage(GtkNotebookPage *, guint);

private:
	Gtk::Notebook notebook;
	SignalSelectionChanged signalSelectionChanged;
	int prev_page;

};

class SignalSelectDialog : public Gtk::Dialog {

public:
	SignalSelectDialog();

	void setChoices(const SignalsQuery & signals);
	std::string getSelection() { return selection; }

protected:
	void addTreeView(GType gtype, const std::vector<std::string> & names);
	void onSelectionChanged(Gtk::TreeView * treeView);

private:
	bool lock;
	std::string selection;

	Gtk::ScrolledWindow scroll;
	Gtk::VBox vBox;

	Gtk::Button * cancelButton;
	Gtk::Button * okButton;

};

class TextInputDialog : public Gtk::Dialog {

public:
	TextInputDialog();

	Glib::ustring getText();
	void setText(ustring_cr text);

	Translatable getMeta();
	void setMeta(const Translatable & trans);

private:
	Gtk::VPaned paned;
	Gtk::ScrolledWindow scrollText;
	Gtk::TextView text;
	Gtk::Frame frame;
	Gtk::CheckButton check;
	Gtk::Alignment align;
	Gtk::VBox box1;
	Gtk::HBox box2;
	Gtk::VBox box3;
	Gtk::Label label1;
	Gtk::Entry entry;
	Gtk::Label label2;
	Gtk::ScrolledWindow scrollComments;
	Gtk::TextView comments;

};

class IconSelectDialog : public Gtk::Dialog {

public:
	IconSelectDialog();

	void setChoices(const std::vector<std::pair<PGdkPixbuf, Glib::ustring> > & choices);

	Glib::ustring getChoice();
	void setChoice(ustring_cr name);

protected:
	void onSelectionChanged();

private:
	Gtk::ScrolledWindow scroll;
	Gtk::TreeView treeView;

	Gtk::Button * cancelButton;
	Gtk::Button * okButton;

};

class FlagsDialog : public Gtk::Dialog {

public:
	FlagsDialog();

	typedef std::list<std::pair<std::string, Bits> > Values;

	void setOptions(const Values & options);
	Bits getValue();
	void setValue(Bits value);

private:
	Gtk::ScrolledWindow scroll;
	Gtk::TreeView treeView;

};

class TranslatablesDialog : public Gtk::Dialog {

public:
	TranslatablesDialog();

	struct String {
		PNode node;
		IdPath path;
		Glib::ustring string;
		Glib::ustring defval;
		Translatable meta;
		bool operator < (const String & s) const { return path < s.path; }
	};

	typedef std::vector<String> Strings;

	void setStrings(const Strings & strings);
	Strings getStrings();

protected:
	void setupColumn(int column, int width);
	void onInvert();

private:
	Gtk::Button invertButton;
	Gtk::ScrolledWindow scroll;
	Gtk::TreeView treeView;

	Strings set;

};

class Enclosure : public Gtk::Alignment {

public:
	Enclosure() { window= NULL; }

	void setWidget(Gtk::Widget * W);
	void unsetWidget();
	Gtk::Widget * getWidget();

private:
	static void onRealize(GtkWidget * widget);
	static void onSizeAllocate(GtkWidget * widget, GtkAllocation * allocation);
	static void onSizeRequest(GtkWidget * widget, GtkRequisition * requisition);
	static void encloseWindow(GtkWidget * widget);
	static void unlooseWindow(GtkWidget * widget);

private:
	Gtk::Window * window;

};

class DesignWindow : public Gtk::Alignment {

public:
	DesignWindow();

	void setBordering(bool b);
	void setTitle(ustring_cr t);
	void setIconName(ustring_cr in);

protected:
	bool on_expose_event(GdkEventExpose * e);
	void on_size_request(Gtk::Requisition * requisition);

	PPangoLayout createLayout(ustring_cr text, int width);
	Rectangle getCaptionRect(const Rectangle & A);
	Rectangle getButtonRect(const Rectangle & A, int pos);

	void drawIcon(PGdkGC gc, const Rectangle & R);
	void drawMinimize(PGdkGC gc, const Rectangle & R);
	void drawMaximize(PGdkGC gc, const Rectangle & R);
	void drawClose(PGdkGC gc, const Rectangle & R);

	void onTitleChanged();
	void onIconNameChanged();
	void loadIcon();
	void updatePadding();

private:
	bool bordering;
	int CS, IS, HS, BS, BI;
	int RW, RH;
	Glib::ustring title;
	Glib::ustring iconName;
	PGdkPixbuf icon;

};

class Containment : public DesignWindow {

public:
	Containment();
	Gtk::Widget * getEnclosure() { return &enclosure; }

	void setWidget(Gtk::Widget * W);
	void unsetWidget();
	Gtk::Widget * getWidget() { return enclosure.getWidget(); }

	Rectangle getShift();

protected:
	void onTitleChanged();
	void onIconNameChanged();

private:
	Enclosure enclosure;
	sigc::connection s1, s2;

};

class Button : public Gtk::FileChooserButton {

public:
	Button();
	~Button();

};

class Custom : public Gtk::DrawingArea {

public:
	Custom();

protected:
	bool on_expose_event(GdkEventExpose * event);
	void on_size_request(Gtk::Requisition * requisition);

};

class ArrowButton : public Gtk::Button {

public:
	ArrowButton();

private:
	Gtk::Arrow arrow;

};

class EditorWidget : public virtual CustomAlignment {

public:
	EditorWidget();
	PEditor getEditor();
	void setEditor(PEditor editor);
	void unsetEditor();

	bool isFrameNeeded() { return frameNeeded; }

protected:
	void setFrameNeeded(bool fn) { frameNeeded= fn; }

private:
	PEditor editor;
	bool frameNeeded;

};

class HierarchyEditorWidget : public EditorWidget {

public:
	HierarchyEditorWidget();

	Polytree * getTreeView() { return &treeView; }

	int getSplit();
	void setSplit(int split);

	void collapse();

private:
	Gtk::ScrolledWindow scrolledWindow;
	Polytree treeView;

};

class PropertyEditorWidget : public EditorWidget, public virtual Polycell {

public:
	PropertyEditorWidget();
	bool isEditable();

};

class PropertyEditorWidgetText : public PropertyEditorWidget, public PolycellText {

public:
	PropertyEditorWidgetText();

};

class PropertyEditorWidgetInput : public PropertyEditorWidget, public PolycellInput {

public:
	PropertyEditorWidgetInput();

};

class PropertyEditorWidgetChoice : public PropertyEditorWidget, public PolycellChoice {

public:
	PropertyEditorWidgetChoice();

};

class ColorPropertyEditorWidget : public PropertyEditorWidget, public PolycellWidget {

public:
	ColorPropertyEditorWidget();

	const Gdk::Color * getColor() { return &color; }
	void setColor(const Gdk::Color & color);

protected:
	bool onDrawingExpose(GdkEventExpose * event);
	void onButtonClicked();

	void onEditingStarted();

	void cellRender(
		Gtk::Widget * widget,
		PGdkWindow window,
		const Gdk::Rectangle & rect,
		Gtk::CellRendererState flags
	);

	void paint(PGdkWindow window, const Gdk::Rectangle & rect);

private:
	Gtk::HBox hBox;
	Gtk::DrawingArea drawing;
	ArrowButton button;
	Gdk::Color color;

};

}

#endif
