/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "application.h"

namespace CrowApp {

// ErrorDialog -----------------------------------------------------------------

ErrorDialog::ErrorDialog() {
	set_title(_("GTK+ Messages"));
	set_default_size(400, 200);
	set_type_hint(Gdk::WINDOW_TYPE_HINT_DIALOG);
	set_position(Gtk::WIN_POS_CENTER_ON_PARENT);

	scroll.set_policy(Gtk::POLICY_NEVER, Gtk::POLICY_ALWAYS);
	add(scroll);
	scroll.show();

	textView.set_editable(false);
	textView.set_cursor_visible(false);
	textView.set_wrap_mode(Gtk::WRAP_WORD);
	scroll.add(textView);
	textView.show();
}

void ErrorDialog::onErrorMessage(ustring_cr message) {
	Glib::RefPtr<Gtk::TextBuffer> buffer= textView.get_buffer();
	if(!get_visible())
		buffer->set_text("");
	buffer->place_cursor(buffer->end());
	buffer->insert_at_cursor(message);
	textView.scroll_to(buffer->get_insert());
	show();
}

// SettingsDialog --------------------------------------------------------------

SettingsDialog::SettingsDialog() {

	SetKeyColor(&defaults, "wce_vc", Gdk::Color("forest green"));
	SetKeyColor(&defaults, "wce_wc", Gdk::Color("indian red"));
	SetKeyBool(&defaults, "show_dw", false);
	SetKeyBool(&defaults, "show_hp", false);
	SetKeyBool(&defaults, "supp_msg", false);

	loader= Gui::Loader::create();

	GUIXML_DECL
		#include "settings.gui.h"
	GUIXML_LOAD_CPP(loader)

	loader->bind_object("window", &window);
	loader->bind_object("widget", &widget);
	loader->bind_object("validColor", &validColor);
	loader->bind_object("wrongColor", &wrongColor);
	loader->bind_object("showDebugWidgets", &showDebugWidgets);
	loader->bind_object("showHiddenProperties", &showHiddenProperties);
	loader->bind_object("suppressMessages", &suppressMessages);
	loader->bind_object("actions", &actions);
	loader->bind_object("folders", &folders);

	loader->connect_signals("onAdd", sigc::mem_fun(*this, &SettingsDialog::onAdd));
	loader->connect_signals("onRemove", sigc::mem_fun(*this, &SettingsDialog::onRemove));
	loader->connect_signals("onMoveUp", sigc::mem_fun(*this, &SettingsDialog::onMoveUp));
	loader->connect_signals("onMoveDown", sigc::mem_fun(*this, &SettingsDialog::onMoveDown));
	loader->connect_signals("onColorChanged", sigc::mem_fun(*this, &SettingsDialog::onColorChanged), 2);
	loader->connect_signals("onCheckChanged", sigc::mem_fun(*this, &SettingsDialog::onCheckChanged), 3);
	loader->connect_signals("onSwitchPage", sigc::mem_fun(*this, &SettingsDialog::onSwitchPage));
	loader->connect_signals("onReset", sigc::mem_fun(*this, &SettingsDialog::onReset));

	Gtk::TreeModelColumn<Glib::ustring> column;
	Gtk::TreeModel::ColumnRecord columns;
	columns.add(column);
	model= Gtk::ListStore::create(columns);

	folders->set_model(model);
	folders->append_column("", column);

	folders->get_selection()->signal_changed().connect(sigc::mem_fun(*this, &SettingsDialog::onSelectionChanged));

}

void SettingsDialog::setSettings(const Settings & settings) {
	original= settings;
	original.insert(defaults.begin(), defaults.end());
	original["icon_folders"];
	updateUi(original);
}

Settings SettingsDialog::getSettings() {
	Settings current= original;

	SetKeyColor(&current, "wce_vc", validColor->get_color());
	SetKeyColor(&current, "wce_wc", wrongColor->get_color());
	SetKeyBool(&current, "show_dw", showDebugWidgets->get_active());
	SetKeyBool(&current, "show_hp", showHiddenProperties->get_active());
	SetKeyBool(&current, "supp_msg", suppressMessages->get_active());

	Glib::ustring vec;
	Gtk::TreeModel::Children rows= model->children();
	for(Gtk::TreeModel::Children::iterator it= rows.begin(); it!=rows.end(); ++it) {
		Glib::ustring str;
		it->get_value(0, str);
		vec += str + '*';
	}
	if(!vec.empty())
		vec.resize(vec.size()-1);
	SetKeyString(&current, "icon_folders", vec.c_str());

	return current;
}

void SettingsDialog::onAdd() {
	Gtk::FileChooserDialog dialog(*window, _("File Chooser"), Gtk::FILE_CHOOSER_ACTION_SELECT_FOLDER);
	dialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	dialog.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);
	dialog.set_current_folder(Glib::get_home_dir());
	if(dialog.run() == Gtk::RESPONSE_OK) {
		Glib::ustring folder= Glib::filename_to_utf8(dialog.get_filename());
		Gtk::TreeRow row= *model->insert_after(getPosition());
		row.set_value(0, folder);
		folders->get_selection()->select(row);
		actionCompleted();
	}
}

void SettingsDialog::onRemove() {
	Gtk::TreeIter it= getPosition();
	Gtk::TreeIter sel= it;
	++sel;
	model->erase(it);
	if(!sel && !model->children().empty())
		--sel;
	if(sel)
		folders->get_selection()->select(sel);
	actionCompleted();
}

void SettingsDialog::onMoveUp() {
	Gtk::TreeIter it= getPosition();
	Gtk::TreeIter dest= it;
	--dest;
	model->move(it, dest);
	actionCompleted();
	updateActions();
}

void SettingsDialog::onMoveDown() {
	Gtk::TreeIter it= getPosition();
	Gtk::TreeIter dest= it;
	++dest;
	++dest;
	model->move(it, dest);
	actionCompleted();
	updateActions();
}

void SettingsDialog::onColorChanged() {
	updateButtons();
}

void SettingsDialog::onCheckChanged() {
	updateButtons();
}

void SettingsDialog::onSelectionChanged() {
	updateActions();
}

void SettingsDialog::onSwitchPage(GtkNotebookPage *, guint p) {
	updateButtons();
	updateActions();
}

void SettingsDialog::onReset() {
	Settings settings= getSettings();
	for(Settings::iterator it= defaults.begin(); it!=defaults.end(); ++it)
		settings[it->first]= it->second;
	updateUi(settings);
}

void SettingsDialog::updateUi(const Settings & settings) {
	validColor->set_color(GetKeyColor(settings, "wce_vc"));
	wrongColor->set_color(GetKeyColor(settings, "wce_wc"));
	showDebugWidgets->set_active(GetKeyBool(settings, "show_dw"));
	showHiddenProperties->set_active(GetKeyBool(settings, "show_hp"));
	suppressMessages->set_active(GetKeyBool(settings, "supp_msg"));

	StringList vec= Split(GetKeyString(settings, "icon_folders"), "*");
	model->clear();
	for(StringList::iterator it=vec.begin(); it!=vec.end(); ++it) {
		Gtk::TreeModel::iterator iter= model->append();
		iter->set_value(0, *it);
	}

	updateButtons();
	updateActions();
}

void SettingsDialog::updateButtons() {
	Settings current= getSettings();
	window->set_response_sensitive(Gtk::RESPONSE_OK, original != current);
	bool canReset= false;
	if(widget->get_current_page()==0)
		for(Settings::iterator it= defaults.begin(); it!=defaults.end(); ++it)
			if(current[it->first]!=it->second) {
				canReset= true;
				break;
			}
	actions->get_action("Reset")->set_sensitive(canReset);
}

void SettingsDialog::updateActions() {
	Gtk::TreeIter it= getPosition();
	Gtk::TreeIter next;
	if(it) {
		next= it;
		++next;
	}
	bool page= widget->get_current_page() == 1;
	actions->get_action("Add")->set_sensitive(page);
	actions->get_action("Remove")->set_sensitive(page && it);
	actions->get_action("MoveUp")->set_sensitive(page && it && it!=model->children().begin());
	actions->get_action("MoveDown")->set_sensitive(page && it && next!=model->children().end());
}

Gtk::TreeIter SettingsDialog::getPosition() {
	if(folders->get_selection()->count_selected_rows() == 0)
		return model->children().end();
	else
		return folders->get_selection()->get_selected();
}

void SettingsDialog::actionCompleted() {
	if(folders->get_selection()->count_selected_rows() != 0) {
		Gtk::TreeIter it= folders->get_selection()->get_selected();
		folders->scroll_to_row(model->get_path(it));
	}
	updateButtons();
}

// AppWindow -------------------------------------------------------------------

AppWindow::State AppWindow::getState() {
	State state;
	get_size(state.width, state.height);
	state.maximized= getMaximized();
	return state;
}

void AppWindow::setState(const State & state) {
	deiconify();
	if(state.maximized==1) {
		unfullscreen();
		maximize();
	} else if(state.maximized==2) {
		unmaximize();
		fullscreen();
	} else {
		unfullscreen();
		unmaximize();
	}
	resize(state.width, state.height);
}

int AppWindow::getMaximized() {
	Gdk::WindowState state= getGdkWindow()->get_state();
	if(state & Gdk::WINDOW_STATE_MAXIMIZED)
		return 1;
	else if(state & Gdk::WINDOW_STATE_FULLSCREEN)
		return 2;
	else
		return 0;
}

Glib::RefPtr<Gdk::Window> AppWindow::getGdkWindow() {
	Glib::RefPtr<Gdk::Window> window= get_window();
	CHECK(window);
	return window;
}

// Application -----------------------------------------------------------------

Application::Application(const std::vector<std::string> & args1, const std::vector<std::string> & args2) {

	{
		#include "icons.h"
		Gtk::IconTheme::add_builtin_icon("crow-icon", 16, Gdk::Pixbuf::create_from_inline(sizeof(crow16), crow16));
		Gtk::IconTheme::add_builtin_icon("crow-icon", 32, Gdk::Pixbuf::create_from_inline(sizeof(crow32), crow32));
		Gtk::IconTheme::add_builtin_icon("crow-icon", 48, Gdk::Pixbuf::create_from_inline(sizeof(crow48), crow48));
		gtk_window_set_icon_name(window.gobj(), "crow-icon");
	}

	defaultIconPath= currentIconPath();

	errorDialog.set_transient_for(window);

	setupHandler("GLib");
	setupHandler("GLib-GObject");
	setupHandler("Glib-GIO");
	setupHandler("Gdk");
	setupHandler("GdkPixbuf");
	setupHandler("Pango");
	setupHandler("Gtk");
	setupHandler("GuiLoader");
	setupHandler("glibmm");
	setupHandler("gdkmm");
	setupHandler("pangomm");
	setupHandler("gtkmm");
	setupHandler("Crow");

	Glib::add_exception_handler(sigc::mem_fun(*this, &Application::onException));

	startupArgs1= args1;
	startupArgs2= args2;
	startupDirectory= Glib::get_current_dir();

	loadSettings();
	AppWindow::State state;
	state.width= GetKeyInt(settings, "$window_width");
	state.height= GetKeyInt(settings, "$window_height");
	state.maximized= GetKeyInt(settings, "$window_maximized");
	window.setState(state);
	window.signal_delete_event().connect(sigc::mem_fun(*this, &Application::onDeleteEvent));

	createActionGroup();
	createUIManager();

	window.add(vBox);
	vBox.show();

	Gtk::MenuBar * menuBar= dynamic_cast<Gtk::MenuBar *>(uiManager->get_widget("/MainMenubar"));
	vBox.pack_start(*menuBar, Gtk::PACK_SHRINK);

	Gtk::Toolbar * toolbar= dynamic_cast<Gtk::Toolbar *>(uiManager->get_widget("/MainToolbar"));
	toolbar->set_toolbar_style(Gtk::TOOLBAR_ICONS);
	vBox.pack_start(*toolbar, Gtk::PACK_SHRINK);

	statusbar.push("");
	vBox.pack_end(statusbar, Gtk::PACK_SHRINK);
	statusbar.show();

	statusbar.pack_end(infoFrame, Gtk::PACK_SHRINK);

	infoLabel.set_text("");
	infoFrame.add(infoLabel);
	infoLabel.show();

	designer= CreateDesigner();
	designer->connect_actions_updated(sigc::mem_fun(*this, &Application::onActionsUpdated));
	designer->connect_status_updated(sigc::mem_fun(*this, &Application::onStatusUpdated));
	designer->connect_info_updated(sigc::mem_fun(*this, &Application::onInfoUpdated));
	designer->connect_show_url(sigc::mem_fun(*this, &Application::onShowUrl));

	Gtk::Widget * dwidget= designer->get_widget();
	vBox.pack_start(*dwidget, Gtk::PACK_EXPAND_WIDGET);

	Glib::RefPtr<Gtk::RecentAction> recent= Glib::RefPtr<Gtk::RecentAction>::cast_dynamic(actionGroup->get_action("Recent"));
	recent->signal_item_activated().connect(sigc::mem_fun(*this, &Application::onOpenRecent));

	designer->apply_settings(settings);
	window.add_accel_group(uiManager->get_accel_group());

	updateTitle();

	if(args2.size()==2)
		openFile(File(args2[1]));

}

Application::~Application() {
	designer->store_settings(&settings);
	saveSettings(true);
	for(Hids::iterator it=hids.begin(); it!=hids.end(); ++it)
		g_log_remove_handler(it->first, it->second);
}

void Application::setupHandler(const char * domain) {
	int hid= g_log_set_handler(domain,
		GLogLevelFlags(
			G_LOG_LEVEL_CRITICAL | G_LOG_LEVEL_WARNING |
			G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION
		),
		&Application::logHandler, this
	);
	hids.push_back(std::make_pair(domain, hid));
}

void Application::logHandler(
	const char * logDomain,
	GLogLevelFlags logLevel,
	const char * message,
	gpointer userData
) {
	static_cast<Application *>(userData)->onErrorMessage(logDomain, logLevel, message);
}

void Application::onErrorMessage(
	ustring_cr domain,
	GLogLevelFlags level,
	ustring_cr message
) {
	Glib::ustring errstr= "[" + domain;
	if(level & G_LOG_LEVEL_CRITICAL)
		errstr += " CRITICAL";
	if(level & G_LOG_LEVEL_WARNING)
		errstr += " WARNING";
	errstr += "] " + message + "\n\n";
	std::cerr << errstr << std::flush;
	if(!GetKeyBool(settings, "supp_msg"))
		errorDialog.onErrorMessage(errstr);
}

void Application::createActionGroup() {

	Glib::RefPtr<Gtk::RecentAction> recent= Gtk::RecentAction::create("Recent", _("Open _Recent"));
	{
		Gtk::RecentFilter filter;
		filter.add_pattern("*.gui");
		filter.add_pattern("*.gui.h");
		recent->set_sort_type(Gtk::RECENT_SORT_MRU);
		recent->add_filter(filter);
		recent->set_show_icons(false);
		recent->set_show_tips(true);
		recent->set_property("show-numbers", true);
	}

	Glib::RefPtr<Gtk::ActionGroup> actions= Gtk::ActionGroup::create();

	actions->add(Gtk::Action::create("Save", Gtk::Stock::SAVE, _("_Save"), _("Save the current file")), sigc::mem_fun(*this, &Application::onSave));

	actions->add(Gtk::Action::create("Undo", Gtk::Stock::UNDO, _("_Undo"), _("Undo the last action")), Gtk::AccelKey("<control>Z"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_UNDO));
	actions->add(Gtk::Action::create("Redo", Gtk::Stock::REDO, _("_Redo"), _("Redo the undone action")), Gtk::AccelKey("<shift><control>Z"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_REDO));
	actions->add(Gtk::Action::create("Cut", Gtk::Stock::CUT, _("Cu_t"), _("Cut the selection")), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_CUT));
	actions->add(Gtk::Action::create("Copy", Gtk::Stock::COPY, _("_Copy"), _("Copy the selection")), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_COPY));
	actions->add(Gtk::Action::create("Paste", Gtk::Stock::PASTE, _("_Paste..."), _("Paste the clipboard")), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_PASTE));
	actions->add(Gtk::Action::create("PasteToRoot", _("Paste to _Root")), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_PASTE_TO_ROOT));
	actions->add(Gtk::Action::create("Delete", Gtk::Stock::DELETE, _("_Delete"), _("Delete the selection")), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_DELETE));

	actions->add(Gtk::Action::create("About", Gtk::Stock::INFO, _("_About")), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_ABOUT));

	actions->add(Gtk::Action::create("Preview", Gtk::Stock::EXECUTE, _("_Preview..."), _("Show widgets in a preview window")), Gtk::AccelKey("<control>T"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_PREVIEW));
	actions->add(Gtk::Action::create("DefaultSize", Gtk::Stock::ZOOM_100, _("D_efault Size"), _("Restore default widget size")), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_DEFAULT_SIZE));
	actions->add(Gtk::Action::create("ShiftLeft", Gtk::Stock::GO_BACK, _("Shift _Left"), _("Shift the widget left")), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_SHIFT_LEFT));
	actions->add(Gtk::Action::create("ShiftRight", Gtk::Stock::GO_FORWARD, _("Shift _Right"), _("Shift the widget right")), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_SHIFT_RIGHT));
	actions->add(Gtk::Action::create("Translatables", Gtk::Stock::EDIT, _("_Strings..."), _("Manage translatable strings")), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_TRANSLATABLES));

	actions->add(Gtk::Action::create("Default", Gtk::Stock::CLEAR, _("_Default"), _("Set the property to its default value")), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_DEFAULT));
	actions->add(Gtk::Action::create("Add", Gtk::Stock::ADD, _("_Add"), _("Add a New Element")), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_ADD));
	actions->add(Gtk::Action::create("Remove", Gtk::Stock::REMOVE, _("Re_move"), _("Remove the element")), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_REMOVE));
	actions->add(Gtk::Action::create("MoveUp", Gtk::Stock::GO_UP, _("Move _Up"), _("Move the element up")), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_MOVE_UP));
	actions->add(Gtk::Action::create("MoveDown", Gtk::Stock::GO_DOWN, _("Move Do_wn"), _("Move the element down")), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_MOVE_DOWN));

	{
		std::vector<Glib::RefPtr<Gtk::Action> > V= actions->get_actions();
		for(int i=0; i<SIGNED(V.size()); ++i)
			V[i]->set_sensitive(false);
	}

	actions->add(Gtk::Action::create("File", _("_File")));
	actions->add(Gtk::Action::create("Clear", Gtk::Stock::NEW, _("_New"), _("Start a new file")), sigc::mem_fun(*this, &Application::onClear));
	actions->add(Gtk::Action::create("Clone", _("New _Window")), sigc::mem_fun(*this, &Application::onClone));
	actions->add(Gtk::Action::create("Open", Gtk::Stock::OPEN, _("_Open"), _("Open a file")), sigc::mem_fun(*this, &Application::onOpen));
	actions->add(recent);
	actions->add(Gtk::Action::create("SaveAs", _("Save _As...")), sigc::mem_fun(*this, &Application::onSaveAs));
	actions->add(Gtk::Action::create("Quit", Gtk::Stock::QUIT, _("_Quit")), sigc::mem_fun(*this, &Application::onQuit));

	actions->add(Gtk::Action::create("Edit", _("_Edit")));
	actions->add(Gtk::Action::create("Settings", Gtk::Stock::PREFERENCES, _("_Settings...")), sigc::mem_fun(*this, &Application::onSettings));
	actions->add(Gtk::Action::create("Help", _("_Help")));
	actions->add(Gtk::Action::create("Home", Gtk::Stock::HOME, _("_Home")), Gtk::AccelKey("<control>H"), sigc::mem_fun(*this, &Application::onHome));
	actions->add(Gtk::Action::create("API", Gtk::Stock::HELP, _("API _Reference")), Gtk::AccelKey("F1"), sigc::mem_fun(*this, &Application::onApi));
	actions->add(Gtk::Action::create("Actions", _("_Actions")));

	actionGroup= actions;
}

void Application::createUIManager() {

	Glib::RefPtr<Gtk::UIManager> manager= Gtk::UIManager::create();
	manager->insert_action_group(actionGroup);

	Glib::ustring uistr=
		"<ui>"
		"  <menubar name='MainMenubar'>"
		"    <menu action='File'>"
		"      <menuitem action='Clear'/>"
		"      <menuitem action='Clone'/>"
		"      <separator/>"
		"      <menuitem action='Open'/>"
		"      <menuitem action='Recent'/>"
		"      <separator/>"
		"      <menuitem action='Save'/>"
		"      <menuitem action='SaveAs'/>"
		"      <separator/>"
		"      <menuitem action='Quit'/>"
		"    </menu>"
		"    <menu action='Edit'>"
		"      <menuitem action='Undo'/>"
		"      <menuitem action='Redo'/>"
		"      <separator/>"
		"      <menuitem action='Cut'/>"
		"      <menuitem action='Copy'/>"
		"      <menuitem action='Paste'/>"
		"      <menuitem action='PasteToRoot'/>"
		"      <menuitem action='Delete'/>"
		"      <separator/>"
		"      <menuitem action='Settings'/>"
		"    </menu>"
		"    <menu action='Actions'>"
		"      <menuitem action='Preview'/>"
		"      <menuitem action='DefaultSize'/>"
		"      <menuitem action='ShiftLeft'/>"
		"      <menuitem action='ShiftRight'/>"
		"      <menuitem action='Translatables'/>"
		"      <separator/>"
		"      <menuitem action='Default'/>"
		"      <menuitem action='Add'/>"
		"      <menuitem action='Remove'/>"
		"      <menuitem action='MoveUp'/>"
		"      <menuitem action='MoveDown'/>"
		"    </menu>"
		"    <menu action='Help'>"
		"      <menuitem action='Home'/>"
		"      <menuitem action='API'/>"
		"      <separator/>"
		"      <menuitem action='About'/>"
		"    </menu>"
		"  </menubar>"
		"  <toolbar  name='MainToolbar'>"
		"    <toolitem action='Clear'/>"
		"    <toolitem action='Open'/>"
		"    <toolitem action='Save'/>"
		"    <separator/>"
		"    <toolitem action='Cut'/>"
		"    <toolitem action='Copy'/>"
		"    <toolitem action='Paste'/>"
		"    <toolitem action='Delete'/>"
		"    <separator/>"
		"    <toolitem action='Undo'/>"
		"    <toolitem action='Redo'/>"
		"    <separator/>"
		"    <toolitem action='Preview'/>"
		"    <toolitem action='DefaultSize'/>"
		"    <toolitem action='ShiftLeft'/>"
		"    <toolitem action='ShiftRight'/>"
		"    <toolitem action='Translatables'/>"
		"  </toolbar>"
		"  <toolbar name='ExplorerToolbar'>"
		"    <toolitem action='Default'/>"
		"    <separator/>"
		"    <toolitem action='Add'/>"
		"    <toolitem action='Remove'/>"
		"    <toolitem action='MoveUp'/>"
		"    <toolitem action='MoveDown'/>"
		"  </toolbar>"
		"  <popup name='MainPopup'>"
		"    <menuitem action='Cut'/>"
		"    <menuitem action='Copy'/>"
		"    <menuitem action='Delete'/>"
		"    <separator/>"
		"    <menuitem action='DefaultSize'/>"
		"    <menuitem action='ShiftLeft'/>"
		"    <menuitem action='ShiftRight'/>"
		"  </popup>"
		"</ui>"
	;

	try {
		manager->add_ui_from_string(uistr);
	} catch(const Glib::Error & x) {
		FatalError(_("Creating UIManager failed: ") + x.what());
	}

	uiManager= manager;
}

void Application::setupGtkAction(CrowActions actions, CrowActions changed, CrowActions ca, const char * str) {
	if(changed & ca)
		actionGroup->get_action(str)->set_sensitive(actions & ca);
}

void Application::onActionsUpdated(CrowActions changed) {
	CrowActions actions= designer->get_enabled_actions();
	if(changed & CROW_ACTIONS_SAVE)
		setupGtkAction(actions, changed, CROW_ACTION_SAVE, "Save");
	if(changed & CROW_ACTIONS_BASIC) {
		setupGtkAction(actions, changed, CROW_ACTION_UNDO, "Undo");
		setupGtkAction(actions, changed, CROW_ACTION_REDO, "Redo");
		setupGtkAction(actions, changed, CROW_ACTION_CUT, "Cut");
		setupGtkAction(actions, changed, CROW_ACTION_COPY, "Copy");
		setupGtkAction(actions, changed, CROW_ACTION_PASTE, "Paste");
		setupGtkAction(actions, changed, CROW_ACTION_PASTE_TO_ROOT, "PasteToRoot");
		setupGtkAction(actions, changed, CROW_ACTION_DELETE, "Delete");
		setupGtkAction(actions, changed, CROW_ACTION_ABOUT, "About");
		setupGtkAction(actions, changed, CROW_ACTION_PREVIEW, "Preview");
		setupGtkAction(actions, changed, CROW_ACTION_DEFAULT_SIZE, "DefaultSize");
		setupGtkAction(actions, changed, CROW_ACTION_SHIFT_LEFT, "ShiftLeft");
		setupGtkAction(actions, changed, CROW_ACTION_SHIFT_RIGHT, "ShiftRight");
		setupGtkAction(actions, changed, CROW_ACTION_TRANSLATABLES, "Translatables");
	}
	if(changed & CROW_ACTIONS_EXPLORER) {
		setupGtkAction(actions, changed, CROW_ACTION_DEFAULT, "Default");
		setupGtkAction(actions, changed, CROW_ACTION_ADD, "Add");
		setupGtkAction(actions, changed, CROW_ACTION_REMOVE, "Remove");
		setupGtkAction(actions, changed, CROW_ACTION_MOVE_UP, "MoveUp");
		setupGtkAction(actions, changed, CROW_ACTION_MOVE_DOWN, "MoveDown");
	}
}

void Application::onStatusUpdated(ustring_cr str) {
	statusbar.pop();
	statusbar.push(str);
}

void Application::onInfoUpdated(ustring_cr str) {
	infoFrame.property_visible()= !str.empty();
	infoLabel.set_text(str);
}

void Application::onShowUrl(ustring_cr str) {
	showUrl(str);
}

void Application::onClear() {
	if(askProceed()) {
		file.clear();
		designer->clear();
		updateTitle();
	}
}

void Application::onClone() {

	std::vector<std::string> args= startupArgs1;
	for(int i=1; i<SIGNED(startupArgs2.size()); ++i) {
		string_cr arg= startupArgs2[i];
		std::vector<std::string>::iterator it= std::find(args.begin(), args.end(), arg);
		if(it!=args.end())
			args.erase(it);
	}

	try {
		Glib::spawn_async(startupDirectory, args, Glib::SPAWN_SEARCH_PATH);
	} catch(const Glib::Error & x) {
		std::string str= "spawn_async(" + startupDirectory + ",";
		for(int i=0; i<SIGNED(args.size()); ++i)
			str += " " + args[i];
		showMessage(_("'New Window' action failed: ") + x.what() + "\n" + str + ")");
	}
}

void Application::onOpen() {
	if(askProceed())
		openFile(askFilename(Gtk::FILE_CHOOSER_ACTION_OPEN));
}

void Application::onOpenRecent() {
	if(askProceed()) {
		Glib::RefPtr<Gtk::RecentAction> recent= Glib::RefPtr<Gtk::RecentAction>::cast_dynamic(actionGroup->get_action("Recent"));
		std::string uri= recent->get_current_uri();
		if(uri.compare(0, 7, "file://") == 0)
			openFile(uri.substr(7));
		else
			showMessage(_("Invalid URI: ") + uri);
	}
}

void Application::openFile(const File & filename) {
	File name= filename;
	if(!name.empty()) {
		Glib::ustring data;
		bool ok= loadFile(name, &data);
		if(ok) {
			try {
				designer->load(data);
				name.header= data[0]=='"';
				file= name;
				addRecentFile();
			} catch(const Glib::Exception & x) {
				file.clear();
				showMessage(_("Loading failed: ") + x.what());
			}
			updateTitle();
		}
	}
}

void Application::onSave() {
	File name= file;
	if(name.empty())
		name= askFilename(Gtk::FILE_CHOOSER_ACTION_SAVE);
	if(!name.empty()) {
		Glib::ustring data= name.header ? designer->save_header() : designer->save();
		bool ok= saveFile(name, data);
		if(ok) {
			file= name;
			addRecentFile();
			designer->saved();
			updateTitle();
		}
	}
}

void Application::onSaveAs() {
	File name= askFilename(Gtk::FILE_CHOOSER_ACTION_SAVE);
	if(!name.empty()) {
		Glib::ustring data= name.header ? designer->save_header() : designer->save();
		bool ok= saveFile(name, data);
		if(ok) {
			file= name;
			addRecentFile();
			designer->saved();
			updateTitle();
		}
	}
}

void Application::onQuit() {
	if(askProceed())
		window.hide();
}

void Application::onSettings() {
	Glib::ustring path= settings["icon_folders"];
	designer->store_settings(&settings);
	SettingsDialog dialog;
	dialog.getWindow()->set_transient_for(window);
	dialog.setSettings(settings);
	if(dialog.getWindow()->run() == Gtk::RESPONSE_OK) {
		settings= dialog.getSettings();
		if(settings["icon_folders"]!=path)
			updateIconPath();
		if(GetKeyBool(settings, "supp_msg"))
			errorDialog.hide();
		designer->apply_settings(settings);
		designer->store_settings(&settings);
		saveSettings();
	}
}

void Application::onHome() {
	showUrl("http://www.crowdesigner.org");
}

void Application::onApi() {
	int major= GetCrowVersion()/1000000;
	int minor= GetCrowVersion()/1000 - major*1000;
	minor= ((minor + 1) / 2) * 2;
	std::ostringstream oss;
	oss << "http://library.gnome.org/devel/gtk/" << major << "." << minor << "/";
	showUrl(oss.str());
}

bool Application::onDeleteEvent(GdkEventAny *) {
	onQuit();
	return true;
}

void Application::onException() {
	Glib::ustring msg;
	try {
		throw;
	} catch(const Glib::Exception & x) {
		msg= x.what();
	} catch(...) {
		msg= _("no message");
	}
	FatalError(msg);
}

bool Application::loadFile(const File & name, Glib::ustring * data) {
	char * contents;
	if(g_file_get_contents(name.name.c_str(), &contents, NULL, NULL)) {
		*data= contents;
		g_free(contents);
		return true;
	} else {
		showMessage(_("Loading failed"));
		*data= "";
		return false;
	}
}

bool Application::saveFile(const File & name, ustring_cr data) {
	if(g_file_set_contents(name.name.c_str(), data.c_str(), data.bytes(), NULL)) {
		return true;
	} else {
		showMessage(_("Saving failed"));
		return false;
	}
}

Application::File Application::askFilename(Gtk::FileChooserAction action) {
	Gtk::FileChooserDialog dialog(window, _("File Chooser"), action);
	Gtk::CheckButton checkHeader(_("Save as a C header"));

	dialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	if(action==Gtk::FILE_CHOOSER_ACTION_SAVE) {
		checkHeader.set_active(file.header);
		dialog.set_extra_widget(checkHeader);
		dialog.add_button(Gtk::Stock::SAVE, Gtk::RESPONSE_OK);
		dialog.set_do_overwrite_confirmation(true);
		checkHeader.signal_toggled().connect(
			sigc::bind<Gtk::FileChooser *, Gtk::CheckButton *>(
				sigc::mem_fun(*this, &Application::onCheckHeader), &dialog, &checkHeader
			)
		);
	} else
		dialog.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);

	Gtk::FileFilter filter_gui;
	filter_gui.set_name(_("GuiXml files"));
	filter_gui.add_pattern("*.gui");
	filter_gui.add_pattern("*.gui.h");
	dialog.add_filter(filter_gui);

	Gtk::FileFilter filter_any;
	filter_any.set_name(_("Any files"));
	filter_any.add_pattern("*");
	dialog.add_filter(filter_any);

	if(!file.empty())
		dialog.set_filename(file.name);
	else
		dialog.set_current_folder(Glib::get_home_dir());

	if(dialog.run() == Gtk::RESPONSE_OK) {
		File F(dialog.get_filename(), checkHeader.get_active());
		if(action==Gtk::FILE_CHOOSER_ACTION_SAVE)
			if(Glib::path_get_basename(F.name).find('.') == std::string::npos)
				F.name += F.header ? ".gui.h" : ".gui";
		return F;
	} else
		return File();
}

void Application::onCheckHeader(Gtk::FileChooser * chooser, Gtk::CheckButton * button) {
	bool header= button->get_active();
	std::string name= Glib::path_get_basename(chooser->get_filename());
	if(header && g_str_has_suffix(name.c_str(), ".gui")) {
		name += ".h";
		chooser->set_current_name(name);
	} else if(!header && g_str_has_suffix(name.c_str(), ".gui.h")) {
		name.resize(name.size() - 2);
		chooser->set_current_name(name);
	}
}

bool Application::askProceed() {
	if(designer->get_enabled_actions() & CROW_ACTION_SHOULD_SAVE) {
		Glib::ustring msg= file.empty() ? _("\nAbandon new document?\n") : _("\nAbandon changes?\n");
		Gtk::MessageDialog dialog(window, msg, false, Gtk::MESSAGE_WARNING, Gtk::BUTTONS_YES_NO);
		dialog.set_title(_("Document is not saved"));
		return dialog.run() == Gtk::RESPONSE_YES;
	} else
		return true;
}

void Application::updateTitle() {
	Glib::ustring name= file.empty() ? "" : " - " + Glib::filename_to_utf8(file.name);
	window.set_title(_("Crow") + name);
}

std::vector<Glib::ustring> Application::currentIconPath() {
	return Gtk::IconTheme::get_default()->get_search_path();
}

void Application::updateIconPath() {
	StringList custom= Split(settings["icon_folders"], "*");
	std::vector<Glib::ustring> path;
	path.reserve(defaultIconPath.size() + custom.size());
	path.insert(path.end(), defaultIconPath.begin(), defaultIconPath.end());
	path.insert(path.end(), custom.begin(), custom.end());
	if(path != currentIconPath())
		Gtk::IconTheme::get_default()->set_search_path(path);
}

std::string Application::getConfigPath() {
	return Glib::build_filename(Glib::get_user_config_dir(), "crow.conf");
}

void Application::loadKeys(Glib::KeyFile * file, Settings * skeys) {
	try {
		file->load_from_file(getConfigPath());
		std::vector<Glib::ustring> keys= file->get_keys("settings");
		for(int i=0; i<SIGNED(keys.size()); ++i) {
			Glib::ustring str= file->get_string("settings", keys[i]);
			(*skeys)[keys[i]]= str;
		}
	} catch(const Glib::Exception &) { }
}

void Application::loadSettings() {

	Glib::KeyFile file;
	loadKeys(&file, &settings);

	EnsureKeyString(&settings, "$window_width", "800");
	EnsureKeyString(&settings, "$window_height", "450");
	EnsureKeyString(&settings, "$window_maximized", "0");
	EnsureKeyString(&settings, "$hpaned", "160");
	EnsureKeyString(&settings, "$hpaned_ce", "400");
	EnsureKeyString(&settings, "$vpaned_hp", "200");
	EnsureKeyString(&settings, "$hier_split", "90");
	EnsureKeyString(&settings, "$expl_split", "90");

	EnsureKeyString(&settings, "supp_msg", "false");

	updateIconPath();

}

void Application::saveSettings(bool quit) {

	AppWindow::State state= window.getState();
	SetKeyInt(&settings, "$window_width", state.width);
	SetKeyInt(&settings, "$window_height", state.height);
	SetKeyInt(&settings, "$window_maximized", state.maximized);

	Glib::KeyFile file;
	Settings skeys;
	loadKeys(&file, &skeys);

	if(quit)
		for(Settings::iterator it= settings.begin(); it!=settings.end(); ++it) {
			CHECK(!it->first.empty());
			if(it->first[0] != '$') {
				Settings::iterator jt= skeys.find(it->first);
				if(jt!=skeys.end())
					it->second= jt->second;
			}
		}

	if(skeys == settings)
		return;

	Glib::ustring path= getConfigPath();
	for(Settings::iterator it= settings.begin(); it!=settings.end(); ++it)
		file.set_string("settings", it->first, it->second);
	Glib::ustring data= file.to_data();
	g_mkdir_with_parents(Glib::path_get_dirname(path).c_str(), ModeDirPrivate);
	if(!g_file_set_contents(path.c_str(), data.c_str(), data.size(), NULL))
		showMessage(_("Saving settings failed"));

}

void Application::addRecentFile() {
	Gtk::RecentManager::get_default()->add_item("file://" + file.name);
}

void Application::showMessage(ustring_cr message) {
	Gtk::MessageDialog dialog(window, message);
	dialog.set_title(_("Crow"));
	dialog.run();
}

void Application::showUrl(ustring_cr url) {
	#ifdef CROW_WIN32
		HINSTANCE result= ShellExecute(NULL, NULL, url.c_str(), NULL, NULL, SW_SHOWNORMAL);
		if(reinterpret_cast<UINT_PTR>(result) <= 32)
			showMessage(_("Error opening url:\n") + url);
	#else
		gboolean result= gtk_show_uri(NULL, url.c_str(), GDK_CURRENT_TIME, NULL);
		if(!result)
			showMessage(_("Error opening url:\n") + url);
	#endif
}

void Application::designer_run_action(CrowActions action) {
	designer->run_action(action);
}

}
