/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#ifndef CROWAPP_APPLICATION_H
#define CROWAPP_APPLICATION_H

#include "util.h"

namespace CrowApp {

class ErrorDialog : public Gtk::Window {

public:
	ErrorDialog();

	void onErrorMessage(ustring_cr message);

private:
	Gtk::ScrolledWindow scroll;
	Gtk::TextView textView;

};

class SettingsDialog : public sigc::trackable {

public:
	SettingsDialog();

	void setSettings(const Settings & settings);
	Settings getSettings();

	Gtk::Dialog * getWindow() { return window; }

protected:
	void onAdd();
	void onRemove();
	void onMoveUp();
	void onMoveDown();
	void onColorChanged();
	void onCheckChanged();
	void onSelectionChanged();
	void onSwitchPage(GtkNotebookPage *, guint);
	void onReset();

	void updateUi(const Settings & settings);
	void updateButtons();
	void updateActions();

	Gtk::TreeIter getPosition();
	void actionCompleted();

private:
	Gui::PLoader loader;

	Settings original, defaults;

	Gtk::Dialog * window;
	Gtk::Notebook * widget;
	Gtk::ColorButton * validColor;
	Gtk::ColorButton * wrongColor;
	Gtk::CheckButton * showDebugWidgets;
	Gtk::CheckButton * showHiddenProperties;
	Gtk::CheckButton * suppressMessages;
	Glib::RefPtr<Gtk::ActionGroup> actions;
	Gtk::TreeView * folders;

	Glib::RefPtr<Gtk::ListStore> model;

};

class AppWindow : public Gtk::Window {

public:
	struct State {
		State() { width= height= maximized= -1; }
		int width;
		int height;
		int maximized;
	};

	State getState();
	void setState(const State & state);

protected:
	int getMaximized();
	Glib::RefPtr<Gdk::Window> getGdkWindow();

};

class Application : public sigc::trackable {

public:
	Application(const std::vector<std::string> & args1, const std::vector<std::string> & args2);
	~Application();

	Gtk::Window * getWindow() { return &window; }

protected:
	struct File {
		File() { header= false; }
		File(string_cr fn, bool hdr=false) { name= fn; header= hdr; }
		bool empty() { return name.empty(); }
		void clear() { name.clear(); header= false; }
		std::string name;
		bool header;
	};

	void setupHandler(const char * domain);

	static void logHandler(
		const char * logDomain,
		GLogLevelFlags logLevel,
		const char * message,
		gpointer userData
	);

	void onErrorMessage(
		ustring_cr domain,
		GLogLevelFlags level,
		ustring_cr message
	);

	void createActionGroup();
	void createUIManager();
	void setupGtkAction(CrowActions actions, CrowActions changed, CrowActions ca, const char * str);

	void onActionsUpdated(CrowActions changed);
	void onStatusUpdated(ustring_cr str);
	void onInfoUpdated(ustring_cr str);
	void onShowUrl(ustring_cr str);

	void onClear();
	void onClone();
	void onOpen();
	void onOpenRecent();
	void openFile(const File & filename);
	void onSave();
	void onSaveAs();
	void onQuit();
	void onSettings();
	void onHome();
	void onApi();

	bool onDeleteEvent(GdkEventAny *);
	void onException();

	bool loadFile(const File & name, Glib::ustring * data);
	bool saveFile(const File & name, ustring_cr data);
	File askFilename(Gtk::FileChooserAction action);
	void onCheckHeader(Gtk::FileChooser * chooser, Gtk::CheckButton * button);
	bool askProceed();
	void updateTitle();

	std::vector<Glib::ustring> currentIconPath();
	void updateIconPath();
	std::string getConfigPath();
	void loadKeys(Glib::KeyFile * file, Settings * skeys);
	void loadSettings();
	void saveSettings(bool quit=false);

	void addRecentFile();
	void showMessage(ustring_cr message);
	void showUrl(ustring_cr url);

	void designer_run_action(CrowActions action);

private:
	std::vector<Glib::ustring> defaultIconPath;
	Settings settings;
	ErrorDialog errorDialog;

	typedef std::list<std::pair<const char *, guint> > Hids;
	Hids hids;

	std::string startupDirectory;
	std::vector<std::string> startupArgs1, startupArgs2;

	File file;

	AppWindow window;

	Glib::RefPtr<Gtk::UIManager> uiManager;
	Glib::RefPtr<Gtk::ActionGroup> actionGroup;

	Gtk::VBox vBox;
	Gtk::Statusbar statusbar;
	Gtk::Frame infoFrame;
	Gtk::Label infoLabel;

	PDesigner designer;

};

}

#endif
