/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "container.h"

namespace CrowSvc {

#include "glue.h"
#include "marshal.h"

// -----------------------------------------------------------------------------

GType crow_container_get_type(void) {
	static GType object_type= 0;
	if(object_type==0) {
		static const GTypeInfo object_info= {
			sizeof(CrowContainerClass),
			NULL,
			NULL,
			(GClassInitFunc)crow_container_class_init,
			NULL,
			NULL,
			sizeof(CrowContainer),
			0,
			(GInstanceInitFunc)crow_container_init,
		};
	  object_type= g_type_register_static(G_TYPE_OBJECT, "CrowContainer", &object_info, (GTypeFlags)0);
	}
	return object_type;
}

void crow_container_class_init(CrowContainerClass * meta) {
	GObjectClass * object_class;
	meta->parent_class= (GObjectClass *)g_type_class_peek_parent(meta);
	object_class= G_OBJECT_CLASS(meta);
	object_class->finalize= crow_container_finalize;

	g_signal_new(
		"actions_updated",
		G_OBJECT_CLASS_TYPE(object_class),
		G_SIGNAL_RUN_LAST, 0,
		NULL, NULL,
		crow_marshal_VOID__INT_INT64,
		G_TYPE_NONE, 2,
		G_TYPE_INT, G_TYPE_INT64
	);
	g_signal_new(
		"status_updated",
		G_OBJECT_CLASS_TYPE(object_class),
		G_SIGNAL_RUN_LAST, 0,
		NULL, NULL,
		crow_marshal_VOID__INT_STRING,
		G_TYPE_NONE, 2,
		G_TYPE_INT, G_TYPE_STRING
	);
	g_signal_new(
		"info_updated",
		G_OBJECT_CLASS_TYPE(object_class),
		G_SIGNAL_RUN_LAST, 0,
		NULL, NULL,
		crow_marshal_VOID__INT_STRING,
		G_TYPE_NONE, 2,
		G_TYPE_INT, G_TYPE_STRING
	);
	g_signal_new(
		"show_url",
		G_OBJECT_CLASS_TYPE(object_class),
		G_SIGNAL_RUN_LAST, 0,
		NULL, NULL,
		crow_marshal_VOID__INT_STRING,
		G_TYPE_NONE, 2,
		G_TYPE_INT, G_TYPE_STRING
	);
	g_signal_new(
		"modal_enter",
		G_OBJECT_CLASS_TYPE(object_class),
		G_SIGNAL_RUN_LAST, 0,
		NULL, NULL,
		g_cclosure_marshal_VOID__INT,
		G_TYPE_NONE, 1,
		G_TYPE_INT
	);
	g_signal_new(
		"modal_leave",
		G_OBJECT_CLASS_TYPE(object_class),
		G_SIGNAL_RUN_LAST, 0,
		NULL, NULL,
		g_cclosure_marshal_VOID__INT,
		G_TYPE_NONE, 1,
		G_TYPE_INT
	);

}

void crow_container_init(CrowContainer * obj) {
	ContainerInit(obj);
}

void crow_container_finalize(GObject * object) {
	CrowContainer * container= CROW_CONTAINER(object);
	CrowContainerClass * meta= CROW_CONTAINER_GET_CLASS(container);
	ContainerDone(container);
	(*meta->parent_class->finalize)(object);
}

// Container -------------------------------------------------------------------

Container::Container() {
	counter= 1;
	iconPath= GetIconSearchPath();
}

Container::~Container() {
	if(!designers.empty())
		std::cout << "~Container: designers.size() = " << designers.size() << std::endl;
}

PDesigner Container::lookup(int id) {
	Table::iterator it= designers.find(id);
	CHECK(it!=designers.end());
	return it->second;
}

void Container::remove(int id) {
	Table::iterator it= designers.find(id);
	CHECK(it!=designers.end());
	freePlug(it->second);
	designers.erase(it);
}

int Container::insert(PDesigner designer) {
	int id= counter;
	++counter;
	designers[id]= designer;
	return id;
}

void Container::freePlug(PDesigner designer) {
	Gtk::Plug * plug= dynamic_cast<Gtk::Plug *>(designer->get_widget()->get_parent());
	delete plug;
}

// -----------------------------------------------------------------------------

CrowContainer *& ContainerPtr() {
	static CrowContainer * ptr= NULL;
	return ptr;
}

CrowContainer * ContainerGet() {
	return ContainerPtr();
}

void ContainerEnter() {
	dbus_g_object_type_install_info(CROW_TYPE_CONTAINER, &dbus_glib_crow_container_object_info);
	dbus_g_error_domain_register(CROW_ERROR, NULL, CROW_TYPE_ERROR);
	ContainerPtr()= (CrowContainer *)g_object_new(CROW_TYPE_CONTAINER, NULL);
}

void ContainerLeave() {
	g_object_unref(ContainerGet());
	ContainerPtr()= NULL;
}

void ContainerInit(CrowContainer * obj) {
	obj->container= new Container();
}

void ContainerDone(CrowContainer * obj) {
	delete obj->container;
	#ifdef CROW_SANDBOX
		std::cout << "ContainerDone()" << std::endl;
	#endif
}

PDesigner MakeDesigner(GdkNativeWindow handle) {
	PDesigner designer= CreateDesigner();
	Gtk::Widget * widget= designer->get_widget();
	Gtk::Plug * plug= new Gtk::Plug(handle);
	plug->add(*widget);
	return designer;
}

PDesigner FindDesigner(CrowContainer * obj, int id) {
	CHECK(obj && id > 0);
	return obj->container->lookup(id);
}

void OnActionsUpdated(CrowActions changed, int id, CrowContainer * container) {
	g_signal_emit_by_name(container, "actions-updated", id, changed);
}

void OnStatusUpdated(ustring_cr status, int id, CrowContainer * container) {
	g_signal_emit_by_name(container, "status-updated", id, status.c_str());
}

void OnInfoUpdated(ustring_cr info, int id, CrowContainer * container) {
	g_signal_emit_by_name(container, "info-updated", id, info.c_str());
}

void OnShowUrl(ustring_cr url, int id, CrowContainer * container) {
	g_signal_emit_by_name(container, "show-url", id, url.c_str());
}

void OnModalEnter(int id, CrowContainer * container) {
	g_signal_emit_by_name(container, "modal-enter", id);
}

void OnModalLeave(int id, CrowContainer * container) {
	g_signal_emit_by_name(container, "modal-leave", id);
}

gboolean crow_container_get_version(CrowContainer * obj, int * version, GError ** error) {
	*version= GetCrowVersion();
	return TRUE;
}

gboolean crow_container_shutdown(CrowContainer * obj, GError ** error) {
	Gtk::Main::quit();
	return TRUE;
}

gboolean crow_container_create(CrowContainer * obj, gint64 handle, int * id, GError ** error) {
	PDesigner designer= MakeDesigner((GdkNativeWindow)handle);
	*id= obj->container->insert(designer);
	designer->connect_actions_updated(sigc::bind<int, CrowContainer *>(sigc::ptr_fun(&OnActionsUpdated), *id, obj));
	designer->connect_status_updated(sigc::bind<int, CrowContainer *>(sigc::ptr_fun(&OnStatusUpdated), *id, obj));
	designer->connect_info_updated(sigc::bind<int, CrowContainer *>(sigc::ptr_fun(&OnInfoUpdated), *id, obj));
	designer->connect_show_url(sigc::bind<int, CrowContainer *>(sigc::ptr_fun(&OnShowUrl), *id, obj));
	designer->connect_modal_enter(sigc::bind<int, CrowContainer *>(sigc::ptr_fun(&OnModalEnter), *id, obj));
	designer->connect_modal_leave(sigc::bind<int, CrowContainer *>(sigc::ptr_fun(&OnModalLeave), *id, obj));
	return TRUE;
}

gboolean crow_container_remove(CrowContainer * obj, int id, GError ** error) {
	obj->container->remove(id);
	return TRUE;
}

gboolean crow_container_apply_settings(CrowContainer * obj, int id, GHashTable * settings, GError ** error) {
	PDesigner designer= FindDesigner(obj, id);
	Gtk::Plug * plug= dynamic_cast<Gtk::Plug *>(designer->get_widget()->get_parent());
	bool visible= plug->is_visible();

	Settings set= HashTableToSettings(settings);

	Glib::ustring folders= GetKeyString(set, "icon_folders");
	StringSet custom= SplitToSet(folders, "*");
	StringSet path= SetUnion(*obj->container->defIconPath(), custom);
	StringSet current= GetIconSearchPath();
	if(path != current)
		Gtk::IconTheme::get_default()->set_search_path(path);

	if(!visible) {
		int winw= GetKeyInt(set, "$window_width");
		int winh= GetKeyInt(set, "$window_height");
		if(winw > -1 && winh > -1)
			plug->set_default_size(winw, winh);
	}

	designer->apply_settings(set);

	if(!visible)
		plug->show();

	return TRUE;
}

gboolean crow_container_store_settings(CrowContainer * obj, int id, GHashTable * original, GHashTable ** settings, GError ** error) {
	Settings set= HashTableToSettings(original);
	FindDesigner(obj, id)->store_settings(&set);
	*settings= SettingsToHashTable(set);
	return TRUE;
}

gboolean crow_container_clear(CrowContainer * obj, int id, GError ** error) {
	FindDesigner(obj, id)->clear();
	return TRUE;
}

gboolean crow_container_load(CrowContainer * obj, int id, const char * data, GError ** error) {
	try {
		FindDesigner(obj, id)->load(data);
		return TRUE;
	} catch(const Glib::Exception & x) {
		g_set_error(error, CROW_ERROR, CROW_ERROR_RUNTIME, x.what().c_str());
		return FALSE;
	}
}

gboolean crow_container_save(CrowContainer * obj, int id, char ** data, GError ** error) {
	*data= g_strdup(FindDesigner(obj, id)->save().c_str());
	return TRUE;
}

gboolean crow_container_save_header(CrowContainer * obj, int id, char ** data, GError ** error) {
	*data= g_strdup(FindDesigner(obj, id)->save_header().c_str());
	return TRUE;
}

gboolean crow_container_saved(CrowContainer * obj, int id, GError ** error) {
	FindDesigner(obj, id)->saved();
	return TRUE;
}

gboolean crow_container_get_enabled_actions(CrowContainer * obj, int id, gint64 * actions, GError ** error) {
	*actions= FindDesigner(obj, id)->get_enabled_actions();
	return TRUE;
}

gboolean crow_container_run_action(CrowContainer * obj, int id, gint64 action, GError ** error) {
	FindDesigner(obj, id)->run_action(action);
	return TRUE;
}

}
