/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "container.h"

namespace CrowSvc {

void OnNewConnection(
	DBusServer * server,
	DBusConnection * connection,
	void * data
) {

	static gboolean connected= FALSE;

	if(!connected) {

		connected= TRUE;

		dbus_connection_ref(connection);
		dbus_connection_setup_with_g_main(connection, NULL);

		DBusGConnection * bus= dbus_connection_get_g_connection(connection);
		CrowContainer * obj= ContainerGet();

		dbus_g_connection_register_g_object(bus, "/CrowContainer", G_OBJECT(obj));

	} else {

		dbus_connection_close(connection);

	}

}

int Main(int argc, char ** argv) {

	Gtk::Main gtkmain(&argc, &argv);

	#if ENABLE_NLS
		bindtextdomain(PACKAGE, LOCALEDIR);
		bind_textdomain_codeset(PACKAGE, "UTF-8");
		textdomain(PACKAGE);
	#endif

	Glib::ustring address;
	Glib::OptionContext context(_("- Crow Designer DBus Service"));
	Glib::OptionGroup main("", "");

	Glib::OptionEntry entryAddress;
	entryAddress.set_long_name("address");
	entryAddress.set_short_name('a');
	entryAddress.set_description(_("specify DBus address"));
	entryAddress.set_arg_description("arg");
	main.add_entry(entryAddress, address);

	context.set_main_group(main);

	try {
		context.parse(argc, argv);
	} catch(const Glib::Error & e) {
		std::cerr << _("Error parsing program options: ") << e.what() << std::endl;
		return EXIT_FAILURE;
	}

	if(address.empty()) {
		std::cerr << _("Missing DBus address") << std::endl;
		return EXIT_FAILURE;
	}

	Crow::Enter enter;
	ContainerBlock block;

	DBusError error;
	dbus_error_init(&error);
	DBusServer * server= dbus_server_listen(address.c_str(), &error);

	if(!server) {
		if(dbus_error_has_name(&error, DBUS_ERROR_ADDRESS_IN_USE))
			std::cerr << "DBUS_ERROR_ADDRESS_IN_USE: " << error.message << std::endl;
		else
			std::cerr << _("Creating server failed: ") << error.message << std::endl;
		return EXIT_FAILURE;
	}

	dbus_server_set_new_connection_function(server, OnNewConnection, NULL, NULL);
	dbus_server_setup_with_g_main(server, NULL);

	std::cout << address << std::endl;
	gtkmain.run();

	dbus_server_disconnect(server);
	dbus_server_unref(server);

	return EXIT_SUCCESS;
}

}

int main(int argc, char ** argv)
	{ return CrowSvc::Main(argc, argv); }
