/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "util.h"

namespace CrowSvc {

CROW_API GQuark crow_error_quark() {
	static GQuark quark= 0;
	if(!quark)
		quark= g_quark_from_static_string("crow_error");
	return quark;
}

CROW_API GType crow_error_get_type() {
	static GType etype= 0;
	if(etype == 0) {
		static const GEnumValue values[]= {
			{ CROW_ERROR_RUNTIME, "CROW_ERROR_RUNTIME", "Runtime" },
			{ 0, NULL, NULL }
		};
		etype= g_enum_register_static("CrowError", values);
	}
	return etype;
}

void CheckFailed(const char * expr, const char * file, int line) {
	std::ostringstream oss;
	oss << boost::format(_("Check failed: %s, file: %s, line: %d")) % expr % file % line;
	FatalError(oss.str());
}

void FatalError(string_cr message) {
	g_critical(message.c_str());
	#ifdef CROW_SANDBOX
		assert(0);
	#endif
	Gtk::MessageDialog dialog(
		_("Fatal Error") + std::string(" [" G_LOG_DOMAIN "]"), false,
		Gtk::MESSAGE_ERROR, Gtk::BUTTONS_CLOSE, true
	);
	dialog.set_secondary_text(message);
	dialog.set_keep_above(true);
	dialog.run();
	abort();
}

GHashTable * SettingsToHashTable(const Settings & settings) {
	GHashTable * table= g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
	for(Settings::const_iterator it= settings.begin(); it!=settings.end(); ++it) {
		char * key= g_strdup(it->first.c_str());
		char * value= g_strdup(it->second.c_str());
		g_hash_table_insert(table, key, value);
	}
	return table;
}

Settings HashTableToSettings(GHashTable * table) {
	Settings result;
	struct tmp { static void add(gpointer key, gpointer value, gpointer user_data) {
		(*static_cast<Settings *>(user_data))[static_cast<char *>(key)]= static_cast<char *>(value);
	} };
	g_hash_table_foreach(table, &tmp::add, &result);
	return result;
}

StringSet GetIconSearchPath() {
	Glib::ArrayHandle<Glib::ustring> list= Gtk::IconTheme::get_default()->get_search_path();
	return StringSet(list.begin(), list.end());
}

StringSet SplitToSet(string_cr str, const char * dlm) {
	StringSet result;
	typedef boost::tokenizer<boost::char_separator<char> > Tok;
	boost::char_separator<char> sep(dlm, "", boost::keep_empty_tokens);
	Tok tok(str, sep);
	for(Tok::iterator it= tok.begin(); it!=tok.end(); ++it)
		result.insert(*it);
	return result;
}

Glib::ustring GetKeyString(const Settings & settings, string_cr key) {
	Settings::const_iterator it= settings.find(key);
	return it!=settings.end() ? it->second : "";
}

int GetKeyInt(const Settings & settings, string_cr key) {
	Settings::const_iterator it= settings.find(key);
	if(it!=settings.end()) {
		std::istringstream iss(it->second);
		int value;
		iss >> value;
		if(!iss.fail())
			return value;
	}
	return -1;
}

}
